Package: a5R
Title: 'A5' Discrete Global Grid System
Version: 0.5.0
Authors@R: c(
    person("Hugh", "Graham", role = c("aut", "cre"),
           email = "hugh@belian.earth"),
    person("belian.earth", role = "cph")
    )
Description: Bindings for the "A5 geospatial index"
    <https://a5geo.org/>. 'A5' partitions the Earth's surface into
    pentagonal cells across 31 resolution levels using an equal-area
    projection onto a dodecahedron. Provides functions for indexing
    coordinates to cells, traversing the cell hierarchy, computing cell
    boundaries, and compacting/uncompacting cell sets. Powered by the
    'A5' 'Rust' crate via 'extendr'.
License: Apache License (>= 2)
URL: https://github.com/belian-earth/a5R,
        https://belian-earth.github.io/a5R/
BugReports: https://github.com/belian-earth/a5R/issues
Depends: R (>= 4.2)
Imports: cli, rlang (>= 1.1.0), units, vctrs (>= 0.6.0), wk (>= 0.9.0)
Suggests: arrow, knitr, pillar, rmarkdown, sf, terra, testthat (>=
        3.0.0), tibble, withr
VignetteBuilder: knitr
Config/testthat/edition: 3
Config/rextendr/version: 0.5.0
SystemRequirements: Cargo (Rust's package manager), rustc
Encoding: UTF-8
RoxygenNote: 8.0.0
Config/roxygen2/version: 8.0.0
NeedsCompilation: yes
Packaged: 2026-07-01 21:26:54 UTC; hugh
Author: Hugh Graham [aut, cre],
  belian.earth [cph]
Maintainer: Hugh Graham <hugh@belian.earth>
Repository: CRAN
Date/Publication: 2026-07-02 03:20:30 UTC
Built: R 4.6.1; aarch64-apple-darwin23; 2026-07-02 08:02:09 UTC; unix
Archs: a5R.so.dSYM

Package: a5R
Title: 'A5' Discrete Global Grid System
Version: 0.4.0
Authors@R: c(
    person("Hugh", "Graham", role = c("aut", "cre"),
           email = "hugh@belian.earth"),
    person("belian.earth", role = "cph")
    )
Description: Bindings for the "A5 geospatial index"
    <https://a5geo.org/>. 'A5' partitions the Earth's surface into
    pentagonal cells across 31 resolution levels using an equal-area
    projection onto a dodecahedron. Provides functions for indexing
    coordinates to cells, traversing the cell hierarchy, computing cell
    boundaries, and compacting/uncompacting cell sets. Powered by the
    'A5' 'Rust' crate via 'extendr'.
License: Apache License (>= 2)
URL: https://github.com/belian-earth/a5R,
        https://belian-earth.github.io/a5R/
BugReports: https://github.com/belian-earth/a5R/issues
Depends: R (>= 4.2)
Imports: cli, lifecycle, rlang (>= 1.1.0), units, vctrs (>= 0.6.0), wk
        (>= 0.9.0)
Suggests: arrow, knitr, pillar, rmarkdown, sf, terra, testthat (>=
        3.0.0), tibble, withr
VignetteBuilder: knitr
Config/testthat/edition: 3
Config/rextendr/version: 0.5.0
SystemRequirements: Cargo (Rust's package manager), rustc
Encoding: UTF-8
RoxygenNote: 8.0.0
Config/roxygen2/version: 8.0.0
NeedsCompilation: yes
Packaged: 2026-05-14 12:38:49 UTC; hugh
Author: Hugh Graham [aut, cre],
  belian.earth [cph]
Maintainer: Hugh Graham <hugh@belian.earth>
Repository: CRAN
Date/Publication: 2026-05-14 13:00:07 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-05-14 15:41:17 UTC; unix
Archs: a5R.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(format,a5_cell)
S3method(is.na,a5_cell)
S3method(pillar::pillar_shaft,a5_cell)
S3method(vctrs::vec_proxy_compare,a5_cell)
S3method(vctrs::vec_proxy_order,a5_cell)
S3method(vctrs::vec_ptype_abbr,a5_cell)
S3method(vctrs::vec_ptype_full,a5_cell)
S3method(vec_cast,a5_cell.a5_cell)
S3method(vec_cast,a5_cell.character)
S3method(vec_cast,character.a5_cell)
S3method(vec_ptype2,a5_cell.a5_cell)
S3method(vec_ptype2,a5_cell.character)
S3method(vec_ptype2,character.a5_cell)
S3method(wk::wk_crs,a5_cell)
S3method(wk::wk_handle,a5_cell)
export(a5_cell)
export(a5_cell_area)
export(a5_cell_distance)
export(a5_cell_from_arrow)
export(a5_cell_to_arrow)
export(a5_cell_to_boundary)
export(a5_cell_to_children)
export(a5_cell_to_lonlat)
export(a5_cell_to_parent)
export(a5_compact)
export(a5_get_num_cells)
export(a5_get_num_children)
export(a5_get_res0_cells)
export(a5_get_resolution)
export(a5_get_threads)
export(a5_grid_disk)
export(a5_hex_to_u64)
export(a5_is_valid)
export(a5_linestring_to_cells)
export(a5_lonlat_to_cell)
export(a5_polygon_to_cells)
export(a5_set_threads)
export(a5_spherical_cap)
export(a5_u64_to_hex)
export(a5_uncompact)
export(as_a5_cell)
export(is_a5_cell)
importFrom(rlang,caller_env)
importFrom(vctrs,vec_cast)
importFrom(vctrs,vec_ptype2)
useDynLib(a5R, .registration = TRUE)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# a5R <a href="https://belian-earth.github.io/a5R/"><img src="man/figures/logo.png" align="right" height="200" alt="a5R website" /></a>

<!-- badges: start -->

[![R-CMD-check](https://github.com/belian-earth/a5R/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/belian-earth/a5R/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/belian-earth/a5R/graph/badge.svg)](https://app.codecov.io/gh/belian-earth/a5R)
[![extendr](https://img.shields.io/badge/extendr-%5E0.8.1-276DC2)](https://extendr.rs/extendr/extendr_api/)
[![Lifecycle:
stable](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html#stable)
[![License:Apache](https://img.shields.io/github/license/belian-earth/a5R)](https://www.apache.org/licenses/LICENSE-2.0)
[![Github
stars](https://img.shields.io/github/stars/belian-earth/a5R?style=social)](https://github.com/belian-earth/a5R/stargazers)
[![GitHub
Issues](https://img.shields.io/github/issues/belian-earth/a5R)](https://github.com/belian-earth/a5R/issues)
[![CRAN
status](https://www.r-pkg.org/badges/version/a5R)](https://CRAN.R-project.org/package=a5R)
<!-- badges: end -->

a5R provides R bindings for the [A5](https://a5geo.org/) pentagonal
geospatial index, powered by the [a5 Rust
crate](https://github.com/felixpalmer/a5-rs) via
[extendr](https://extendr.rs/extendr/extendr_api/).

A5 partitions the Earth’s surface into pentagonal cells across 31
resolution levels. Cells are equal-area, encoded as 64-bit integers, and
achieve millimetre-level precision at the finest resolution.

## Installation

``` r
install.packages("a5R")
```

Or install the development version from GitHub:

``` r
# install.packages("pak")
pak::pak("belian-earth/a5R")
```

You will need a working [Rust toolchain](https://rustup.rs/) (`cargo`
and `rustc`).

## Quick example

``` r
library(a5R)

# Index a point to a cell
cell <- a5_lonlat_to_cell(-3.19, 55.95, resolution = 10)
cell
#> <a5_cell[1]>
#> [1] 6344be8000000000

# Get the boundary polygon
a5_cell_to_boundary(cell)
#> <wk_wkb[1] with CRS=OGC:CRS84>
#> [1] <POLYGON ((-3.175718 55.93546, -3.145905 55.97569, -3.151641 56.01921, -3.219413 56.00818, -3.226037 55.96443, -3.175718 55.93546...>

# Navigate the hierarchy
a5_cell_to_parent(cell)
#> <a5_cell[1]>
#> [1] 6344be0000000000
a5_cell_to_children(cell)
#> <a5_cell[4]>
#> [1] 6344be2000000000 6344be6000000000 6344bea000000000 6344bee000000000
```

``` r
# Create a collection of cells whose centres fall within a great-circle distance of 100km from the origin cell
cells <- a5_spherical_cap(cell, radius = 100000) |> 
  a5_uncompact(resolution = 10)
plot(a5_cell_to_boundary(cells), col = "#206ead20", border = "#206ead", asp = 1)
```

<img src="man/figures/README-grid-plot-1.png" alt="A5 grid plot showing a collection of cells around a point" width="100%" />

See `vignette("a5R")` for a full walkthrough of indexing, boundaries,
hierarchy, traversal, and grid generation.

## Features

- **Vectorised Rust core** — all operations implemented in Rust and
  called from R via extendr.
  [Benchmarks](https://github.com/belian-earth/a5R/blob/main/benchmarks/RESULTS.md)
  confirm identical results to the Python, JavaScript, and DuckDB A5
  implementations.
- **vctrs + wk integration** — cell indices use an `a5_cell` type with
  tibble support; geometries return as `wk_wkb`/`wk_wkt` vectors
  compatible with sf and terra.
- **Grid generation** — `a5_grid()` fills any bounding box or geometry
  with cells at a target resolution using hierarchical flood-fill.
- **Traversal** — `a5_grid_disk()` and `a5_spherical_cap()` select
  neighbours by hop count or great-circle distance.
- **Multi-threading** — opt-in parallel processing via rayon for
  vectorised operations. See `vignette("multithreading")`.

## Acknowledgements

A5 was created by [Felix Palmer](https://github.com/felixpalmer). This
package is a thin R wrapper around his work and would not exist without
it. The [Query-farm](https://github.com/Query-farm/a5) team maintain the
DuckDB A5 extension, which wraps the same Rust crate and provided a
valuable reference for this project.

<!-- README.md is generated from README.Rmd. Please edit that file -->

# a5R <a href="https://belian-earth.github.io/a5R/"><img src="man/figures/logo.png" align="right" height="200" alt="a5R website" /></a>

<!-- badges: start -->

[![R-CMD-check](https://github.com/belian-earth/a5R/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/belian-earth/a5R/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/belian-earth/a5R/graph/badge.svg)](https://app.codecov.io/gh/belian-earth/a5R)
[![extendr](https://img.shields.io/badge/extendr-%5E0.8.1-276DC2)](https://extendr.rs/extendr/extendr_api/)
[![Lifecycle:
stable](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html#stable)
[![License:Apache](https://img.shields.io/github/license/belian-earth/a5R)](https://www.apache.org/licenses/LICENSE-2.0)
[![Github
stars](https://img.shields.io/github/stars/belian-earth/a5R?style=social)](https://github.com/belian-earth/a5R/stargazers)
[![GitHub
Issues](https://img.shields.io/github/issues/belian-earth/a5R)](https://github.com/belian-earth/a5R/issues)
[![CRAN
status](https://www.r-pkg.org/badges/version/a5R)](https://CRAN.R-project.org/package=a5R)
<!-- badges: end -->

a5R provides R bindings for the [A5](https://a5geo.org/) pentagonal
geospatial index, powered by the [a5 Rust
crate](https://github.com/felixpalmer/a5-rs) via
[extendr](https://extendr.rs/extendr/extendr_api/).

A5 partitions the Earth’s surface into pentagonal cells across 31
resolution levels. Cells are equal-area, encoded as 64-bit integers, and
achieve millimetre-level precision at the finest resolution.

## Installation

``` r
install.packages("a5R")
```

Or install the development version from GitHub:

``` r
# install.packages("pak")
pak::pak("belian-earth/a5R")
```

You will need a working [Rust toolchain](https://rustup.rs/) (`cargo`
and `rustc`).

## Quick example

``` r
library(a5R)

# Index a point to a cell
cell <- a5_lonlat_to_cell(-3.19, 55.95, resolution = 10)
cell
#> <a5_cell[1]>
#> [1] 6344be8000000000

# Get the boundary polygon
a5_cell_to_boundary(cell)
#> <wk_wkb[1] with CRS=OGC:CRS84>
#> [1] <POLYGON ((-3.175718 55.93546, -3.145905 55.97569, -3.151641 56.01921, -3.219413 56.00818, -3.226037 55.96443, -3.175718 55.93546...>

# Navigate the hierarchy
a5_cell_to_parent(cell)
#> <a5_cell[1]>
#> [1] 6344be0000000000
a5_cell_to_children(cell)
#> <a5_cell[4]>
#> [1] 6344be2000000000 6344be6000000000 6344bea000000000 6344bee000000000
```

``` r
# Create a collection of cells whose centres fall within a great-circle distance of 100km from the origin cell
cells <- a5_spherical_cap(cell, radius = 100000) |> 
  a5_uncompact(resolution = 10)
plot(a5_cell_to_boundary(cells), col = "#206ead20", border = "#206ead", asp = 1)
```

<img src="man/figures/README-grid-plot-1.png" alt="A5 grid plot showing a collection of cells around a point" width="100%" />

See `vignette("a5R")` for a full walkthrough of indexing, boundaries,
hierarchy, traversal, and geometry-to-cell conversion.

## Features

- **Vectorised Rust core** — all operations implemented in Rust and
  called from R via extendr.
  [Benchmarks](https://github.com/belian-earth/a5R/blob/main/benchmarks/RESULTS.md)
  confirm identical results to the Python, JavaScript, and DuckDB A5
  implementations.
- **vctrs + wk integration** — cell indices use an `a5_cell` type with
  tibble support; geometries return as `wk_wkb`/`wk_wkt` vectors
  compatible with sf and terra.
- **Geometry indexing** — `a5_polygon_to_cells()` and
  `a5_linestring_to_cells()` map polygons and great-circle lines to the
  cells that cover them.
- **Traversal** — `a5_grid_disk()` and `a5_spherical_cap()` select
  neighbours by hop count or great-circle distance.
- **Multi-threading** — opt-in parallel processing via rayon for
  vectorised operations. See `vignette("multithreading")`.

## Acknowledgements

A5 was created by [Felix Palmer](https://github.com/felixpalmer). This
package is a thin R wrapper around his work and would not exist without
it. The [Query-farm](https://github.com/Query-farm/a5) team maintain the
DuckDB A5 extension, which wraps the same Rust crate and provided a
valuable reference for this project.

# make a hex logo
library(devtools)
library(gdalraster)
library(ggplot2)
library(magick)
library(shadowtext)
load_all()

# ----- make raster ------

esri_sat <- "<GDAL_WMS><Service name=\"TMS\"><ServerUrl>http://services.arcgisonline.com/ArcGIS/rest/services/World_Imagery/MapServer/tile/${z}/${y}/${x}</ServerUrl></Service><DataWindow><UpperLeftX>-20037508.34</UpperLeftX><UpperLeftY>20037508.34</UpperLeftY><LowerRightX>20037508.34</LowerRightX><LowerRightY>-20037508.34</LowerRightY><TileLevel>17</TileLevel><TileCountX>1</TileCountX><TileCountY>1</TileCountY><YOrigin>top</YOrigin></DataWindow><Projection>EPSG:900913</Projection><BlockSizeX>256</BlockSizeX><BlockSizeY>256</BlockSizeY><BandsCount>3</BandsCount><MaxConnections>10</MaxConnections><Cache /></GDAL_WMS>"

ds <- new(GDALRaster, esri_sat)
tras <- fs::file_temp(ext = "tif")
wds <- gdalraster::warp(
  ds,
  dst_filename = tras,
  t_srs = "EPSG:4326",
  cl_arg = c("-tr", "0.1", "0.1")
)

vrtility::plot_raster_src(tras, c(1, 2, 3))
# ----------------------------------------

# ------ make a5 grid ------

resolution <- 3

origin <- a5_lonlat_to_cell(0, 0, resolution = resolution)

disk <- a5_spherical_cap(origin, radius = 1e7) |>
  a5_uncompact(resolution = resolution)


ts <- fs::file_temp(ext = "fgb")
a5sf <- tibble::tibble(
  cell_id = as.character(disk),
  geometry = a5R::a5_cell_to_boundary(disk)
) |>
  sf::st_as_sf(crs = 4326) |>
  sf::st_write(ts, delete_dsn = TRUE)


# ----- zonal stats -----
ao <- fs::file_temp(ext = "parquet")
args <- list(
  input = tras,
  zones = ts,
  output = ao,
  format = "PARQUET",
  stat = "mean",
  overwrite = TRUE,
  "include-field" = "cell_id"
)

alg <- gdalraster::gdal_run("raster zonal-stats", args)
alg$release()

aodf <- arrow::read_parquet(ao)

# ----- make plot -----
sf_wc <- sf::st_as_sf(dplyr::left_join(aodf, a5sf, by = "cell_id")) |>
  tidyr::drop_na() |>
  dplyr::mutate(
    rgbhex = {
      hsv_mat <- rgb2hsv(
        mean_band_1,
        mean_band_2,
        mean_band_3,
        maxColorValue = 255
      )
      hsv(hsv_mat[1, ], pmin(hsv_mat[2, ] * 0.45, 1), hsv_mat[3, ])
    }
  ) |>
  sf::st_transform(crs = "+proj=ortho")


ggp <- ggplot(sf_wc) +
  geom_sf(aes(fill = rgbhex), colour = "#74ac9071") +
  scale_fill_identity() +
  annotate(
    "text",
    x = 0,
    y = 0,
    label = "a5R",
    size = 20,
    colour = "#ffffffff",
    fontface = "bold",
    family = "Manjari Thin"
  ) +
  theme_void() +
  theme(
    legend.position = "none",
    plot.margin = margin(40, 40, 40, 40)
  )
ggp
tf <- fs::file_temp(ext = "png")


ggsave(tf, width = 100, height = 100, units = "mm", dpi = 300)

t2 <- fs::file_temp(ext = "png")
cropcircles::crop_hex(
  tf,
  to = t2,
  border_size = 10,
  border_colour = "#74ac90ff",
  bg_fill = "#000000ff"
)

usethis::use_logo(t2)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup--------------------------------------------------------------------
library(a5R)

## -----------------------------------------------------------------------------
origin <- a5_lonlat_to_cell(35, 55, resolution = 14)

disk <- a5_grid_disk(origin, k = 20, vertex = FALSE)
disk_vertex <- a5_grid_disk(origin, k = 20, vertex = TRUE)
cap_compact <- a5_spherical_cap(origin, radius = 5000)
cap <- a5_uncompact(cap_compact, resolution = 14)

## ----heatmap, fig.width = 10, fig.height = 10---------------------------------
dcap <- a5_cell_distance(origin, cap, units = "km")
dcap_compact <- a5_cell_distance(origin, cap_compact, units = "km")
ddisk <- a5_cell_distance(origin, disk, units = "km")
ddisk_vertex <- a5_cell_distance(origin, disk_vertex, units = "km")

pal <- hcl.colors(256, "Inferno")

oldpar <- par(mfrow = c(2, 2), mar = c(2, 2, 2, 1))
for (info in list(
  list(s = disk, d = ddisk, lab = "Grid disk (edges)"),
  list(s = disk_vertex, d = ddisk_vertex, lab = "Grid disk (vertices)"),
  list(s = cap_compact, d = dcap_compact, lab = "Spherical cap (compact)"),
  list(s = cap, d = dcap, lab = "Spherical cap")
)) {
  brk <- seq(0, max(as.numeric(info$d)), length.out = 257)
  cols <- pal[findInterval(as.numeric(info$d), brk, all.inside = TRUE)]

  plot(a5_cell_to_boundary(info$s), col = cols, border = NA, asp = 1,
       main = info$lab)
  plot(a5_cell_to_boundary(origin), border = "#ffffffff", add = TRUE)
}

## ----include = FALSE----------------------------------------------------------
par(oldpar)

## ----methods, fig.width = 10, fig.height = 10---------------------------------
wide_origin <- a5_lonlat_to_cell(35, 55, resolution = 8)
wide_cap <- a5_uncompact(
  a5_spherical_cap(wide_origin, radius = 500000),
  resolution = 8
)

hav <- a5_cell_distance(wide_origin, wide_cap, units = "m", method = "haversine")
geo <- a5_cell_distance(wide_origin, wide_cap, units = "m", method = "geodesic")

diff_m <- as.numeric(geo - hav)
brk <- seq(min(diff_m), max(diff_m), length.out = 257)
cols <- pal[findInterval(diff_m, brk, all.inside = TRUE)]

oldpar <- par(no.readonly = TRUE)
layout(matrix(c(1, 2), nrow = 1), widths = c(4, 1))
par(mar = c(2, 2, 2, 1))
plot(a5_cell_to_boundary(wide_cap), col = cols, border = NA, asp = 1,
     main = "Geodesic \u2212 Haversine (m)")

par(mar = c(9, 0, 9, 9))
image(1, seq(min(diff_m), max(diff_m), length.out = 256),
      t(seq_along(pal)), col = pal, axes = FALSE, xlab = "", ylab = "")
axis(4, las = 1)

## ----include = FALSE----------------------------------------------------------
par(oldpar)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----naive, eval = requireNamespace("arrow", quietly = TRUE), message = FALSE----
library(arrow)
library(tibble)
library(a5R)

# A real A5 cell: Edinburgh at resolution 20
cell <- a5_lonlat_to_cell(-3.19, 55.95, resolution = 20)
a5_u64_to_hex(cell)

# Write to Parquet as uint64 (the standard interchange format)
tf <- tempfile(fileext = ".parquet")
arrow::write_parquet(
  arrow::arrow_table(cell_id = a5_cell_to_arrow(cell)),
  tf
)

# Read it back naively: arrow silently converts uint64 to double
(naive <- tibble(arrow::read_parquet(tf)))

cell_as_dbl <- naive$cell_id

# The double can't distinguish this cell from nearby IDs
cell_as_dbl == cell_as_dbl + 1   # TRUE: silent corruption
cell_as_dbl == cell_as_dbl + 100 # still TRUE

## ----bridge, eval = requireNamespace("arrow", quietly = TRUE)-----------------
library(a5R)
library(tibble)

# Six cities across the globe; some will have bit 63 set (origin >= 6)
cities <- tibble(
  name = c("Edinburgh", "Tokyo", "São Paulo", "Nairobi", "Anchorage", "Sydney"),
  lon  = c(   -3.19,     139.69,     -46.63,     36.82,    -149.90,    151.21),
  lat  = c(   55.95,      35.69,     -23.55,     -1.29,      61.22,    -33.87)
)

cities$cell <- a5_lonlat_to_cell(cities$lon, cities$lat, resolution = 10)
cities

## ----enrich, eval = requireNamespace("arrow", quietly = TRUE)-----------------
edinburgh <- cities$cell[1]

cities$resolution <- a5_get_resolution(cities$cell)
cities$dist_from_edinburgh_km <- as.numeric(
  a5_cell_distance(cities$cell, rep(edinburgh, nrow(cities)), units = "km")
)

cities

## ----parquet_write, eval = requireNamespace("arrow", quietly = TRUE)----------
tf <- tempfile(fileext = ".parquet")

arrow_tbl <- arrow::arrow_table(
  name = cities$name,
  cell_id = a5_cell_to_arrow(cities$cell),
  cell_res = cities$resolution,
  dist_from_edinburgh_km = cities$dist_from_edinburgh_km
)
arrow_tbl$schema
arrow::write_parquet(arrow_tbl, tf)

## ----parquet_read, eval = requireNamespace("arrow", quietly = TRUE)-----------
pq <- arrow::read_parquet(tf, as_data_frame = FALSE)

# Recover cells from the uint64 column, bind with the rest of the data
recovered_cells <- a5_cell_from_arrow(pq$column(1))
result <- as.data.frame(pq)
result$cell <- recovered_cells
result <- tibble::as_tibble(result[c("name", "cell", "cell_res", "dist_from_edinburgh_km")])
result

## ----verify, eval = requireNamespace("arrow", quietly = TRUE)-----------------
identical(format(cities$cell), format(result$cell))

#' `r lifecycle::badge('deprecated')`
#'
#' Generate a grid of A5 cells covering an area
#'
#' Returns all cells at the target resolution that intersect the given
#' geometry. Uses hierarchical flood-fill: starting from the 12 resolution-0
#' root cells, the algorithm repeatedly expands and prunes by spatial
#' intersection until the target resolution is reached.
#'
#' @param x An area specification. One of:
#'   - A numeric vector of length 4 (`c(xmin, ymin, xmax, ymax)`) interpreted
#'     as a WGS 84 bounding box.
#'   - Any geometry handleable by [wk::wk_handle()] (e.g. [wk::wkt()],
#'     [wk::wkb()], `sfc`, `sf`, [a5_cell]).
#' @param resolution Integer scalar target resolution (0--30).
#'
#' @returns An [a5_cell] vector of cells at `resolution` that intersect `x`.
#'
#' @details
#' Grid generation runs entirely in Rust via hierarchical flood-fill with
#' bounding-box pruning. For non-bbox geometry inputs, an exact intersection
#' filter removes cells that fall outside the target shape. No cell count
#' limit is imposed — high resolutions over large areas can consume
#' significant memory.
#'
#' Input geometries must use WGS 84 coordinates; projected geometries are
#' not reprojected and will produce incorrect results. Multiple geometries
#' are collected into a GEOMETRYCOLLECTION automatically.
#' Antimeridian-crossing bounding boxes are supported: when `xmin > xmax`
#' (e.g. `c(170, -50, -170, -30)`), the bbox is split at the antimeridian.
#'
#' **Limitation:** spatial filtering uses planar geometry on lon/lat
#' coordinates, which can produce incomplete results very close to the poles
#' (above ~88° latitude) or the antimeridian. Use a larger target geometry
#' to ensure complete coverage in these areas.
#'
#' @seealso [a5_cell_to_boundary()] to convert result cells to geometries.
#' @export
#' @examples
#' # Grid from a bounding box
#' cells <- a5_grid(c(-3.3, 55.9, -3.1, 56.0), resolution = 5)
#' cells
#'
#' # Grid from a WKT polygon
#' poly <- wk::wkt("POLYGON ((-3.3 55.9, -3.1 55.9, -3.1 56, -3.3 56, -3.3 55.9))")
#' cells <- a5_grid(poly, resolution = 5)
a5_grid <- function(x, resolution) {
  lifecycle::deprecate_warn(
    when = "0.4.0",
    what = "a5_grid()",
    with = "a5_polygon_to_cells()",
    details = c(
      i = paste0(
        "`a5_polygon_to_cells()` uses centre-in-polygon containment, ",
        "whereas `a5_grid()` uses boundary intersection. ",
        "Results can differ for cells crossed by the polygon edge."
      ),
      i = paste0(
        "For a bounding box, build a closed polygon first, e.g. ",
        "`a5_polygon_to_cells(wk::rct(xmin, ymin, xmax, ymax), res)`."
      )
    )
  )
  resolution <- vctrs::vec_cast(resolution, integer())
  check_resolution(resolution)
  vctrs::vec_assert(resolution, size = 1L)

  is_bbox <- is.numeric(x)

  if (is_bbox) {
    validate_bbox(x)
  }

  # Get bbox for Rust grid generation
  if (is_bbox && x[[1]] > x[[3]]) {
    # Antimeridian-crossing: two halves
    cells1 <- cells_from_rs(a5_grid_bbox_rs(
      x[[1]],
      x[[2]],
      180,
      x[[4]],
      resolution
    ))
    cells2 <- cells_from_rs(a5_grid_bbox_rs(
      -180,
      x[[2]],
      x[[3]],
      x[[4]],
      resolution
    ))
    cells <- vctrs::vec_c(cells1, cells2)
  } else {
    if (is_bbox) {
      bb <- list(xmin = x[[1]], ymin = x[[2]], xmax = x[[3]], ymax = x[[4]])
    } else {
      bb <- unclass(wk::wk_bbox(x))
    }
    cells <- cells_from_rs(
      a5_grid_bbox_rs(bb$xmin, bb$ymin, bb$xmax, bb$ymax, resolution)
    )
  }

  # Final exact filter for non-bbox inputs
  if (!is_bbox) {
    target_wkt <- as_target_wkt(x)
    filtered <- a5_grid_intersects_rs(cell_data(cells), target_wkt)
    cells <- cells_from_rs(filtered)
  }

  cells
}

# -- internal helpers ----------------------------------------------------------

#' Validate a numeric bounding box
#' @noRd
validate_bbox <- function(x, call = rlang::caller_env()) {
  if (length(x) != 4L) {
    cli::cli_abort(
      "Numeric {.arg x} must have length 4 ({.code c(xmin, ymin, xmax, ymax)}), not {length(x)}.",
      call = call
    )
  }
  if (anyNA(x)) {
    cli::cli_abort(
      "{.arg x} must not contain {.code NA} values.",
      call = call
    )
  }
  if (x[[2]] >= x[[4]]) {
    cli::cli_abort(
      "{.code ymin} ({x[[2]]}) must be less than {.code ymax} ({x[[4]]}).",
      call = call
    )
  }
  if (x[[1]] == x[[3]]) {
    cli::cli_abort(
      "{.code xmin} and {.code xmax} must not be equal ({x[[1]]}).",
      call = call
    )
  }
  invisible(x)
}

#' Convert geometry input to a single WKT string for Rust
#' @noRd
as_target_wkt <- function(x) {
  wkt_vec <- as.character(wk::as_wkt(x))
  if (length(wkt_vec) == 1L) {
    wkt_vec
  } else {
    paste0("GEOMETRYCOLLECTION (", paste(wkt_vec, collapse = ", "), ")")
  }
}

# nocov start
.onLoad <- function(libname, pkgname) {
  n <- getOption("a5R.threads")
  if (is.null(n)) {
    env <- Sys.getenv("A5R_NUM_THREADS", unset = "")
    if (nzchar(env)) {
      n <- as.integer(env)
    }
  }
  if (!is.null(n) && !is.na(n) && n >= 1L) {
    a5_set_threads_rs(as.integer(n))
  }
}
# nocov end

#' Get cell boundary polygons
#'
#' Returns the boundary of each cell as a [wk::wkt()] or [wk::wkb()]
#' polygon geometry. Boundaries are pentagonal polygons on the WGS 84
#' ellipsoid.
#'
#' @param cell An [a5_cell] vector.
#' @param format Character scalar, either `"wkb"` (default) or `"wkt"`.
#' @param closed Logical scalar; if `TRUE` (default) the ring is closed
#'   (first vertex repeated at end).
#' @param segments Integer scalar or `NULL`. Number of interpolation segments
#'   per edge for geodesic accuracy. `NULL` uses the default (straight edges).
#' @returns A `wk_wkt` or `wk_wkb` vector of polygon geometries with
#'   `wk::wk_crs_longlat()` CRS.
#'
#' @seealso [a5_cell_to_lonlat()] for cell centroids.
#' @export
#' @examples
#' cell <- a5_lonlat_to_cell(-3.19, 55.95, resolution = 5)
#' a5_cell_to_boundary(cell)
#' a5_cell_to_boundary(cell, format = "wkt")
a5_cell_to_boundary <- function(
  cell,
  format = c("wkb", "wkt"),
  closed = TRUE,
  segments = NULL
) {
  cell <- as_a5_cell(cell)
  format <- rlang::arg_match(format)
  closed <- vctrs::vec_cast(closed, logical())
  vctrs::vec_assert(closed, size = 1L)
  if (!is.null(segments)) {
    segments <- vctrs::vec_cast(segments, integer())
    vctrs::vec_assert(segments, size = 1L)
  }
  cd <- cell_data(cell)
  if (format == "wkb") {
    raw_list <- a5_cell_to_boundary_wkb_rs(cd, closed, segments)
    wk::new_wk_wkb(raw_list, crs = wk::wk_crs_longlat())
  } else {
    wkt <- a5_cell_to_boundary_rs(cd, closed, segments)
    wk::new_wk_wkt(wkt, crs = wk::wk_crs_longlat())
  }
}

test_that("cell_to_boundary returns wk_wkb by default", {
  cell <- a5_lonlat_to_cell(-3.19, 55.95, resolution = 5)
  boundary <- a5_cell_to_boundary(cell)
  expect_s3_class(boundary, "wk_wkb")
  expect_length(boundary, 1L)
})

test_that("boundary has lonlat CRS", {
  cell <- a5_lonlat_to_cell(0, 0, resolution = 5)
  boundary <- a5_cell_to_boundary(cell)
  crs <- wk::wk_crs(boundary)
  expect_identical(crs, wk::wk_crs_longlat())
})

test_that("boundary with closed = FALSE omits repeated vertex", {
  cell <- a5_lonlat_to_cell(0, 0, resolution = 5)
  open <- a5_cell_to_boundary(cell, format = "wkt", closed = FALSE)
  closed <- a5_cell_to_boundary(cell, format = "wkt", closed = TRUE)
  # closed ring repeats the first vertex, so its WKT is longer
  expect_true(nchar(as.character(closed)) > nchar(as.character(open)))
})

test_that("boundary with segments param works", {
  cell <- a5_lonlat_to_cell(0, 0, resolution = 5)
  b_default <- a5_cell_to_boundary(cell, format = "wkt")
  b_seg <- a5_cell_to_boundary(cell, format = "wkt", segments = 3L)
  # more segments = more vertices = longer WKT
  expect_true(nchar(as.character(b_seg)) > nchar(as.character(b_default)))
})

test_that("boundary format = wkt returns wk_wkt", {
  cell <- a5_lonlat_to_cell(-3.19, 55.95, resolution = 5)
  boundary <- a5_cell_to_boundary(cell, format = "wkt")
  expect_s3_class(boundary, "wk_wkt")
  expect_length(boundary, 1L)
  expect_true(grepl("^POLYGON", as.character(boundary)))
})

test_that("boundary format rejects invalid values", {
  cell <- a5_lonlat_to_cell(0, 0, resolution = 5)
  expect_error(a5_cell_to_boundary(cell, format = "geojson"), "format")
})

test_that("boundary is vectorised", {
  cells <- a5_lonlat_to_cell(c(0, 10, -10), c(0, 20, -20), resolution = 5)
  boundaries <- a5_cell_to_boundary(cells)
  expect_length(boundaries, 3L)
  expect_s3_class(boundaries, "wk_wkb")
})

test_that("boundary handles NA cell", {
  cells <- a5_cell(c("0800000000000006", NA))
  boundaries <- a5_cell_to_boundary(cells, format = "wkt")
  expect_length(boundaries, 2L)
  expect_true(is.na(as.character(boundaries)[2]))
})

test_that("cell_area decreases with resolution", {
  areas <- a5_cell_area(0:5)
  expect_true(all(diff(areas) < units::as_units(0, "m^2")))
})


test_that("cell_area converts units", {
  m2 <- a5_cell_area(5)
  km2 <- a5_cell_area(5, units = "km^2")
  expect_equal(as.numeric(m2) / 1e6, as.numeric(km2), tolerance = 1e-10)
})

test_that("cell_area rejects non-area units", {
  expect_error(a5_cell_area(5, units = "kg"), "area unit")
  expect_error(a5_cell_area(5, units = "nonsense"), "area unit")
})

#' Cells whose centres lie inside a polygon
#'
#' Returns A5 cells at `resolution` whose centres fall inside the polygon.
#' Multi-feature inputs (a `MULTIPOLYGON`, an `sfc` of multiple polygons,
#' or a `POLYGON` with holes) are handled natively: per polygon part, the
#' outer-ring cells are computed and any hole-ring cells are subtracted,
#' then the results are unioned across parts. The final cell set is
#' compacted; use [a5_uncompact()] to expand to a uniform-resolution grid.
#'
#' @param x A polygon-like geometry. One of:
#'   - Any geometry handleable by [wk::wk_handle()] (e.g. [wk::wkt()],
#'     [wk::wkb()], [wk::rct()], `sf`, `sfc`) containing one or more
#'     `POLYGON` / `MULTIPOLYGON` features.
#'   - A `SpatVector` of polygons (requires the `terra` package).
#'   - A two-column numeric matrix (`cbind(lon, lat)`) of vertices,
#'     interpreted as a single outer ring.
#'   - A `data.frame` with columns `lon` and `lat`, interpreted as a
#'     single outer ring.
#' @param resolution Integer scalar target resolution (0-30).
#'
#' @returns An [a5_cell] vector at or coarser than `resolution`.
#'
#' @details
#' Membership is determined by **centre-point containment**: a cell is
#' included iff its centroid lies inside the polygon, with hole rings
#' properly subtracted. This is distinct from [a5_grid()]'s
#' boundary-intersection semantics; for the same polygon the two
#' functions can return slightly different cell sets near the boundary.
#'
#' Coordinates must be WGS 84 longitude/latitude in degrees. Rings are
#' closed automatically; a trailing duplicate vertex is dropped if
#' present.
#'
#' Where no A5 cell centroids at the specified `resolution` fall within
#' the geometry, an empty `a5_cell` vector is returned.
#'
#' Matrix and `data.frame` inputs are treated as a single ring; for
#' multi-feature data or polygons with holes, pass an `sf`, `sfc`, wk,
#' or `SpatVector` geometry instead.
#'
#' @seealso [a5_linestring_to_cells()], [a5_uncompact()].
#' @export
#' @examples
#' poly <- wk::wkt(
#'   "POLYGON ((-3.3 55.9, -3.1 55.9, -3.1 56, -3.3 56, -3.3 55.9))"
#' )
#' cells <- a5_polygon_to_cells(poly, resolution = 8)
#' length(cells)
a5_polygon_to_cells <- function(x, resolution) {
  resolution <- vctrs::vec_cast(resolution, integer())
  check_resolution(resolution)
  vctrs::vec_assert(resolution, size = 1L)

  bundle <- prepare_polygon_input(x)

  cells_from_rs(a5_polygon_to_cells_rs(
    bundle$lon,
    bundle$lat,
    bundle$offsets,
    bundle$part_id,
    bundle$is_outer,
    resolution
  ))
}

#' Cells traced by a great-circle linestring
#'
#' Returns the A5 cells at `resolution` whose pentagons are intersected by
#' the great-circle polyline connecting the supplied waypoints. Output is
#' uncompacted, in discovery order along the path, with duplicates removed
#' first-seen. Multi-feature inputs (a `MULTILINESTRING` or an `sfc` of
#' multiple linestrings) are handled natively: per-feature cell sequences
#' are concatenated in feature order with first-seen deduplication.
#'
#' @param x A linestring-like geometry. One of:
#'   - Any geometry handleable by [wk::wk_handle()] containing one or
#'     more `LINESTRING` / `MULTILINESTRING` features.
#'   - A `SpatVector` of linestrings (requires the `terra` package).
#'   - A two-column numeric matrix (`cbind(lon, lat)`) of waypoints.
#'   - A `data.frame` with columns `lon` and `lat`.
#' @param resolution Integer scalar target resolution (0--30).
#'
#' @returns An [a5_cell] vector at `resolution`.
#'
#' @details
#' Consecutive waypoints are connected by great-circle arcs (not rhumb
#' lines or planar segments), so antimeridian-crossing paths work
#' transparently when written in unwrapped lon/lat.
#'
#' Matrix and `data.frame` inputs are treated as a single linestring;
#' for multi-feature data, pass an `sf`, `sfc`, wk, or `SpatVector`
#' geometry instead.
#'
#' @seealso [a5_polygon_to_cells()], [a5_grid_disk()].
#' @export
#' @examples
#' line <- wk::wkt("LINESTRING (2.35 48.86, -0.13 51.51)")
#' cells <- a5_linestring_to_cells(line, resolution = 6)
#' length(cells)
a5_linestring_to_cells <- function(x, resolution) {
  resolution <- vctrs::vec_cast(resolution, integer())
  check_resolution(resolution)
  vctrs::vec_assert(resolution, size = 1L)

  bundle <- prepare_linestring_input(x)

  cells_from_rs(a5_linestring_to_cells_rs(
    bundle$lon,
    bundle$lat,
    bundle$offsets,
    resolution
  ))
}

# -- input preparation ---------------------------------------------------------

#' Flatten polygon input to per-ring (lon, lat, offsets, part_id, is_outer)
#'
#' Returns `list(lon, lat, offsets, part_id, is_outer)`. `offsets` is
#' cumulative; ring `i` occupies `offsets[i] + 1` to `offsets[i + 1]`.
#' `part_id[i]` groups rings by polygon part. `is_outer[i]` is `1L` for
#' an outer ring and `0L` for a hole ring.
#' @noRd
prepare_polygon_input <- function(x, call = rlang::caller_env()) {
  x <- maybe_coerce_spatvector(x, call = call)
  bundle <- if (is.matrix(x)) {
    coords_from_matrix(x, call = call)
  } else if (is.data.frame(x) && !inherits(x, "sf")) {
    coords_from_data_frame(x, call = call)
  } else {
    coords_from_wk_polygons(x, call = call)
  }

  if (length(bundle$offsets) < 2L) {
    cli::cli_abort("{.arg x} contains no polygon rings.", call = call)
  }
  validate_feature_sizes(bundle, expected = "polygon", call = call)
  if (anyNA(bundle$lon) || anyNA(bundle$lat)) {
    cli::cli_abort(
      "{.arg x} must not contain {.code NA} coordinates.",
      call = call
    )
  }
  bundle
}

#' Flatten linestring input to (lon, lat, offsets)
#' @noRd
prepare_linestring_input <- function(x, call = rlang::caller_env()) {
  x <- maybe_coerce_spatvector(x, call = call)
  bundle <- if (is.matrix(x)) {
    coords_from_matrix(x, call = call)
  } else if (is.data.frame(x) && !inherits(x, "sf")) {
    coords_from_data_frame(x, call = call)
  } else {
    coords_from_wk_linestrings(x, call = call)
  }

  if (length(bundle$offsets) < 2L) {
    cli::cli_abort("{.arg x} contains no linestrings.", call = call)
  }
  validate_feature_sizes(bundle, expected = "linestring", call = call)
  if (anyNA(bundle$lon) || anyNA(bundle$lat)) {
    cli::cli_abort(
      "{.arg x} must not contain {.code NA} coordinates.",
      call = call
    )
  }
  bundle
}

#' @noRd
coords_from_matrix <- function(x, call = rlang::caller_env()) {
  if (ncol(x) != 2L) {
    cli::cli_abort(
      "Matrix input must have exactly 2 columns ({.code lon}, {.code lat}); \\
       got {ncol(x)}.",
      call = call
    )
  }
  if (!is.numeric(x)) {
    cli::cli_abort("Matrix input must be numeric.", call = call)
  }
  lon <- as.numeric(x[, 1L])
  lat <- as.numeric(x[, 2L])
  list(
    lon = lon,
    lat = lat,
    offsets = c(0L, length(lon)),
    part_id = 1L,
    is_outer = 1L
  )
}

#' @noRd
coords_from_data_frame <- function(x, call = rlang::caller_env()) {
  if (!all(c("lon", "lat") %in% names(x))) {
    cli::cli_abort(
      "Data frame input must have columns {.code lon} and {.code lat}.",
      call = call
    )
  }
  lon <- as.numeric(x$lon)
  lat <- as.numeric(x$lat)
  list(
    lon = lon,
    lat = lat,
    offsets = c(0L, length(lon)),
    part_id = 1L,
    is_outer = 1L
  )
}

#' @noRd
coords_from_wk_polygons <- function(x, call = rlang::caller_env()) {
  meta <- safe_wk_meta(x, call = call)

  # wk_meta() geometry_type codes: 3 = POLYGON, 6 = MULTIPOLYGON.
  allowed <- c(POLYGON = 3L, MULTIPOLYGON = 6L)
  bad <- !is.na(meta$geometry_type) & !(meta$geometry_type %in% allowed)
  if (any(bad)) {
    cli::cli_abort(
      "{.arg x} must contain only {paste(names(allowed), collapse = ' or ')} geometries.",
      call = call
    )
  }
  if (any(meta$is_empty)) {
    cli::cli_abort("{.arg x} contains empty geometries.", call = call)
  }

  coords <- wk::wk_coords(x)
  if (nrow(coords) == 0L) {
    return(list(
      lon = numeric(),
      lat = numeric(),
      offsets = 0L,
      part_id = integer(),
      is_outer = integer()
    ))
  }

  # Ring grouping (feature_id, part_id, ring_id) — first-seen order.
  ring_key <- paste(
    coords$feature_id,
    coords$part_id,
    coords$ring_id,
    sep = "/"
  )
  ring_group <- match(ring_key, unique(ring_key))

  # Polygon-part grouping (feature_id, part_id) — first-seen order.
  part_key <- paste(coords$feature_id, coords$part_id, sep = "/")
  row_part_id <- match(part_key, unique(part_key))

  rle_rings <- rle(ring_group)
  offsets <- c(0L, cumsum(rle_rings$lengths))
  ring_starts <- offsets[-length(offsets)] + 1L

  ring_part_id <- row_part_id[ring_starts]
  # Within a polygon part, the first ring (smallest ring_id) is the outer
  # and any later rings are holes. wk emits rings in outer-first order, so
  # the first occurrence of each part_id marks the outer ring.
  ring_is_outer <- as.integer(!duplicated(ring_part_id))

  lon <- coords$x
  lat <- coords$y

  # Drop trailing duplicate vertex per ring (WKT/WKB rings are closed).
  keep <- rep(TRUE, length(lon))
  for (i in seq_along(ring_starts)) {
    a <- offsets[i] + 1L
    b <- offsets[i + 1L]
    if (
      b - a >= 1L &&
        isTRUE(lon[a] == lon[b]) &&
        isTRUE(lat[a] == lat[b])
    ) {
      keep[b] <- FALSE
    }
  }
  if (!all(keep)) {
    lon <- lon[keep]
    lat <- lat[keep]
    kept_groups <- ring_group[keep]
    rle_rings <- rle(kept_groups)
    offsets <- c(0L, cumsum(rle_rings$lengths))
  }

  list(
    lon = as.numeric(lon),
    lat = as.numeric(lat),
    offsets = as.integer(offsets),
    part_id = as.integer(ring_part_id),
    is_outer = as.integer(ring_is_outer)
  )
}

#' @noRd
coords_from_wk_linestrings <- function(x, call = rlang::caller_env()) {
  meta <- safe_wk_meta(x, call = call)

  # wk_meta() geometry_type codes: 2 = LINESTRING, 5 = MULTILINESTRING.
  allowed <- c(LINESTRING = 2L, MULTILINESTRING = 5L)
  bad <- !is.na(meta$geometry_type) & !(meta$geometry_type %in% allowed)
  if (any(bad)) {
    cli::cli_abort(
      "{.arg x} must contain only {paste(names(allowed), collapse = ' or ')} geometries.",
      call = call
    )
  }
  if (any(meta$is_empty)) {
    cli::cli_abort("{.arg x} contains empty geometries.", call = call)
  }

  coords <- wk::wk_coords(x)
  if (nrow(coords) == 0L) {
    return(list(lon = numeric(), lat = numeric(), offsets = 0L))
  }

  # Linestrings: one ring_id per linestring (within a feature/part).
  line_key <- paste(
    coords$feature_id,
    coords$part_id,
    coords$ring_id,
    sep = "/"
  )
  line_group <- match(line_key, unique(line_key))
  rle_lines <- rle(line_group)
  offsets <- c(0L, cumsum(rle_lines$lengths))

  list(
    lon = as.numeric(coords$x),
    lat = as.numeric(coords$y),
    offsets = as.integer(offsets)
  )
}

#' Coerce a terra `SpatVector` to a wk-handleable WKT vector
#'
#' terra does not register `wk_handle()` methods, so we route SpatVector
#' inputs through `terra::geom(x, wkt = TRUE)` and wrap the result in a
#' `wk_wkt` vector. `terra` is in `Suggests`; this path errors with a
#' clear install hint if the package is missing.
#' @noRd
maybe_coerce_spatvector <- function(x, call = rlang::caller_env()) {
  if (!inherits(x, "SpatVector")) {
    return(x)
  }
  rlang::check_installed("terra", reason = "to convert terra SpatVector input.")
  wk::wkt(terra::geom(x, wkt = TRUE))
}

#' @noRd
safe_wk_meta <- function(x, call = rlang::caller_env()) {
  tryCatch(
    wk::wk_meta(x),
    error = function(e) {
      cli::cli_abort(
        c(
          "Could not interpret {.arg x} as a geometry.",
          i = "Pass a matrix, data frame with {.code lon}/{.code lat} \\
               columns, or a {.fn wk::wk_handle}-able object."
        ),
        parent = e,
        call = call
      )
    }
  )
}

#' @noRd
validate_feature_sizes <- function(
  bundle,
  expected,
  call = rlang::caller_env()
) {
  sizes <- diff(bundle$offsets)
  min_size <- switch(expected, polygon = 3L, linestring = 2L)
  short <- which(sizes < min_size)
  if (length(short) > 0L) {
    cli::cli_abort(
      "Each {expected} ring must have at least {min_size} vertices; \\
       ring {short[1]} has {sizes[short[1]]}.",
      call = call
    )
  }
  invisible(bundle)
}

# -- helpers -------------------------------------------------------------------

square_wkt <- function(xmin, ymin, xmax, ymax) {
  sprintf(
    "POLYGON ((%s %s, %s %s, %s %s, %s %s, %s %s))",
    xmin, ymin, xmax, ymin, xmax, ymax, xmin, ymax, xmin, ymin
  )
}

# -- a5_polygon_to_cells: return type ------------------------------------------

test_that("a5_polygon_to_cells returns an a5_cell vector", {
  poly <- wk::wkt(square_wkt(-3.5, 55.5, -2.5, 56.5))
  cells <- a5_polygon_to_cells(poly, resolution = 10)
  expect_s3_class(cells, "a5_cell")
  expect_true(length(cells) > 0)
  expect_true(all(a5_get_resolution(cells) <= 10L))
  expect_equal(length(cells), length(unique(format(cells))))
})

test_that("a5_polygon_to_cells uses centre-point containment (centre cell is inside)", {
  poly <- wk::wkt(square_wkt(-3.5, 55.5, -2.5, 56.5))
  cells <- a5_polygon_to_cells(poly, resolution = 10)
  centre <- a5_lonlat_to_cell(-3.0, 56.0, resolution = 10)
  expanded <- a5_uncompact(cells, resolution = 10)
  expect_true(format(centre) %in% format(expanded))
})

# -- input shapes --------------------------------------------------------------

test_that("a5_polygon_to_cells accepts a matrix shortcut", {
  m <- cbind(c(-3.5, -2.5, -2.5, -3.5), c(55.5, 55.5, 56.5, 56.5))
  cells_m <- a5_polygon_to_cells(m, resolution = 10)
  cells_w <- a5_polygon_to_cells(
    wk::wkt(square_wkt(-3.5, 55.5, -2.5, 56.5)),
    resolution = 10
  )
  expect_setequal(format(cells_m), format(cells_w))
})

test_that("a5_polygon_to_cells accepts a data.frame shortcut", {
  df <- data.frame(
    lon = c(-3.5, -2.5, -2.5, -3.5),
    lat = c(55.5, 55.5, 56.5, 56.5)
  )
  cells_df <- a5_polygon_to_cells(df, resolution = 10)
  cells_w <- a5_polygon_to_cells(
    wk::wkt(square_wkt(-3.5, 55.5, -2.5, 56.5)),
    resolution = 10
  )
  expect_setequal(format(cells_df), format(cells_w))
})

test_that("a5_polygon_to_cells accepts wk::rct", {
  rect_cells <- a5_polygon_to_cells(wk::rct(-3.5, 55.5, -2.5, 56.5),
                                    resolution = 10)
  wkt_cells <- a5_polygon_to_cells(
    wk::wkt(square_wkt(-3.5, 55.5, -2.5, 56.5)),
    resolution = 10
  )
  expect_setequal(format(rect_cells), format(wkt_cells))
})

test_that("a5_polygon_to_cells accepts a terra SpatVector", {
  skip_if_not_installed("terra")
  sv <- terra::vect(square_wkt(-3.5, 55.5, -2.5, 56.5))
  cells_terra <- a5_polygon_to_cells(sv, resolution = 10)
  cells_w <- a5_polygon_to_cells(
    wk::wkt(square_wkt(-3.5, 55.5, -2.5, 56.5)),
    resolution = 10
  )
  expect_setequal(format(cells_terra), format(cells_w))
})

test_that("a5_polygon_to_cells handles a multi-polygon SpatVector", {
  skip_if_not_installed("terra")
  sv <- terra::vect(c(
    square_wkt(-3.5, 55.5, -2.5, 56.5),
    square_wkt(1, 1, 2, 2)
  ))
  cells_terra <- a5_polygon_to_cells(sv, resolution = 8)
  cells_w <- a5_polygon_to_cells(
    wk::wkt(c(
      square_wkt(-3.5, 55.5, -2.5, 56.5),
      square_wkt(1, 1, 2, 2)
    )),
    resolution = 8
  )
  expect_setequal(format(cells_terra), format(cells_w))
})

test_that("a5_linestring_to_cells accepts a terra SpatVector", {
  skip_if_not_installed("terra")
  sv <- terra::vect("LINESTRING (2.35 48.86, -0.13 51.51)")
  cells_terra <- a5_linestring_to_cells(sv, resolution = 5)
  cells_w <- a5_linestring_to_cells(
    wk::wkt("LINESTRING (2.35 48.86, -0.13 51.51)"),
    resolution = 5
  )
  expect_identical(format(cells_terra), format(cells_w))
})

test_that("a5_polygon_to_cells accepts an sfc polygon", {
  skip_if_not_installed("sf")
  sfc <- sf::st_sfc(
    sf::st_polygon(list(matrix(
      c(-3.5, 55.5, -2.5, 55.5, -2.5, 56.5, -3.5, 56.5, -3.5, 55.5),
      ncol = 2, byrow = TRUE
    ))),
    crs = 4326
  )
  cells_sf <- a5_polygon_to_cells(sfc, resolution = 10)
  cells_w <- a5_polygon_to_cells(
    wk::wkt(square_wkt(-3.5, 55.5, -2.5, 56.5)),
    resolution = 10
  )
  expect_setequal(format(cells_sf), format(cells_w))
})

# -- multi-feature: handled natively ------------------------------------------

test_that("MULTIPOLYGON returns the union of cells across parts", {
  mp <- wk::wkt(
    "MULTIPOLYGON (((-3.5 55.5, -2.5 55.5, -2.5 56.5, -3.5 56.5, -3.5 55.5)),
                   ((1 1, 2 1, 2 2, 1 2, 1 1)))"
  )
  union_cells <- a5_polygon_to_cells(mp, resolution = 8)
  part1 <- a5_polygon_to_cells(
    wk::wkt(square_wkt(-3.5, 55.5, -2.5, 56.5)),
    resolution = 8
  )
  part2 <- a5_polygon_to_cells(
    wk::wkt(square_wkt(1, 1, 2, 2)),
    resolution = 8
  )
  u1 <- a5_uncompact(part1, resolution = 8)
  u2 <- a5_uncompact(part2, resolution = 8)
  uu <- a5_uncompact(union_cells, resolution = 8)
  expect_setequal(format(uu), c(format(u1), format(u2)))
})

test_that("sfc of multiple polygons returns the union", {
  skip_if_not_installed("sf")
  sfc <- sf::st_sfc(
    sf::st_polygon(list(matrix(
      c(-3.5, 55.5, -2.5, 55.5, -2.5, 56.5, -3.5, 56.5, -3.5, 55.5),
      ncol = 2, byrow = TRUE
    ))),
    sf::st_polygon(list(matrix(
      c(1, 1, 2, 1, 2, 2, 1, 2, 1, 1),
      ncol = 2, byrow = TRUE
    ))),
    crs = 4326
  )
  cells <- a5_polygon_to_cells(sfc, resolution = 8)
  expect_s3_class(cells, "a5_cell")
  expect_true(length(cells) > 0)
})

# -- holes are properly subtracted --------------------------------------------

test_that("polygon-with-hole subtracts hole cells from the outer", {
  outer <- wk::wkt(square_wkt(0, 0, 4, 4))
  with_hole <- wk::wkt(
    "POLYGON ((0 0, 4 0, 4 4, 0 4, 0 0), (1 1, 2 1, 2 2, 1 2, 1 1))"
  )
  hole <- wk::wkt(square_wkt(1, 1, 2, 2))

  outer_u <- a5_uncompact(a5_polygon_to_cells(outer, resolution = 6),
                          resolution = 6)
  hole_u  <- a5_uncompact(a5_polygon_to_cells(hole, resolution = 6),
                          resolution = 6)
  result_u <- a5_uncompact(a5_polygon_to_cells(with_hole, resolution = 6),
                           resolution = 6)

  # Expected: outer cells minus hole cells.
  expected <- setdiff(format(outer_u), format(hole_u))
  expect_setequal(format(result_u), expected)
})

test_that("hole that lies wholly inside the outer leaves a real gap", {
  # The hole is centred inside the outer; some cells fall fully inside it.
  with_hole <- wk::wkt(
    "POLYGON ((0 0, 8 0, 8 8, 0 8, 0 0), (3 3, 5 3, 5 5, 3 5, 3 3))"
  )
  hole_alone <- wk::wkt(square_wkt(3, 3, 5, 5))

  result_u <- a5_uncompact(a5_polygon_to_cells(with_hole, resolution = 6),
                           resolution = 6)
  hole_u <- a5_uncompact(a5_polygon_to_cells(hole_alone, resolution = 6),
                         resolution = 6)
  # No cell from the hole should appear in the result.
  expect_true(length(intersect(format(result_u), format(hole_u))) == 0L)
})

test_that("polygon with two holes subtracts both", {
  # Outer 0..8 with two disjoint holes.
  outer  <- wk::wkt(square_wkt(0, 0, 8, 8))
  hole_a <- wk::wkt(square_wkt(1, 1, 3, 3))
  hole_b <- wk::wkt(square_wkt(5, 5, 7, 7))
  with_holes <- wk::wkt(
    "POLYGON ((0 0, 8 0, 8 8, 0 8, 0 0),
              (1 1, 3 1, 3 3, 1 3, 1 1),
              (5 5, 7 5, 7 7, 5 7, 5 5))"
  )

  outer_u  <- a5_uncompact(a5_polygon_to_cells(outer,  resolution = 6), 6)
  hole_a_u <- a5_uncompact(a5_polygon_to_cells(hole_a, resolution = 6), 6)
  hole_b_u <- a5_uncompact(a5_polygon_to_cells(hole_b, resolution = 6), 6)
  result_u <- a5_uncompact(a5_polygon_to_cells(with_holes, resolution = 6), 6)

  expected <- setdiff(format(outer_u),
                      c(format(hole_a_u), format(hole_b_u)))
  expect_setequal(format(result_u), expected)
})

test_that("MULTIPOLYGON where one part has a hole subtracts only that hole", {
  # First part: outer 0..8 with a hole 3..5. Second part: simple 10..12 square.
  outer1 <- wk::wkt(square_wkt(0, 0, 8, 8))
  hole1  <- wk::wkt(square_wkt(3, 3, 5, 5))
  part2  <- wk::wkt(square_wkt(10, 10, 12, 12))
  mp <- wk::wkt(
    "MULTIPOLYGON (((0 0, 8 0, 8 8, 0 8, 0 0), (3 3, 5 3, 5 5, 3 5, 3 3)),
                   ((10 10, 12 10, 12 12, 10 12, 10 10)))"
  )

  outer1_u <- a5_uncompact(a5_polygon_to_cells(outer1, resolution = 6), 6)
  hole1_u  <- a5_uncompact(a5_polygon_to_cells(hole1,  resolution = 6), 6)
  part2_u  <- a5_uncompact(a5_polygon_to_cells(part2,  resolution = 6), 6)
  result_u <- a5_uncompact(a5_polygon_to_cells(mp, resolution = 6), 6)

  expected <- c(setdiff(format(outer1_u), format(hole1_u)),
                format(part2_u))
  expect_setequal(format(result_u), expected)
})

test_that("MULTIPOLYGON with holes in every part subtracts each independently", {
  outer1 <- wk::wkt(square_wkt(0, 0, 8, 8))
  hole1  <- wk::wkt(square_wkt(3, 3, 5, 5))
  outer2 <- wk::wkt(square_wkt(10, 0, 18, 8))
  hole2  <- wk::wkt(square_wkt(13, 3, 15, 5))
  mp <- wk::wkt(
    "MULTIPOLYGON (((0 0, 8 0, 8 8, 0 8, 0 0), (3 3, 5 3, 5 5, 3 5, 3 3)),
                   ((10 0, 18 0, 18 8, 10 8, 10 0), (13 3, 15 3, 15 5, 13 5, 13 3)))"
  )

  outer1_u <- a5_uncompact(a5_polygon_to_cells(outer1, resolution = 6), 6)
  hole1_u  <- a5_uncompact(a5_polygon_to_cells(hole1,  resolution = 6), 6)
  outer2_u <- a5_uncompact(a5_polygon_to_cells(outer2, resolution = 6), 6)
  hole2_u  <- a5_uncompact(a5_polygon_to_cells(hole2,  resolution = 6), 6)
  result_u <- a5_uncompact(a5_polygon_to_cells(mp, resolution = 6), 6)

  expected <- c(setdiff(format(outer1_u), format(hole1_u)),
                setdiff(format(outer2_u), format(hole2_u)))
  expect_setequal(format(result_u), expected)
})

test_that("sf MULTIPOLYGON with a hole subtracts the hole", {
  skip_if_not_installed("sf")
  outer_ring <- matrix(c(0, 0, 8, 0, 8, 8, 0, 8, 0, 0), ncol = 2, byrow = TRUE)
  hole_ring  <- matrix(c(3, 3, 5, 3, 5, 5, 3, 5, 3, 3), ncol = 2, byrow = TRUE)
  poly <- sf::st_polygon(list(outer_ring, hole_ring))
  sfc <- sf::st_sfc(poly, crs = 4326)

  outer <- wk::wkt(square_wkt(0, 0, 8, 8))
  hole  <- wk::wkt(square_wkt(3, 3, 5, 5))
  outer_u <- a5_uncompact(a5_polygon_to_cells(outer, resolution = 6), 6)
  hole_u  <- a5_uncompact(a5_polygon_to_cells(hole,  resolution = 6), 6)
  result_u <- a5_uncompact(a5_polygon_to_cells(sfc, resolution = 6), 6)

  expected <- setdiff(format(outer_u), format(hole_u))
  expect_setequal(format(result_u), expected)
})

# -- a5_linestring_to_cells ----------------------------------------------------

test_that("a5_linestring_to_cells returns ordered cells along the path", {
  line <- wk::wkt("LINESTRING (2.35 48.86, -0.13 51.51)")
  cells <- a5_linestring_to_cells(line, resolution = 5)
  expect_s3_class(cells, "a5_cell")
  expect_true(length(cells) >= 2)
  start <- a5_lonlat_to_cell(2.35, 48.86, resolution = 5)
  end   <- a5_lonlat_to_cell(-0.13, 51.51, resolution = 5)
  expect_true(format(start) %in% format(cells))
  expect_true(format(end) %in% format(cells))
  expect_equal(length(cells), length(unique(format(cells))))
})

test_that("a5_linestring_to_cells matrix and data.frame shortcuts match wk input", {
  pts <- rbind(c(2.35, 48.86), c(-0.13, 51.51))
  c_w <- a5_linestring_to_cells(wk::wkt("LINESTRING (2.35 48.86, -0.13 51.51)"),
                                resolution = 5)
  c_m <- a5_linestring_to_cells(pts, resolution = 5)
  c_d <- a5_linestring_to_cells(data.frame(lon = pts[, 1], lat = pts[, 2]),
                                resolution = 5)
  expect_identical(format(c_m), format(c_w))
  expect_identical(format(c_d), format(c_w))
})

test_that("MULTILINESTRING returns the deduped union in feature order", {
  ml <- wk::wkt(
    "MULTILINESTRING ((2.35 48.86, -0.13 51.51),
                      (-0.13 51.51, 4.83 45.76))"
  )
  cells <- a5_linestring_to_cells(ml, resolution = 5)
  expect_s3_class(cells, "a5_cell")
  expect_equal(length(cells), length(unique(format(cells))))
  paris  <- a5_lonlat_to_cell(2.35, 48.86, resolution = 5)
  london <- a5_lonlat_to_cell(-0.13, 51.51, resolution = 5)
  lyon   <- a5_lonlat_to_cell(4.83, 45.76, resolution = 5)
  expect_true(all(c(format(paris), format(london), format(lyon)) %in%
                  format(cells)))
})

test_that("a5_linestring_to_cells handles an antimeridian-crossing path", {
  line <- wk::wkt("LINESTRING (170 0, -170 0)")
  cells <- a5_linestring_to_cells(line, resolution = 5)
  centres <- a5_cell_to_lonlat(cells)
  lons <- wk::wk_coords(centres)$x
  expect_true(any(lons > 160))
  expect_true(any(lons < -160))
})

# -- threading invariance ------------------------------------------------------

test_that("polygon/linestring results are identical with threads > 1", {
  prev <- a5_get_threads()
  on.exit(a5_set_threads(prev), add = TRUE)

  mp <- wk::wkt(
    "MULTIPOLYGON (((-3.5 55.5, -2.5 55.5, -2.5 56.5, -3.5 56.5, -3.5 55.5)),
                   ((1 1, 2 1, 2 2, 1 2, 1 1)))"
  )
  ml <- wk::wkt(
    "MULTILINESTRING ((2.35 48.86, -0.13 51.51),
                      (-0.13 51.51, 4.83 45.76))"
  )

  a5_set_threads(1)
  p1 <- a5_polygon_to_cells(mp, resolution = 8)
  l1 <- a5_linestring_to_cells(ml, resolution = 5)

  a5_set_threads(2)
  p2 <- a5_polygon_to_cells(mp, resolution = 8)
  l2 <- a5_linestring_to_cells(ml, resolution = 5)

  expect_setequal(format(p1), format(p2))
  expect_identical(format(l1), format(l2))
})

# -- empty-result behaviour ---------------------------------------------------

test_that("polygon smaller than a single cell returns an empty a5_cell", {
  # A res-14 cell's pentagon is several orders of magnitude smaller than
  # any res-8 cell, so no res-8 centre can lie inside it.
  tiny <- a5_cell_to_boundary(a5_lonlat_to_cell(10, 50, resolution = 14))
  out <- a5_polygon_to_cells(tiny, resolution = 8)
  expect_s3_class(out, "a5_cell")
  expect_length(out, 0L)
})

# -- input validation: polygon -------------------------------------------------

test_that("polygon matrix input must have exactly 2 columns", {
  expect_error(
    a5_polygon_to_cells(matrix(0, nrow = 4, ncol = 3), resolution = 8),
    "exactly 2 columns"
  )
  expect_error(
    a5_polygon_to_cells(matrix(0, nrow = 4, ncol = 1), resolution = 8),
    "exactly 2 columns"
  )
})

test_that("polygon matrix input must be numeric", {
  m <- matrix(c("a", "b", "c", "d", "e", "f"), ncol = 2)
  expect_error(
    a5_polygon_to_cells(m, resolution = 8),
    "must be numeric"
  )
})

test_that("polygon data.frame input must have lon and lat columns", {
  expect_error(
    a5_polygon_to_cells(data.frame(x = 1:3, y = 1:3), resolution = 8),
    "lon.*lat"
  )
  expect_error(
    a5_polygon_to_cells(data.frame(lon = 1:3), resolution = 8),
    "lon.*lat"
  )
})

test_that("polygon vertex count is enforced", {
  expect_error(
    a5_polygon_to_cells(matrix(c(0, 0, 1, 1), ncol = 2), resolution = 8),
    "at least 3 vertices"
  )
})

test_that("polygon NA coordinates error", {
  expect_error(
    a5_polygon_to_cells(
      data.frame(lon = c(0, 1, NA, 0), lat = c(0, 0, 1, 0)),
      resolution = 8
    ),
    "NA"
  )
  expect_error(
    a5_polygon_to_cells(
      matrix(c(0, 1, NA, 0, 0, 0, 1, 0), ncol = 2),
      resolution = 8
    ),
    "NA"
  )
})

test_that("polygon rejects wrong geometry types", {
  expect_error(
    a5_polygon_to_cells(wk::wkt("POINT (0 0)"), resolution = 8),
    "POLYGON"
  )
  expect_error(
    a5_polygon_to_cells(wk::wkt("LINESTRING (0 0, 1 1)"), resolution = 8),
    "POLYGON"
  )
})

test_that("polygon rejects empty geometries", {
  expect_error(
    a5_polygon_to_cells(wk::wkt("POLYGON EMPTY"), resolution = 8),
    "empty geometries"
  )
})

test_that("polygon rejects empty sfc / wkt vector", {
  expect_error(
    a5_polygon_to_cells(wk::wkt(character()), resolution = 8),
    "no polygon rings"
  )
})

test_that("polygon rejects non-geometry input", {
  # A plain list isn't wk-handleable.
  expect_error(
    a5_polygon_to_cells(list(a = 1, b = 2), resolution = 8),
    "Could not interpret"
  )
})

# -- input validation: linestring ---------------------------------------------

test_that("linestring matrix input must have exactly 2 columns", {
  expect_error(
    a5_linestring_to_cells(matrix(0, nrow = 4, ncol = 3), resolution = 8),
    "exactly 2 columns"
  )
})

test_that("linestring matrix input must be numeric", {
  m <- matrix(c("a", "b", "c", "d"), ncol = 2)
  expect_error(
    a5_linestring_to_cells(m, resolution = 8),
    "must be numeric"
  )
})

test_that("linestring data.frame input must have lon and lat columns", {
  expect_error(
    a5_linestring_to_cells(data.frame(x = 1:3, y = 1:3), resolution = 8),
    "lon.*lat"
  )
})

test_that("linestring vertex count is enforced", {
  expect_error(
    a5_linestring_to_cells(matrix(c(0, 0), ncol = 2), resolution = 8),
    "at least 2 vertices"
  )
})

test_that("linestring NA coordinates error", {
  expect_error(
    a5_linestring_to_cells(
      data.frame(lon = c(0, NA, 1), lat = c(0, 0, 1)),
      resolution = 8
    ),
    "NA"
  )
  expect_error(
    a5_linestring_to_cells(
      matrix(c(0, NA, 1, 0, 0, 1), ncol = 2),
      resolution = 8
    ),
    "NA"
  )
})

test_that("linestring rejects wrong geometry types", {
  expect_error(
    a5_linestring_to_cells(wk::wkt("POINT (0 0)"), resolution = 8),
    "LINESTRING"
  )
  expect_error(
    a5_linestring_to_cells(
      wk::wkt(square_wkt(0, 0, 1, 1)),
      resolution = 8
    ),
    "LINESTRING"
  )
})

test_that("linestring rejects empty geometries", {
  expect_error(
    a5_linestring_to_cells(wk::wkt("LINESTRING EMPTY"), resolution = 8),
    "empty geometries"
  )
})

test_that("linestring rejects empty sfc / wkt vector", {
  expect_error(
    a5_linestring_to_cells(wk::wkt(character()), resolution = 8),
    "no linestrings"
  )
})

test_that("linestring rejects non-geometry input", {
  expect_error(
    a5_linestring_to_cells(list(a = 1, b = 2), resolution = 8),
    "Could not interpret"
  )
})

# -- input validation: shared ------------------------------------------------

test_that("resolution must be in 0..30", {
  poly <- wk::wkt(square_wkt(0, 0, 1, 1))
  line <- wk::wkt("LINESTRING (0 0, 1 1)")
  expect_error(a5_polygon_to_cells(poly, resolution = -1), "resolution")
  expect_error(a5_polygon_to_cells(poly, resolution = 31), "resolution")
  expect_error(a5_linestring_to_cells(line, resolution = -1), "resolution")
  expect_error(a5_linestring_to_cells(line, resolution = 31), "resolution")
})

test_that("resolution must be scalar", {
  poly <- wk::wkt(square_wkt(0, 0, 1, 1))
  line <- wk::wkt("LINESTRING (0 0, 1 1)")
  expect_error(a5_polygon_to_cells(poly, resolution = c(5, 6)), "size 1")
  expect_error(a5_linestring_to_cells(line, resolution = c(5, 6)), "size 1")
})

#' Convert between a5_cell and Arrow uint64 arrays
#'
#' Losslessly convert between [a5_cell] vectors and Arrow `uint64`
#' arrays. This avoids the precision loss that occurs when Arrow
#' converts `uint64` to R's `double` (which can only represent
#' integers exactly up to 2^53, while A5 cell IDs span the full
#' 0--2^64 range).
#'
#' @param x For `a5_cell_from_arrow()`, an Arrow `Array` or
#'   `ChunkedArray` of type `uint64`. For `a5_cell_to_arrow()`, an
#'   [a5_cell] vector.
#' @returns `a5_cell_from_arrow()` returns an [a5_cell] vector.
#'   `a5_cell_to_arrow()` returns an Arrow `Array` of type `uint64`.
#'
#' @details
#' Internally these use Arrow's zero-copy `View()` to reinterpret
#' `uint64` bytes as `fixed_size_binary(8)`, then convert to/from the
#' raw-byte representation used by [a5_cell]. The resulting Arrow
#' arrays can be written directly to Parquet and read correctly by
#' DuckDB, Python, and other Arrow-compatible tools.
#'
#' @seealso [a5_u64_to_hex()] for converting to hex strings instead.
#'
#' @export
#' @examplesIf requireNamespace("arrow", quietly = TRUE)
#' cell <- a5_lonlat_to_cell(135, 0, resolution = 10)
#' arr <- a5_cell_to_arrow(cell)
#' back <- a5_cell_from_arrow(arr)
#' identical(format(cell), format(back))
a5_cell_from_arrow <- function(x) {
  rlang::check_installed("arrow", reason = "to convert Arrow arrays to a5_cell")

  type_str <- x$type$ToString()
  if (!identical(type_str, "uint64")) {
    cli::cli_abort(c(
      "Expected an Arrow {.code uint64} array.",
      "x" = "Got {.code {type_str}}."
    ))
  }

  # Handle ChunkedArray by processing each chunk
  if (inherits(x, "ChunkedArray")) {
    parts <- lapply(seq_len(x$num_chunks) - 1L, function(i) {
      chunk_to_a5(x$chunk(i))
    })
    return(do.call(vctrs::vec_c, parts))
  }

  chunk_to_a5(x)
}

#' @rdname a5_cell_from_arrow
#' @export
#' @examplesIf requireNamespace("arrow", quietly = TRUE)
#' cells <- a5_lonlat_to_cell(c(-3.19, 135), c(55.95, 0), resolution = 10)
#' arr <- a5_cell_to_arrow(cells)
#' arr$type$ToString()
a5_cell_to_arrow <- function(x) {
  rlang::check_installed("arrow", reason = "to convert a5_cell to Arrow arrays")
  x <- as_a5_cell(x)

  raw_list <- raw8_to_blobs_rs(cell_data(x))

  bin_arr <- arrow::Array$create(raw_list, type = arrow::fixed_size_binary(8L))
  bin_arr$View(arrow::uint64())
}

#' Convert a single Arrow uint64 chunk to a5_cell
#' @noRd
chunk_to_a5 <- function(chunk) {
  fsb <- chunk$View(arrow::fixed_size_binary(8L))
  raw_list <- fsb$as_vector()
  rs <- blobs_to_raw8_rs(raw_list)
  cells_from_rs(rs)
}

test_that("a5_cell_to_arrow returns uint64 array", {
  skip_if_not_installed("arrow")
  cells <- a5_lonlat_to_cell(c(-3.19, 135, -60), c(55.95, 0, 45), 10)
  arr <- a5_cell_to_arrow(cells)
  expect_s3_class(arr, "Array")
  expect_equal(arr$type$ToString(), "uint64")
  expect_equal(arr$length(), 3L)
})

test_that("a5_cell_from_arrow rejects non-uint64", {
  skip_if_not_installed("arrow")
  arr <- arrow::Array$create(1:3, type = arrow::int32())
  expect_error(a5_cell_from_arrow(arr), "uint64")
})

test_that("round-trip a5_cell -> arrow -> a5_cell is lossless", {
  skip_if_not_installed("arrow")
  # Include cells with bit 63 set (origin >= 6) — these exceed 2^53
  cells <- a5_lonlat_to_cell(
    c(-3.19, 135, -60, 0, 179),
    c(55.95, 0, 45, 0, -89),
    resolution = 10
  )
  arr <- a5_cell_to_arrow(cells)
  back <- a5_cell_from_arrow(arr)
  expect_equal(format(back), format(cells))
})

test_that("round-trip preserves cells across all origins", {
  skip_if_not_installed("arrow")
  # Systematically test points that hit different icosahedron faces
  lons <- seq(-180, 170, by = 30)
  lats <- rep(c(-45, 0, 45), length.out = length(lons))
  cells <- a5_lonlat_to_cell(lons, lats, resolution = 15)
  arr <- a5_cell_to_arrow(cells)
  back <- a5_cell_from_arrow(arr)
  expect_equal(format(back), format(cells))
})

test_that("NA round-trips through arrow", {
  skip_if_not_installed("arrow")
  cells <- a5_cell(c("0800000000000006", NA, "633e000000000000"))
  arr <- a5_cell_to_arrow(cells)
  expect_true(arr$IsNull(1))
  expect_false(arr$IsNull(0))
  back <- a5_cell_from_arrow(arr)
  expect_equal(format(back), format(cells))
  expect_true(is.na(back[2]))
})

test_that("a5_cell_from_arrow handles ChunkedArray", {
  skip_if_not_installed("arrow")
  cells <- a5_lonlat_to_cell(c(0, 135), c(0, 0), 10)
  arr <- a5_cell_to_arrow(cells)
  chunked <- arrow::chunked_array(arr, arr)
  back <- a5_cell_from_arrow(chunked)
  expect_length(back, 4L)
  expect_equal(format(back), rep(format(cells), 2))
})

test_that("parquet round-trip is lossless", {
  skip_if_not_installed("arrow")
  cells <- a5_lonlat_to_cell(
    c(-3.19, 135, -60, 0, 179),
    c(55.95, 0, 45, 0, -89),
    resolution = 10
  )

  tf <- tempfile(fileext = ".parquet")
  on.exit(unlink(tf))

  # Write
  tbl <- arrow::arrow_table(cell_id = a5_cell_to_arrow(cells))
  arrow::write_parquet(tbl, tf)

  # Read
  back_tbl <- arrow::read_parquet(tf, as_data_frame = FALSE)
  back_cells <- a5_cell_from_arrow(back_tbl$column(0))
  expect_equal(format(back_cells), format(cells))
})

test_that("high resolution survives parquet round-trip", {
  skip_if_not_installed("arrow")
  # High resolution uses many bits of the u64
  cells <- a5_lonlat_to_cell(c(-60, 135, 0), c(45, 0, 0), resolution = 20)
  arr <- a5_cell_to_arrow(cells)
  back <- a5_cell_from_arrow(arr)
  expect_equal(format(back), format(cells))
  expect_true(all(a5_get_resolution(back) == 20L))
})

test_that("resolution 29 round-trips through arrow (vectorised)", {
  skip_if_not_installed("arrow")
  # res 29 uses nearly every bit of the u64 — precision stress test
  # (res 30 has an upstream bug in the a5 crate)
  lons <- seq(-180, 170, by = 10)
  lats <- rep(c(-80, -40, 0, 40, 80), length.out = length(lons))
  cells <- a5_lonlat_to_cell(lons, lats, resolution = 29)
  expect_true(all(a5_get_resolution(cells) == 29L))
  arr <- a5_cell_to_arrow(cells)
  back <- a5_cell_from_arrow(arr)
  expect_equal(format(back), format(cells))
})

#' @title a5R: A5 Pentagonal Geospatial Index for R
#'
#' @description
#' R bindings for the [A5](https://a5geo.org/) pentagonal geospatial index /
#' discrete global grid system, powered by the a5 Rust crate via extendr.
#'
#' @section Cell type:
#' - [a5_cell()] --- construct cell ID vectors
#' - [is_a5_cell()][a5_cell] / [a5_is_valid()][a5_cell] --- type test and
#'   validation
#' - [a5_u64_to_hex()] / [a5_hex_to_u64()][a5_u64_to_hex] --- hex string
#'   conversion
#'
#' @section Indexing:
#' - [a5_lonlat_to_cell()] --- coordinates to cell IDs
#' - [a5_cell_to_lonlat()] --- cell IDs to centre coordinates
#'
#' @section Geometry:
#' - [a5_cell_to_boundary()] --- cell boundary polygons (WKB or WKT)
#' - [a5_cell_area()] --- cell area at a given resolution
#' - [a5_cell_distance()] --- distance between cell centroids
#' - [a5_get_num_cells()] --- total cell count at a resolution
#' - [a5_get_num_children()] --- child count between resolutions
#'
#' @section Hierarchy:
#' - [a5_get_resolution()] --- extract resolution from cell IDs
#' - [a5_cell_to_parent()] --- navigate to coarser cells
#' - [a5_cell_to_children()] --- navigate to finer cells
#' - [a5_get_res0_cells()] --- the 12 root cells
#' - [a5_compact()] / [a5_uncompact()] --- compress and expand cell sets
#'
#' @section Traversal:
#' - [a5_grid_disk()] --- neighbours by hop count
#' - [a5_spherical_cap()] --- neighbours by great-circle distance
#'
#' @section Geometry indexing:
#' - [a5_polygon_to_cells()] --- cells whose centres lie inside a polygon
#' - [a5_linestring_to_cells()] --- cells crossed by a great-circle polyline
#'
#' @section Arrow & Parquet:
#' - [a5_cell_from_arrow()] / [a5_cell_to_arrow()][a5_cell_from_arrow] ---
#'   lossless conversion to/from Arrow `uint64`
#'
#' @section Configuration:
#' - [a5_set_threads()] / [a5_get_threads()] --- multi-threading control
#'
#' @section Vignettes:
#' - `vignette("a5R")` --- getting started
#' - `vignette("multithreading")` --- parallel processing
#' - `vignette("internal-cell-representation")` --- how cell IDs are stored
#' - `vignette("arrow-parquet")` --- Arrow and Parquet interop
#'
#' @keywords internal
"_PACKAGE"

#' Compact a set of A5 cells
#'
#' Merges complete sibling groups into their common parent, reducing the
#' number of cells while preserving coverage.
#'
#' @param cells An [a5_cell] vector.
#' @returns An [a5_cell] vector of compacted cells.
#'
#' @seealso [a5_uncompact()]
#' @export
#' @examples
#' cell <- a5_lonlat_to_cell(-3.19, 55.95, resolution = 5)
#' children <- a5_cell_to_children(cell)
#' a5_compact(children) # back to the parent
a5_compact <- function(cells) {
  cells <- as_a5_cell(cells)
  cells_from_rs(a5_compact_rs(cell_data(cells)))
}

#' Uncompact a set of A5 cells to a target resolution
#'
#' Expands each cell to its descendants at the target resolution.
#'
#' @param cells An [a5_cell] vector.
#' @param resolution Integer scalar target resolution (0--30).
#' @returns An [a5_cell] vector of uncompacted cells.
#'
#' @seealso [a5_compact()]
#' @export
#' @examples
#' cell <- a5_lonlat_to_cell(-3.19, 55.95, resolution = 5)
#' a5_uncompact(cell, resolution = 7)
a5_uncompact <- function(cells, resolution) {
  cells <- as_a5_cell(cells)
  resolution <- vctrs::vec_cast(resolution, integer())
  check_resolution(resolution)
  vctrs::vec_assert(resolution, size = 1L)
  cells_from_rs(a5_uncompact_rs(cell_data(cells), resolution))
}

test_that("compact reverses children", {
  cell <- a5_lonlat_to_cell(0, 0, resolution = 5)
  children <- a5_cell_to_children(cell)
  compacted <- a5_compact(children)
  expect_equal(format(compacted), format(cell))
})

test_that("compact with partial siblings is a no-op", {
  cell <- a5_lonlat_to_cell(0, 0, resolution = 5)
  children <- a5_cell_to_children(cell)
  partial <- children[1:2]
  compacted <- a5_compact(partial)
  expect_length(compacted, 2L)
})

test_that("compact with single cell is a no-op", {
  cell <- a5_lonlat_to_cell(0, 0, resolution = 5)
  compacted <- a5_compact(cell)
  expect_equal(format(compacted), format(cell))
})

test_that("uncompact expands to target resolution", {
  cell <- a5_lonlat_to_cell(0, 0, resolution = 5)
  expanded <- a5_uncompact(cell, resolution = 7)
  expect_length(expanded, 16L) # 4^2 = 16
  expect_true(all(a5_get_resolution(expanded) == 7L))
})

test_that("compact and uncompact round-trip", {
  cell <- a5_lonlat_to_cell(0, 0, resolution = 5)
  expanded <- a5_uncompact(cell, resolution = 7)
  compacted <- a5_compact(expanded)
  expect_equal(format(compacted), format(cell))
})

test_that("compact skips NA cells", {
  cell <- a5_lonlat_to_cell(0, 0, resolution = 5)
  children <- a5_cell_to_children(cell)
  with_na <- vctrs::vec_c(children, a5_cell(NA))
  compacted <- a5_compact(with_na)
  # NA is skipped; 4 siblings still compact to parent
  expect_equal(format(compacted), format(cell))
})

#' @importFrom rlang caller_env
#' @importFrom vctrs vec_cast vec_ptype2
NULL

check_resolution <- function(resolution,
                             min = 0L,
                             max = 30L,
                             call = rlang::caller_env()) {
  bad <- !is.na(resolution) & (resolution < min | resolution > max)
  if (any(bad)) {
    cli::cli_abort(
      "{.arg resolution} must be between {min} and {max}, not {resolution[which(bad)[1]]}.",
      call = call
    )
  }
  invisible(resolution)
}

#' Convert coordinates to A5 cell indices
#'
#' Maps longitude/latitude coordinates to A5 cell indices at the specified
#' resolution.
#'
#' @param lon Numeric vector of longitudes in degrees.
#' @param lat Numeric vector of latitudes in degrees.
#' @param resolution Integer scalar or vector of resolutions (0--30).
#' @returns An [a5_cell] vector of cell indices.
#'
#' @seealso [a5_cell_to_lonlat()] for the inverse operation.
#' @export
#' @examples
#' a5_lonlat_to_cell(-3.19, 55.95, resolution = 5)
a5_lonlat_to_cell <- function(lon, lat, resolution) {
  args <- vctrs::vec_recycle_common(
    lon = vctrs::vec_cast(lon, double()),
    lat = vctrs::vec_cast(lat, double()),
    resolution = vctrs::vec_cast(resolution, integer())
  )
  check_resolution(args$resolution)
  cells_from_rs(a5_lonlat_to_cell_rs(args$lon, args$lat, args$resolution))
}

#' Convert A5 cell indices to coordinates
#'
#' Returns the centre-point longitude and latitude of each cell.
#'
#' @param cell An [a5_cell] vector (or character coercible to one).
#' @param as_dataframe Logical scalar controlling the return container.
#'   When `FALSE` (default), centroids are returned as a [wk::xy()] vector
#'   with WGS 84 CRS, the geographic-typed form that plugs into wk/sf
#'   pipelines. When `TRUE`, centroids are returned as a base `data.frame`
#'   with columns `lon` and `lat`.
#' @returns A [wk::xy()] vector (if `as_dataframe = FALSE`) or a
#'   `data.frame` with columns `lon` and `lat` (if `as_dataframe = TRUE`).
#'
#' @seealso [a5_lonlat_to_cell()] for the inverse operation,
#'   [a5_cell_to_boundary()] for full cell polygons.
#' @export
#' @examples
#' cell <- a5_lonlat_to_cell(-3.19, 55.95, resolution = 5)
#' a5_cell_to_lonlat(cell)
#'
#' # Data frame output
#' cell2 <- a5_lonlat_to_cell(114.8, 4.1, resolution = 5)
#' a5_cell_to_lonlat(cell2, as_dataframe = TRUE)
a5_cell_to_lonlat <- function(cell, as_dataframe = FALSE) {
  cell <- as_a5_cell(cell)
  ll <- a5_cell_to_lonlat_rs(cell_data(cell), TRUE)
  if (as_dataframe) {
    data.frame(lon = ll$lon, lat = ll$lat)
  } else {
    wk::xy(ll$lon, ll$lat, crs = wk::wk_crs_longlat())
  }
}

test_that("default threads is 1", {
    expect_equal(a5_get_threads(), 1L)
})

test_that("set/get threads works", {
    withr::defer(a5_set_threads(1L))
    old <- a5_set_threads(2L)
    expect_equal(old, 1L)
    expect_equal(a5_get_threads(), 2L)
})

test_that("results are identical with 1 vs 2 threads", {
    withr::defer(a5_set_threads(1L))
    cells <- a5_lonlat_to_cell(c(0, 10, -10), c(0, 20, -20), resolution = 5)

    a5_set_threads(1L)
    boundary_1 <- a5_cell_to_boundary(cells, format = "wkt")
    lonlat_1 <- a5_cell_to_lonlat(cells)
    parent_1 <- a5_cell_to_parent(cells)
    res_1 <- a5_get_resolution(cells)
    valid_1 <- a5_is_valid(cells)

    a5_set_threads(2L)
    boundary_2 <- a5_cell_to_boundary(cells, format = "wkt")
    lonlat_2 <- a5_cell_to_lonlat(cells)
    parent_2 <- a5_cell_to_parent(cells)
    res_2 <- a5_get_resolution(cells)
    valid_2 <- a5_is_valid(cells)

    expect_identical(as.character(boundary_1), as.character(boundary_2))
    expect_identical(lonlat_1, lonlat_2)
    expect_identical(as.character(parent_1), as.character(parent_2))
    expect_identical(res_1, res_2)
    expect_identical(valid_1, valid_2)
})

# -- a5_cell_distance ----------------------------------------------------------

test_that("cell_distance returns units vector in metres", {
  a <- a5_lonlat_to_cell(0, 0, resolution = 8)
  b <- a5_lonlat_to_cell(1, 1, resolution = 8)
  d <- a5_cell_distance(a, b)
  expect_s3_class(d, "units")
  expect_equal(units::deparse_unit(d), "m")
  expect_true(d > units::set_units(0, "m"))
})

test_that("cell_distance of a cell to itself is zero", {
  cell <- a5_lonlat_to_cell(10, 50, resolution = 10)
  d <- a5_cell_distance(cell, cell)
  expect_equal(as.numeric(d), 0)
})

test_that("cell_distance supports unit conversion", {
  a <- a5_lonlat_to_cell(0, 0, resolution = 8)
  b <- a5_lonlat_to_cell(1, 1, resolution = 8)
  d_m <- a5_cell_distance(a, b)
  d_km <- a5_cell_distance(a, b, units = "km")
  expect_equal(as.numeric(d_km), as.numeric(d_m) / 1000, tolerance = 1e-10)
})

test_that("cell_distance is vectorised and recycled", {
  origin <- a5_lonlat_to_cell(0, 0, resolution = 8)
  targets <- a5_lonlat_to_cell(c(1, 2, 3), c(1, 2, 3), resolution = 8)
  d <- a5_cell_distance(origin, targets)
  expect_length(d, 3L)
  # distances should be increasing
  expect_true(all(diff(as.numeric(d)) > 0))
  e <- a5_cell_distance(targets, origin)
  expect_equal(as.numeric(d), as.numeric(e))
})

test_that("cell_distance handles NA", {
  a <- a5_lonlat_to_cell(0, 0, resolution = 8)
  d <- a5_cell_distance(a, a5_cell(NA))
  expect_true(is.na(d))
})

test_that("cell_distance is close to s2/sf", {
  skip_if_not_installed("sf")
  uses2 <- sf::sf_use_s2(TRUE)
  withr::defer(sf::sf_use_s2(uses2))

  a <- a5_lonlat_to_cell(-3.19, 55.95, resolution = 10)
  b <- a5_lonlat_to_cell(-2.0, 55.0, resolution = 10)
  d <- as.numeric(a5_cell_distance(a, b))

  p1 <- sf::st_sfc(sf::st_point(as.numeric(a5_cell_to_lonlat(a))), crs = 4326)
  p2 <- sf::st_sfc(sf::st_point(as.numeric(a5_cell_to_lonlat(b))), crs = 4326)

  sf_d <- as.numeric(sf::st_distance(p1, p2))

  expect_equal(d, sf_d, tolerance = 0.1) # within 1 metre.
})

test_that("cell_distance with units = NULL returns plain numeric in metres", {
  a <- a5_lonlat_to_cell(0, 0, resolution = 8)
  b <- a5_lonlat_to_cell(1, 1, resolution = 8)
  d_null <- a5_cell_distance(a, b, units = NULL)
  d_m <- a5_cell_distance(a, b, units = "m")
  expect_type(d_null, "double")
  expect_false(inherits(d_null, "units"))
  expect_equal(d_null, as.numeric(d_m))
})

test_that("cell_distance rejects invalid units", {
  a <- a5_lonlat_to_cell(0, 0, resolution = 8)
  expect_error(a5_cell_distance(a, a, units = "kg"), "distance unit")
})

test_that("cell_distance default method is haversine", {
  a <- a5_lonlat_to_cell(-3.19, 55.95, resolution = 8)
  b <- a5_lonlat_to_cell(-2.0, 55.0, resolution = 8)
  d_default <- as.numeric(a5_cell_distance(a, b))
  d_explicit <- as.numeric(a5_cell_distance(a, b, method = "haversine"))
  expect_equal(d_default, d_explicit)
})

test_that("cell_distance geodesic is close to haversine", {
  a <- a5_lonlat_to_cell(-3.19, 55.95, resolution = 8)
  b <- a5_lonlat_to_cell(-2.0, 55.0, resolution = 8)
  d_hav <- as.numeric(a5_cell_distance(a, b, method = "haversine"))
  d_geo <- as.numeric(a5_cell_distance(a, b, method = "geodesic"))

  # Geodesic distance should be greater than or equal to haversine (great circle) distance
  expect_gte(d_geo, d_hav)
  # But equal within 0.3% of each other
  expect_equal(d_geo, d_hav, tolerance = 0.003)
})

test_that("cell_distance rhumb >= haversine (great circle)", {
  a <- a5_lonlat_to_cell(-3.19, 55.95, resolution = 8)
  b <- a5_lonlat_to_cell(-2.0, 55.0, resolution = 8)
  d_hav <- as.numeric(a5_cell_distance(a, b, method = "haversine"))
  d_rhumb <- as.numeric(a5_cell_distance(a, b, method = "rhumb"))
  expect_true(d_rhumb >= d_hav)
})

test_that("cell_distance rejects invalid method", {
  a <- a5_lonlat_to_cell(0, 0, resolution = 8)
  expect_error(a5_cell_distance(a, a, method = "euclidean"), "must be one of")
})

test_that("lonlat_to_cell returns a5_cell", {
  cell <- a5_lonlat_to_cell(-3.19, 55.95, resolution = 5)
  expect_s3_class(cell, "a5_cell")
  expect_length(cell, 1L)
})

test_that("round-trip lonlat -> cell -> lonlat is stable", {
  lon <- -3.19
  lat <- 55.95
  cell <- a5_lonlat_to_cell(lon, lat, resolution = 15)
  pt <- a5_cell_to_lonlat(cell)
  expect_equal(wk::wk_coords(pt)$x, lon, tolerance = 0.01)
  expect_equal(wk::wk_coords(pt)$y, lat, tolerance = 0.01)
})

test_that("vectorised indexing works", {
  lons <- c(-3.19, 0, 140)
  lats <- c(55.95, 0, 35.68)
  cells <- a5_lonlat_to_cell(lons, lats, resolution = 5)
  expect_length(cells, 3L)
  expect_s3_class(cells, "a5_cell")
})

test_that("NA handling in lonlat_to_cell", {
  cell <- a5_lonlat_to_cell(NA_real_, 55.95, resolution = 5)
  expect_true(is.na(cell[1]))

  cell2 <- a5_lonlat_to_cell(0, NA_real_, resolution = 5)
  expect_true(is.na(cell2[1]))

  cell3 <- a5_lonlat_to_cell(0, 0, resolution = NA_integer_)
  expect_true(is.na(cell3[1]))
})

test_that("cell_to_lonlat handles NA", {
  cells <- a5_cell(c("0800000000000006", NA))
  pts <- a5_cell_to_lonlat(cells)
  coords <- wk::wk_coords(pts)
  expect_false(is.na(coords$x[1]))
  expect_true(is.na(coords$x[2]))
})

test_that("cell_to_lonlat returns xy with longlat CRS", {
  cell <- a5_lonlat_to_cell(0, 0, resolution = 5)
  pt <- a5_cell_to_lonlat(cell)
  expect_s3_class(pt, "wk_xy")
  expect_identical(wk::wk_crs(pt), wk::wk_crs_longlat())
})

test_that("lonlat_to_cell rejects invalid resolution", {
  expect_error(a5_lonlat_to_cell(0, 0, resolution = 31), "resolution")
  expect_error(a5_lonlat_to_cell(0, 0, resolution = -1), "resolution")
})

test_that("cell_to_lonlat as_dataframe = TRUE returns a data frame", {
  cell <- a5_lonlat_to_cell(114.8, 4.1, resolution = 5)
  df <- a5_cell_to_lonlat(cell, as_dataframe = TRUE)
  expect_s3_class(df, "data.frame")
  expect_named(df, c("lon", "lat"))
  expect_true(df$lon >= -180 && df$lon <= 180)
})

test_that("cell_to_lonlat default wraps longitude to [-180, 180]", {
  cell <- a5_lonlat_to_cell(114.8, 4.1, resolution = 5)
  pt <- a5_cell_to_lonlat(cell)
  lon <- wk::wk_coords(pt)$x
  expect_true(lon >= -180 && lon <= 180)
})

test_that("cell_to_lonlat rejects the removed `normalise` argument", {
  cell <- a5_lonlat_to_cell(114.8, 4.1, resolution = 5)
  expect_error(
    a5_cell_to_lonlat(cell, normalise = TRUE),
    "unused argument"
  )
})

# -- deprecation ---------------------------------------------------------------

test_that("a5_grid is soft-deprecated in favour of a5_polygon_to_cells", {
  lifecycle::expect_deprecated(
    a5_grid(c(-3.3, 55.9, -3.1, 56.0), resolution = 5),
    "a5_polygon_to_cells"
  )
})

# -- input validation -----------------------------------------------------------

test_that("a5_grid rejects bad bbox", {
  local_quiet_lifecycle()
  expect_error(a5_grid(c(1, 2, 3), resolution = 3), "length 4")
  expect_error(a5_grid(c(1, 2, NA, 4), resolution = 3), "NA")
  expect_error(a5_grid(c(0, 10, 1, 5), resolution = 3), "ymin")
  expect_error(a5_grid(c(0, 5, 1, 5), resolution = 3), "ymin")
  expect_error(a5_grid(c(5, 0, 5, 1), resolution = 3), "xmin.*xmax.*equal")
})

test_that("a5_grid rejects invalid resolution", {
  local_quiet_lifecycle()
  expect_error(a5_grid(c(0, 0, 1, 1), resolution = -1))
  expect_error(a5_grid(c(0, 0, 1, 1), resolution = 31))
})

# -- return type ---------------------------------------------------------------

test_that("a5_grid returns a5_cell at correct resolution", {
  local_quiet_lifecycle()
  cells <- a5_grid(c(-3.3, 55.9, -3.1, 56.0), resolution = 5)
  expect_s3_class(cells, "a5_cell")
  expect_true(length(cells) > 0)
  expect_true(all(a5_get_resolution(cells) == 5L))
})

# -- correctness ---------------------------------------------------------------

test_that("res 0 with global bbox returns 12 cells", {
  local_quiet_lifecycle()
  cells <- a5_grid(c(-180, -90, 180, 90), resolution = 0)
  expect_length(cells, 12L)
})

test_that("result cells intersect the target (bbox path)", {
  local_quiet_lifecycle()
  bbox <- c(-3.3, 55.9, -3.1, 56.0)
  cells <- a5_grid(bbox, resolution = 5)
  target_wkt <- sprintf(
    "POLYGON ((%s %s, %s %s, %s %s, %s %s, %s %s))",
    bbox[1], bbox[2], bbox[3], bbox[2], bbox[3], bbox[4],
    bbox[1], bbox[4], bbox[1], bbox[2]
  )
  filtered <- cells_from_rs(a5_grid_intersects_rs(
    cell_data(cells), target_wkt
  ))
  expect_equal(length(filtered), length(cells))
})

test_that("geometry input filters cells by intersection", {
  local_quiet_lifecycle()
  # Triangle is strictly smaller than its bounding box, so the exact

  # intersection filter should return fewer cells than a bbox grid
  tri <- wk::wkt("POLYGON ((-3.3 55.9, -3.1 55.95, -3.3 56, -3.3 55.9))")
  tri_cells <- a5_grid(tri, resolution = 8)
  bbox_cells <- a5_grid(c(-3.3, 55.9, -3.1, 56.0), resolution = 8)
  expect_true(length(tri_cells) < length(bbox_cells))
  # All triangle cells should be a subset of the bbox cells
  expect_true(all(format(tri_cells) %in% format(bbox_cells)))
})

test_that("interior point is covered by a returned cell", {
  local_quiet_lifecycle()
  bbox <- c(-3.3, 55.9, -3.1, 56.0)
  cells <- a5_grid(bbox, resolution = 5)
  # the centroid of the bbox should fall inside one of the cells
  centre <- a5_lonlat_to_cell(-3.2, 55.95, resolution = 5)
  expect_true(any(format(cells) == format(centre)))
})

test_that("no duplicate cells", {
  local_quiet_lifecycle()
  cells <- a5_grid(c(-3.3, 55.9, -3.1, 56.0), resolution = 5)
  expect_equal(length(cells), length(unique(format(cells))))
})

# -- input types ---------------------------------------------------------------

test_that("a5_grid accepts wkt polygon", {
  local_quiet_lifecycle()
  poly <- wk::wkt("POLYGON ((-3.3 55.9, -3.1 55.9, -3.1 56, -3.3 56, -3.3 55.9))")
  cells <- a5_grid(poly, resolution = 5)
  expect_s3_class(cells, "a5_cell")
  expect_true(length(cells) > 0)
})

test_that("a5_grid accepts a5_cell as area", {
  local_quiet_lifecycle()
  cell <- a5_lonlat_to_cell(-3.19, 55.95, resolution = 3)
  children <- a5_grid(cell, resolution = 5)
  expect_s3_class(children, "a5_cell")
  expect_true(length(children) > 0)
  expect_true(all(a5_get_resolution(children) == 5L))
})

# -- edge cases ----------------------------------------------------------------

test_that("res 1 works (below filter threshold)", {
  local_quiet_lifecycle()
  cells <- a5_grid(c(-180, -90, 180, 90), resolution = 1)
  expect_s3_class(cells, "a5_cell")
  expect_true(all(a5_get_resolution(cells) == 1L))
})

test_that("tiny bbox returns at least 1 cell", {
  local_quiet_lifecycle()
  cells <- a5_grid(c(-3.19, 55.95, -3.189, 55.951), resolution = 8)
  expect_true(length(cells) >= 1L)
})

test_that("antimeridian-crossing bbox works", {
  local_quiet_lifecycle()
  # Fiji: bbox straddles the antimeridian (xmin > xmax)
  cells <- a5_grid(c(177, -19, -178, -17), resolution = 5)
  expect_s3_class(cells, "a5_cell")
  expect_true(length(cells) > 0)
  expect_true(all(a5_get_resolution(cells) == 5L))
  # cells on both sides of the antimeridian
  centres <- a5_cell_to_lonlat(cells)
  lons <- wk::wk_coords(centres)$x
  expect_true(any(lons > 0))
  expect_true(any(lons < 0))
})

test_that("near-pole bbox returns cells", {
  local_quiet_lifecycle()
  # Rust bbox grid skips filtering near poles (buf >= 45) to avoid false negatives
  cells <- a5_grid(c(0, 89.999, 0.001, 90), resolution = 5)
  expect_s3_class(cells, "a5_cell")
  expect_true(length(cells) >= 1L)
})

test_that("a5_grid accepts multiple geometries (GEOMETRYCOLLECTION path)", {
  local_quiet_lifecycle()
  # Two small polygons far apart — should return cells near both
  polys <- wk::wkt(c(
    "POLYGON ((-3.3 55.9, -3.1 55.9, -3.1 56, -3.3 56, -3.3 55.9))",
    "POLYGON ((135 -33, 136 -33, 136 -32, 135 -32, 135 -33))"
  ))
  cells <- a5_grid(polys, resolution = 5)
  expect_s3_class(cells, "a5_cell")
  expect_true(length(cells) > 0)
  centres <- a5_cell_to_lonlat(cells)
  lons <- wk::wk_coords(centres)$x
  # Cells near both polygons
  expect_true(any(lons < 0))
  expect_true(any(lons > 100))
})

test_that("a5_grid coerces double resolution to integer", {
  local_quiet_lifecycle()
  cells <- a5_grid(c(-3.3, 55.9, -3.1, 56.0), resolution = 5.0)
  expect_s3_class(cells, "a5_cell")
  expect_true(all(a5_get_resolution(cells) == 5L))
})

test_that("a5_grid rejects vector resolution", {
  local_quiet_lifecycle()
  expect_error(a5_grid(c(0, 0, 1, 1), resolution = c(3, 5)), "size 1")
})

#' Set the number of threads used by a5R
#'
#' Controls the number of threads used for parallel processing in vectorised
#' functions. Set to 1 (the default) for sequential processing with zero
#' overhead, or higher for parallel execution via rayon.
#'
#' @param n Integer scalar. Number of threads. Must be >= 1.
#' @returns Invisibly returns the previous thread count.
#' @rdname a5_threads
#' @export
a5_set_threads <- function(n = 1L) {
  n <- vctrs::vec_cast(n, integer())
  vctrs::vec_assert(n, size = 1L)
  old <- a5_get_threads_rs()
  a5_set_threads_rs(n)
  invisible(old)
}

#' Get the current number of threads
#'
#' @rdname a5_threads
#' @returns Integer scalar.
#' @export
a5_get_threads <- function() {
  a5_get_threads_rs()
}

# Generated by extendr: Do not edit by hand
#
# This file was created with the following call:
#   .Call("wrap__make_a5R_wrappers", use_symbols = TRUE, package_name = "a5R")

#' @usage NULL
#' @useDynLib a5R, .registration = TRUE
NULL

a5_set_threads_rs <- function(n) invisible(.Call(wrap__a5_set_threads_rs, n))

a5_get_threads_rs <- function() .Call(wrap__a5_get_threads_rs)

#' Convert cell raw bytes to hex strings (zero-padded to 16 chars).
#' @noRd
#' @keywords internal
raw8_to_hex_rs <- function(cells) .Call(wrap__raw8_to_hex_rs, cells)

#' Convert hex strings to cell raw bytes.
#' Returns list(b1 = raw(), ..., b8 = raw()).
#' @noRd
#' @keywords internal
hex_to_raw8_rs <- function(cells) .Call(wrap__hex_to_raw8_rs, cells)

#' Convert a list of raw(8) blobs (from Arrow) to cell raw bytes.
#' @noRd
#' @keywords internal
blobs_to_raw8_rs <- function(blobs) .Call(wrap__blobs_to_raw8_rs, blobs)

#' Convert cell raw bytes to a list of raw(8) blobs (for Arrow).
#' NA cells produce NULL elements.
#' @noRd
#' @keywords internal
raw8_to_blobs_rs <- function(cells) .Call(wrap__raw8_to_blobs_rs, cells)

#' Convert longitude/latitude coordinates to A5 cell indices.
#'
#' @param lon Numeric vector of longitudes (degrees).
#' @param lat Numeric vector of latitudes (degrees).
#' @param resolution Integer vector of resolutions (0--30).
#' @return A list with b1..b8 raw vectors.
#' @noRd
#' @keywords internal
a5_lonlat_to_cell_rs <- function(lon, lat, resolution) .Call(wrap__a5_lonlat_to_cell_rs, lon, lat, resolution)

#' Convert A5 cell indices to longitude/latitude coordinates.
#'
#' @param cells List with b1..b8 raw vectors.
#' @param normalise Logical: if TRUE, wrap longitudes to the standard range.
#' @return A list with `lon` and `lat` numeric vectors.
#' @noRd
#' @keywords internal
a5_cell_to_lonlat_rs <- function(cells, normalise) .Call(wrap__a5_cell_to_lonlat_rs, cells, normalise)

#' Get boundary polygons for A5 cells as WKT strings.
#'
#' @param cells List with b1..b8 raw vectors.
#' @param closed_ring Logical: should the polygon ring be closed?
#' @param segments Integer: number of interpolation segments per edge.
#' @return A character vector of WKT POLYGON strings.
#' @noRd
#' @keywords internal
a5_cell_to_boundary_rs <- function(cells, closed_ring, segments) .Call(wrap__a5_cell_to_boundary_rs, cells, closed_ring, segments)

#' Get boundary polygons for A5 cells as WKB raw vectors.
#'
#' @param cells List with b1..b8 raw vectors.
#' @param closed_ring Logical: should the polygon ring be closed?
#' @param segments Integer: number of interpolation segments per edge.
#' @return A list of raw vectors (WKB bytes) or NULL for NA cells.
#' @noRd
#' @keywords internal
a5_cell_to_boundary_wkb_rs <- function(cells, closed_ring, segments) .Call(wrap__a5_cell_to_boundary_wkb_rs, cells, closed_ring, segments)

#' Get the area (in square metres) of cells at a given resolution.
#'
#' @param resolution Integer vector of resolutions (0--30).
#' @return Numeric vector of areas in square metres.
#' @noRd
#' @keywords internal
a5_cell_area_rs <- function(resolution) .Call(wrap__a5_cell_area_rs, resolution)

#' Get total number of cells at a given resolution.
#'
#' @param resolution Integer scalar (0--30).
#' @return Numeric scalar (as double, since R has no u64).
#' @noRd
#' @keywords internal
a5_get_num_cells_rs <- function(resolution) .Call(wrap__a5_get_num_cells_rs, resolution)

#' Number of children per parent cell between two resolutions.
#'
#' @param parent_resolution Integer scalar.
#' @param child_resolution Integer scalar.
#' @return Numeric scalar (as double).
#' @noRd
#' @keywords internal
a5_get_num_children_rs <- function(parent_resolution, child_resolution) .Call(wrap__a5_get_num_children_rs, parent_resolution, child_resolution)

#' Validate cell IDs stored as raw bytes.
#'
#' @param cells List with b1..b8 raw vectors.
#' @return Logical vector indicating validity.
#' @noRd
#' @keywords internal
a5_is_valid_cell_rs <- function(cells) .Call(wrap__a5_is_valid_cell_rs, cells)

#' Validate hex cell ID strings (for use before conversion to rcrd).
#'
#' @param cell Character vector of hex strings to validate.
#' @return Logical vector indicating validity.
#' @noRd
#' @keywords internal
a5_is_valid_hex_rs <- function(cell) .Call(wrap__a5_is_valid_hex_rs, cell)

#' Distance between pairs of cell centroids.
#'
#' @param from_cells List with b1..b8 raw vectors for `from` cells.
#' @param to_cells List with b1..b8 raw vectors for `to` cells.
#' @param method Distance method: "haversine", "geodesic", or "rhumb".
#' @return Numeric vector of distances in metres.
#' @noRd
#' @keywords internal
a5_cell_distance_rs <- function(from_cells, to_cells, method) .Call(wrap__a5_cell_distance_rs, from_cells, to_cells, method)

#' Get the resolution of A5 cell indices.
#'
#' @param cells List with b1..b8 raw vectors.
#' @return Integer vector of resolutions.
#' @noRd
#' @keywords internal
a5_get_resolution_rs <- function(cells) .Call(wrap__a5_get_resolution_rs, cells)

#' Navigate to parent cell(s).
#'
#' @param cells List with b1..b8 raw vectors.
#' @param parent_resolution Integer: target parent resolution. NULL for
#'   immediate parent.
#' @return List with b1..b8 raw vectors.
#' @noRd
#' @keywords internal
a5_cell_to_parent_rs <- function(cells, parent_resolution) .Call(wrap__a5_cell_to_parent_rs, cells, parent_resolution)

#' Get child cells.
#'
#' @param cell List with b1..b8 raw vectors (length 1).
#' @param child_resolution Integer: target child resolution. NULL for
#'   immediate children.
#' @return List with b1..b8 raw vectors.
#' @noRd
#' @keywords internal
a5_cell_to_children_rs <- function(cell, child_resolution) .Call(wrap__a5_cell_to_children_rs, cell, child_resolution)

#' Get all 12 resolution-0 root cells.
#'
#' @return List with b1..b8 raw vectors.
#' @noRd
#' @keywords internal
a5_get_res0_cells_rs <- function() .Call(wrap__a5_get_res0_cells_rs)

#' Compact a set of A5 cell IDs.
#'
#' @param cells List with b1..b8 raw vectors.
#' @return List with b1..b8 raw vectors.
#' @noRd
#' @keywords internal
a5_compact_rs <- function(cells) .Call(wrap__a5_compact_rs, cells)

#' Uncompact a set of A5 cell IDs to a target resolution.
#'
#' @param cells List with b1..b8 raw vectors.
#' @param target_resolution Integer: the resolution to expand to.
#' @return List with b1..b8 raw vectors.
#' @noRd
#' @keywords internal
a5_uncompact_rs <- function(cells, target_resolution) .Call(wrap__a5_uncompact_rs, cells, target_resolution)

#' Generate a grid of A5 cells covering a bounding box.
#'
#' @param xmin,ymin,xmax,ymax Bounding box coordinates.
#' @param resolution Target resolution (0--30).
#' @return List with b1..b8 raw vectors.
#' @noRd
#' @keywords internal
a5_grid_bbox_rs <- function(xmin, ymin, xmax, ymax, resolution) .Call(wrap__a5_grid_bbox_rs, xmin, ymin, xmax, ymax, resolution)

#' Filter cell IDs to those whose boundary polygons intersect a target geometry.
#'
#' @param cells List with b1..b8 raw vectors.
#' @param target_wkt WKT string of the target geometry.
#' @return List with b1..b8 raw vectors (filtered).
#' @noRd
#' @keywords internal
a5_grid_intersects_rs <- function(cells, target_wkt) .Call(wrap__a5_grid_intersects_rs, cells, target_wkt)

#' Get all cells within k hops of a centre cell.
#'
#' @param cell List with b1..b8 raw vectors (length 1).
#' @param k Number of hops.
#' @param vertex If TRUE, include vertex-sharing (8-connected) neighbours.
#' @return List with b1..b8 raw vectors.
#' @noRd
#' @keywords internal
a5_grid_disk_rs <- function(cell, k, vertex) .Call(wrap__a5_grid_disk_rs, cell, k, vertex)

#' Get all cells within a great-circle radius of a centre cell.
#'
#' @param cell List with b1..b8 raw vectors (length 1).
#' @param radius Great-circle radius in metres.
#' @return List with b1..b8 raw vectors.
#' @noRd
#' @keywords internal
a5_spherical_cap_rs <- function(cell, radius) .Call(wrap__a5_spherical_cap_rs, cell, radius)

#' Convert one or more polygon parts (with optional holes) to A5 cells.
#'
#' `lon`, `lat` are flat coordinate vectors; `offsets` is cumulative
#' (length `n_rings + 1`) so ring `i` is `lon[offsets[i]..offsets[i+1]]`.
#' `part_id` (length `n_rings`) groups rings by polygon part. `is_outer`
#' (length `n_rings`, 1 = outer / 0 = hole) classifies each ring.
#'
#' For each polygon part, the outer ring is converted to cells, the
#' uncompacted cells inside any hole rings are subtracted, and the
#' remaining cells are unioned across parts. The result is recompacted
#' at the end. A single-outer-no-holes input takes a fast path that
#' passes the upstream compacted result through unchanged.
#'
#' @noRd
#' @keywords internal
a5_polygon_to_cells_rs <- function(lon, lat, offsets, part_id, is_outer, resolution) .Call(wrap__a5_polygon_to_cells_rs, lon, lat, offsets, part_id, is_outer, resolution)

#' Convert one or more linestring waypoint sequences to A5 cells.
#'
#' Per-feature `line_string_to_cells` results are concatenated in feature
#' order with first-seen deduplication, preserving discovery order along
#' the path.
#'
#' @noRd
#' @keywords internal
a5_linestring_to_cells_rs <- function(lon, lat, offsets, resolution) .Call(wrap__a5_linestring_to_cells_rs, lon, lat, offsets, resolution)

#' Cell area at a given resolution
#'
#' Returns the area of a single cell in square metres at the given
#' resolution(s). Because A5 is an equal-area DGGS, all cells at the same
#' resolution have identical area.
#'
#' @param resolution Integer vector of resolutions (0--30).
#' @param units Character scalar specifying the output area unit (default
#'   `"m^2"`). Any unit convertible from `m^2` via [units::set_units()] is
#'   accepted (e.g. `"km^2"`, `"ha"`, `"acre"`).  If NULL, the area is
#'   returned as a numeric vector in m^2.
#' @returns A [units::units] vector of areas.
#'
#' @export
#' @examples
#' a5_cell_area(0:5)
#' a5_cell_area(5, units = "km^2")
a5_cell_area <- function(resolution, units = "m^2") {
  resolution <- vctrs::vec_cast(resolution, integer())
  check_resolution(resolution)
  if (!is.null(units) && !units::ud_are_convertible("m^2", units)) {
    cli::cli_abort(
      "{.arg units} must be an area unit convertible from m^2 (or NULL), not {.val {units}}."
    )
  }
  r <- a5_cell_area_rs(resolution)

  if (is.null(units)) {
    return(r)
  }

  units::set_units(r, "m^2", mode = "standard") |>
    units::set_units(units, mode = "standard")
}

#' Total number of cells at a given resolution
#'
#' @param resolution Integer scalar resolution (0--30).
#' @returns A numeric scalar (double) giving the total count. Returned as
#'   double because the count can exceed R's integer range.
#'
#' @export
#' @examples
#' a5_get_num_cells(0)
#' a5_get_num_cells(10)
a5_get_num_cells <- function(resolution) {
  resolution <- vctrs::vec_cast(resolution, integer())
  check_resolution(resolution)
  vctrs::vec_assert(resolution, size = 1L)
  a5_get_num_cells_rs(resolution)
}

#' Number of children between two resolutions
#'
#' Returns the number of child cells each parent cell contains when
#' expanding from one resolution to another.
#'
#' @param parent_resolution Integer scalar (0--30).
#' @param child_resolution Integer scalar (0--30), must be >=
#'   `parent_resolution`.
#' @returns A numeric scalar. Returned as double because the count can
#'   exceed R's integer range at large resolution deltas.
#'
#' @seealso [a5_get_num_cells()], [a5_cell_to_children()],
#'   [a5_uncompact()]
#' @export
#' @examples
#' a5_get_num_children(5, 8)   # 4^3 = 64
#' a5_get_num_children(0, 5)
a5_get_num_children <- function(parent_resolution, child_resolution) {
  parent_resolution <- vctrs::vec_cast(parent_resolution, integer())
  child_resolution <- vctrs::vec_cast(child_resolution, integer())
  check_resolution(parent_resolution)
  check_resolution(child_resolution)
  vctrs::vec_assert(parent_resolution, size = 1L)
  vctrs::vec_assert(child_resolution, size = 1L)
  a5_get_num_children_rs(parent_resolution, child_resolution)
}

#' Get all resolution-0 root cells
#'
#' Returns the 12 root cells corresponding to the 12 faces of the
#' dodecahedron.
#'
#' @returns An [a5_cell] vector of length 12.
#'
#' @export
#' @examples
#' a5_get_res0_cells()
a5_get_res0_cells <- function() {
  cells_from_rs(a5_get_res0_cells_rs())
}

# -- a5_get_num_children ------------------------------------------------------

test_that("get_num_children returns correct count for aperture 4", {
  # Each resolution step is aperture 4, so 3 steps = 4^3 = 64
  expect_equal(a5_get_num_children(5, 8), 64)
})

test_that("get_num_children same resolution returns 1", {
  expect_equal(a5_get_num_children(5, 5), 1)
})

test_that("get_num_children child < parent returns 0", {
  expect_equal(a5_get_num_children(8, 5), 0)
})

test_that("get_num_children from resolution 0", {
  # res 0 has 12 cells, res 1 has 60 cells -> 5 children per res-0 cell
  expect_equal(a5_get_num_children(0, 1), 5)
})

test_that("get_num_children is consistent with get_num_cells", {
  # num_cells(child) / num_cells(parent) should equal num_children
  parent <- 3L
  child <- 7L
  ratio <- a5_get_num_cells(child) / a5_get_num_cells(parent)
  expect_equal(a5_get_num_children(parent, child), ratio)
})

# -- a5_cell_area --------------------------------------------------------------

test_that("cell_area with units = NULL returns plain numeric in m^2", {
  a_null <- a5_cell_area(5, units = NULL)
  a_m2 <- a5_cell_area(5, units = "m^2")
  expect_type(a_null, "double")
  expect_false(inherits(a_null, "units"))
  expect_equal(a_null, as.numeric(a_m2))
})

test_that("get_num_children validates resolution", {
  expect_error(a5_get_num_children(-1, 5), "resolution")
  expect_error(a5_get_num_children(5, 31), "resolution")
})

#' Cells within k hops of a cell
#'
#' Returns all cells reachable within `k` edge hops of a centre cell,
#' including the centre cell itself.
#'
#' @param cell A single [a5_cell] value.
#' @param k Integer scalar, number of hops.
#' @param vertex Logical scalar. If `FALSE` (default), only edge-sharing
#'   neighbours (4-connected) are traversed. If `TRUE`, vertex-sharing
#'   neighbours are included (8-connected).
#' @returns A compacted [a5_cell] vector.
#'
#' @seealso [a5_spherical_cap()] for distance-based selection.
#' @export
#' @examples
#' cell <- a5_lonlat_to_cell(-3.19, 55.95, resolution = 8)
#' a5_grid_disk(cell, k = 1)
a5_grid_disk <- function(cell, k, vertex = FALSE) {
  cell <- as_a5_cell(cell)
  vctrs::vec_assert(cell, size = 1L)
  k <- vctrs::vec_cast(k, integer())
  vctrs::vec_assert(k, size = 1L)
  cells_from_rs(a5_grid_disk_rs(cell_data(cell), k, vertex))
}

#' Cells within a great-circle radius
#'
#' Returns all cells whose centres fall within a great-circle distance
#' of a given cell's centre.
#'
#' @param cell A single [a5_cell] value.
#' @param radius Numeric scalar, great-circle radius in metres.
#' @returns A compacted [a5_cell] vector.
#'
#' @seealso [a5_grid_disk()] for hop-based selection.
#' @export
#' @examples
#' cell <- a5_lonlat_to_cell(-3.19, 55.95, resolution = 8)
#' a5_spherical_cap(cell, radius = 1000)
a5_spherical_cap <- function(cell, radius) {
  cell <- as_a5_cell(cell)
  vctrs::vec_assert(cell, size = 1L)
  radius <- vctrs::vec_cast(radius, double())
  vctrs::vec_assert(radius, size = 1L)
  cells_from_rs(a5_spherical_cap_rs(cell_data(cell), radius))
}

cell <- a5_lonlat_to_cell(0, 0, resolution = 8)

test_that("grid_disk k=0 returns only the centre cell", {
  result <- a5_grid_disk(cell, k = 0)
  expect_length(result, 1L)
  expect_equal(format(result), format(cell))
})

test_that("grid_disk k=1 returns centre + neighbours", {
  result <- a5_grid_disk(cell, k = 1)
  expect_true(length(result) > 1L)
})

test_that("grid_disk vertex=TRUE returns more cells than vertex=FALSE", {
  edge <- a5_grid_disk(cell, k = 1, vertex = FALSE)
  vert <- a5_grid_disk(cell, k = 1, vertex = TRUE)
  expect_true(length(vert) > length(edge))
})

test_that("spherical_cap returns at least the centre cell", {
  result <- a5_spherical_cap(cell, radius = 1)
  expect_true(length(result) >= 1L)
})

test_that("spherical_cap with larger radius returns more cells", {
  small <- a5_spherical_cap(cell, radius = 100000)
  large <- a5_spherical_cap(cell, radius = 500000)
  expect_true(length(large) > length(small))
})

test_that("grid_disk errors on invalid cell", {
  expect_error(a5_grid_disk(a5_cell("zzzz"), k = 1))
})

test_that("spherical_cap errors on invalid cell", {
  expect_error(a5_spherical_cap(a5_cell("zzzz"), radius = 100))
})

test_that("grid_disk errors on NA cell", {
  expect_error(a5_grid_disk(a5_cell(NA), k = 1))
})

test_that("spherical_cap errors on NA cell", {
  expect_error(a5_spherical_cap(a5_cell(NA), radius = 100))
})

#' A5 Cell Index Vector
#'
#' Create, test, and coerce A5 cell index vectors. Cells are stored as
#' a record with eight raw-byte fields (`b1`--`b8`) representing the
#' little-endian bytes of the u64 cell ID. This avoids the precision
#' loss of floating-point storage and keeps memory compact.
#'
#' @param x A character vector of hex-encoded A5 cell IDs, or an object
#'   coercible to one.
#' @returns An `a5_cell` vector (`a5_cell`, `as_a5_cell`), a logical
#'   scalar (`is_a5_cell`), or a logical vector (`a5_is_valid`).
#'
#' @export
#' @examples
#' cells <- a5_cell(c("0800000000000006", "0800000000000016"))
#' cells
a5_cell <- function(x = character()) {
  x <- vctrs::vec_cast(x, character())
  check_hex16(x)
  rs <- hex_to_raw8_rs(x)
  cells_from_rs(rs)
}

new_a5_cell <- function(
  b1 = raw(),
  b2 = raw(),
  b3 = raw(),
  b4 = raw(),
  b5 = raw(),
  b6 = raw(),
  b7 = raw(),
  b8 = raw()
) {
  vctrs::new_rcrd(
    list(
      b1 = b1,
      b2 = b2,
      b3 = b3,
      b4 = b4,
      b5 = b5,
      b6 = b6,
      b7 = b7,
      b8 = b8
    ),
    class = "a5_cell"
  )
}

#' Construct an a5_cell from Rust list(b1=, ..., b8=) output
#' @noRd
cells_from_rs <- function(x) {
  new_a5_cell(
    b1 = x$b1,
    b2 = x$b2,
    b3 = x$b3,
    b4 = x$b4,
    b5 = x$b5,
    b6 = x$b6,
    b7 = x$b7,
    b8 = x$b8
  )
}

#' Check that all non-NA hex strings are exactly 16 characters
#' @noRd
check_hex16 <- function(x) {
  non_na <- !is.na(x)
  bad <- non_na & nchar(x) != 16L
  if (any(bad)) {
    first <- x[which(bad)[1L]]
    cli::cli_abort(
      "Cell hex strings must be exactly 16 characters, not {nchar(first)}. {.val {first}} is invalid."
    )
  }
}

#' Pass cell fields to Rust as a named list
#' @noRd
cell_data <- function(x) {
  vctrs::vec_data(x)
}

#' @export
#' @rdname a5_cell
is_a5_cell <- function(x) {
  inherits(x, "a5_cell")
}

#' @export
#' @rdname a5_cell
as_a5_cell <- function(x) {
  if (is_a5_cell(x)) {
    return(x)
  }
  a5_cell(x)
}

#' @export
#' @rdname a5_cell
#' @examples
#' a5_is_valid(c("0800000000000006", "not_a_cell", NA))
a5_is_valid <- function(x) {
  if (is_a5_cell(x)) {
    a5_is_valid_cell_rs(cell_data(x))
  } else {
    x <- vctrs::vec_cast(x, character())
    a5_is_valid_hex_rs(x)
  }
}

#' Coerce between hex strings and A5 cell vectors
#'
#' `a5_u64_to_hex()` converts an [a5_cell] vector to 16-character
#' zero-padded hex strings. `a5_hex_to_u64()` converts hex strings to
#' an [a5_cell] vector.
#'
#' @param x For `a5_u64_to_hex()`, an [a5_cell] vector (or object
#'   coercible to one). For `a5_hex_to_u64()`, a character vector of
#'   hex-encoded cell IDs.
#' @returns `a5_u64_to_hex()` returns a character vector. `a5_hex_to_u64()`
#'   returns an [a5_cell] vector.
#'
#' @details
#' These are named to match `u64_to_hex` / `hex_to_u64` in the upstream
#' Python, JavaScript, and DuckDB A5 bindings. In those languages the
#' functions convert between a native 64-bit unsigned integer and its hex
#' representation. Because R has no native `uint64` type, `a5_u64_to_hex()`
#' accepts an [a5_cell] (which stores the `u64` internally as eight raw
#' bytes) instead of a bare integer.
#'
#' @seealso [a5_cell_from_arrow()] and [a5_cell_to_arrow()] for lossless
#'   conversion between [a5_cell] and Arrow `uint64` arrays.
#'
#' @export
#' @examples
#' cell <- a5_lonlat_to_cell(-3.19, 55.95, resolution = 5)
#' hex <- a5_u64_to_hex(cell)
#' hex
#'
#' a5_hex_to_u64(hex)
a5_u64_to_hex <- function(x) {
  x <- as_a5_cell(x)
  raw8_to_hex_rs(cell_data(x))
}

#' @rdname a5_u64_to_hex
#' @export
a5_hex_to_u64 <- function(x) {
  a5_cell(x)
}

# --- vctrs methods ---

#' @exportS3Method vctrs::vec_ptype_abbr
#' @noRd
#' @keywords internal
vec_ptype_abbr.a5_cell <- function(x, ...) "a5_cell"

#' @exportS3Method vctrs::vec_ptype_full
#' @noRd
#' @keywords internal
vec_ptype_full.a5_cell <- function(x, ...) "a5_cell"

#' @export
#' @noRd
#' @keywords internal
format.a5_cell <- function(x, ...) {
  raw8_to_hex_rs(cell_data(x))
}

#' @export
#' @noRd
#' @keywords internal
is.na.a5_cell <- function(x) {
  # NA sentinel: last byte (b8) == 0xFC
  vctrs::field(x, "b8") == as.raw(0xFC)
}

# --- ordering: big-endian byte order = u64 numeric order ---

#' @exportS3Method vctrs::vec_proxy_compare
#' @noRd
#' @keywords internal
vec_proxy_compare.a5_cell <- function(x, ...) {
  data.frame(
    b8 = as.integer(vctrs::field(x, "b8")),
    b7 = as.integer(vctrs::field(x, "b7")),
    b6 = as.integer(vctrs::field(x, "b6")),
    b5 = as.integer(vctrs::field(x, "b5")),
    b4 = as.integer(vctrs::field(x, "b4")),
    b3 = as.integer(vctrs::field(x, "b3")),
    b2 = as.integer(vctrs::field(x, "b2")),
    b1 = as.integer(vctrs::field(x, "b1"))
  )
}

#' @exportS3Method vctrs::vec_proxy_order
#' @noRd
#' @keywords internal
vec_proxy_order.a5_cell <- function(x, ...) {
  vec_proxy_compare.a5_cell(x, ...)
}

# --- coercion: a5_cell <-> character ---

#' @export
#' @noRd
#' @keywords internal
vec_ptype2.a5_cell.a5_cell <- function(x, y, ...) new_a5_cell()

#' @export
#' @noRd
#' @keywords internal
vec_ptype2.a5_cell.character <- function(x, y, ...) new_a5_cell()

#' @export
#' @noRd
#' @keywords internal
vec_ptype2.character.a5_cell <- function(x, y, ...) new_a5_cell()

#' @export
#' @noRd
#' @keywords internal
vec_cast.a5_cell.a5_cell <- function(x, to, ...) x

#' @export
#' @noRd
#' @keywords internal
vec_cast.a5_cell.character <- function(x, to, ...) {
  check_hex16(x)
  rs <- hex_to_raw8_rs(x)
  cells_from_rs(rs)
}

#' @export
#' @noRd
#' @keywords internal
vec_cast.character.a5_cell <- function(x, to, ...) {
  raw8_to_hex_rs(cell_data(x))
}

# --- pillar formatting for tibbles ---

#' @exportS3Method pillar::pillar_shaft
#' @noRd
#' @keywords internal
pillar_shaft.a5_cell <- function(x, ...) {
  out <- format(x)
  pillar::new_pillar_shaft_simple(out, align = "left")
}

test_that("a5_cell constructor", {
  cell <- a5_cell("0800000000000006")
  expect_s3_class(cell, "a5_cell")
  expect_s3_class(cell, "vctrs_rcrd")
  expect_true(is.raw(vctrs::field(cell, "b1")))
  expect_true(is.raw(vctrs::field(cell, "b8")))
})

test_that("a5_cell coercion with character", {
  cell <- a5_cell("0800000000000006")
  combined <- vctrs::vec_c(cell, "0800000000000016")
  expect_s3_class(combined, "a5_cell")
  expect_length(combined, 2L)
})

test_that("is_a5_cell works", {
  expect_true(is_a5_cell(a5_cell("0000000000000abc")))
  expect_false(is_a5_cell("0000000000000abc"))
})

test_that("a5_cell rejects hex strings not exactly 16 characters", {
  expect_error(a5_cell("12346"), "exactly 16 characters")
  expect_error(a5_cell("0800000000000006f"), "exactly 16 characters")
  expect_error(a5_cell(c("0800000000000006", "abc")), "exactly 16 characters")
  # NA and valid 16-char hex pass

  expect_length(a5_cell(c("0800000000000006", NA)), 2L)
  # vec_cast path also validates
  expect_error(vctrs::vec_cast("short", a5_cell()), "exactly 16 characters")
})

test_that("a5_is_valid validates", {
  result <- a5_is_valid(c("0800000000000006", "not_valid", NA))
  expect_equal(result[1], TRUE)
  expect_equal(result[2], FALSE)
  expect_true(is.na(result[3]))
})

test_that("a5_is_valid works with a5_cell input", {
  cell <- a5_cell("0800000000000006")
  expect_true(a5_is_valid(cell))
})

test_that("as_a5_cell passes through a5_cell unchanged", {
  cell <- a5_cell("0800000000000006")
  expect_identical(as_a5_cell(cell), cell)
})

test_that("as_a5_cell coerces character", {
  cell <- as_a5_cell("0800000000000006")
  expect_s3_class(cell, "a5_cell")
})

test_that("a5_cell handles NA values", {
  cell <- a5_cell(c("0800000000000006", NA))
  expect_length(cell, 2L)
  # NA is represented by the sentinel value (b8 == 0xFC)
  expect_true(is.na(cell[2]))
})

test_that("format.a5_cell preserves hex strings", {
  cell <- a5_cell(c("0800000000000006", NA))
  formatted <- format(cell)
  expect_equal(formatted[1], "0800000000000006")
  expect_true(is.na(formatted[2]))
})

test_that("a5_u64_to_hex works", {
  cell <- a5_cell("0800000000000006")
  expect_equal(a5_u64_to_hex(cell), "0800000000000006")
})

test_that("a5_hex_to_u64 works", {
  cell <- a5_hex_to_u64("0800000000000006")
  expect_s3_class(cell, "a5_cell")
  expect_equal(format(cell), "0800000000000006")
})

test_that("vec_cast round-trips a5_cell <-> character", {
  cell <- a5_cell("0800000000000006")
  chr <- vctrs::vec_cast(cell, character())
  expect_equal(chr, "0800000000000006")
  back <- vctrs::vec_cast(chr, a5_cell())
  expect_s3_class(back, "a5_cell")
  expect_equal(format(back), "0800000000000006")
})

test_that("vec_c combines a5_cell + a5_cell", {
  a <- a5_cell("0800000000000006")
  b <- a5_cell("0800000000000016")
  combined <- vctrs::vec_c(a, b)
  expect_s3_class(combined, "a5_cell")
  expect_length(combined, 2L)
  expect_equal(
    format(combined),
    c("0800000000000006", "0800000000000016")
  )
})

test_that("vec_c combines character + a5_cell (character first)", {
  cell <- a5_cell("0800000000000016")
  combined <- vctrs::vec_c("0800000000000006", cell)
  expect_s3_class(combined, "a5_cell")
  expect_length(combined, 2L)
})

test_that("vec_ptype_abbr and vec_ptype_full dispatch correctly", {
  cell <- a5_cell("0800000000000006")
  expect_equal(vec_ptype_abbr.a5_cell(cell), "a5_cell")
  expect_equal(vec_ptype_full.a5_cell(cell), "a5_cell")
})

test_that("pillar_shaft.a5_cell formats correctly", {
  skip_if_not_installed("pillar")
  cell <- a5_cell(c("0800000000000006", NA))
  shaft <- pillar_shaft.a5_cell(cell)
  expect_s3_class(shaft, "pillar_shaft")
})

test_that("a5_cell displays correctly in tibbles", {
  skip_if_not_installed("tibble")
  skip_if_not_installed("pillar")
  tbl <- tibble::tibble(
    cell = a5_cell(c("0800000000000006", "0800000000000016"))
  )

  testthat::expect_snapshot(tbl)

  out <- format(tbl)
  expect_true(any(grepl("a5_cell", out)))
  expect_true(any(grepl("0800000000000006", out)))
})

test_that("str works on a5_cell", {
  cell <- a5_cell(c("0800000000000006", "0800000000000016"))
  out <- capture.output(str(cell))
  expect_true(any(grepl("a5_cell", out)))
})

test_that("wk_crs.a5_cell returns longlat CRS", {
  cell <- a5_lonlat_to_cell(0, 0, resolution = 5)
  expect_identical(wk::wk_crs(cell), wk::wk_crs_longlat())
})

test_that("wk_handle.a5_cell produces boundary geometry", {
  cell <- a5_lonlat_to_cell(0, 0, resolution = 5)
  wkt <- wk::wk_handle(cell, wk::wkt_writer())
  expect_s3_class(wkt, "wk_wkt")
  expect_true(grepl("^POLYGON", as.character(wkt)))
})

# -- sentinel / NA byte-level tests -------------------------------------------

test_that("NA sentinel has b8 == 0xFC", {
  cell <- a5_cell(c("0800000000000006", NA))
  # Non-NA cell: b8 should NOT be 0xFC
  expect_false(vctrs::field(cell, "b8")[1] == as.raw(0xFC))
  # NA cell: b8 IS the sentinel byte

  expect_equal(vctrs::field(cell, "b8")[2], as.raw(0xFC))
})

test_that("sentinel hex value fc00000000000000 becomes NA", {
  cell <- a5_cell("fc00000000000000")
  expect_true(is.na(cell))
  expect_true(is.na(format(cell)))
})

test_that("known hex encodes to correct little-endian bytes", {
  cell <- a5_cell("0800000000000006")
  # 0x0800000000000006 in little-endian: 06 00 00 00 00 00 00 08
  expect_equal(vctrs::field(cell, "b1"), as.raw(0x06))
  expect_equal(vctrs::field(cell, "b2"), as.raw(0x00))
  expect_equal(vctrs::field(cell, "b7"), as.raw(0x00))
  expect_equal(vctrs::field(cell, "b8"), as.raw(0x08))
})

# -- empty / all-NA edge cases ------------------------------------------------

test_that("empty a5_cell works", {
  cell <- a5_cell(character())
  expect_length(cell, 0L)
  expect_s3_class(cell, "a5_cell")
  expect_equal(format(cell), character())
})

test_that("all-NA a5_cell works", {
  cell <- a5_cell(c(NA, NA, NA))
  expect_length(cell, 3L)
  expect_true(all(is.na(cell)))
})

test_that("single NA a5_cell works", {
  cell <- a5_cell(NA)
  expect_length(cell, 1L)
  expect_true(is.na(cell))
})

# -- vctrs operations: unique, duplicated, match ------------------------------

test_that("unique works on a5_cell", {
  cells <- a5_cell(c(
    "0800000000000006",
    "0800000000000016",
    "0800000000000006"
  ))
  u <- unique(cells)
  expect_length(u, 2L)
  expect_equal(sort(format(u)), c("0800000000000006", "0800000000000016"))
})

test_that("duplicated works on a5_cell", {
  cells <- a5_cell(c(
    "0800000000000006",
    "0800000000000016",
    "0800000000000006"
  ))
  expect_equal(duplicated(cells), c(FALSE, FALSE, TRUE))
})

test_that("match works on a5_cell", {
  haystack <- a5_cell(c("0800000000000006", "0800000000000016"))
  needle <- a5_cell("0800000000000016")
  expect_equal(match(needle, haystack), 2L)
})

test_that("%in% works on a5_cell", {
  cells <- a5_cell(c(
    "0800000000000006",
    "0800000000000016",
    "0800000000000026"
  ))
  subset <- a5_cell(c("0800000000000006", "0800000000000026"))
  expect_equal(cells %in% subset, c(TRUE, FALSE, TRUE))
})

test_that("unique handles NA correctly", {
  cells <- a5_cell(c("0800000000000006", NA, NA))
  u <- unique(cells)
  expect_length(u, 2L)
  expect_equal(sum(is.na(u)), 1L)
})

# -- ordering -----------------------------------------------------------------

test_that("sort orders by u64 value", {
  # Create cells with known ordering: lower hex = lower u64
  cells <- a5_cell(c(
    "0800000000000016",
    "0800000000000006",
    "0800000000000026"
  ))
  sorted <- sort(cells)
  expect_equal(
    format(sorted),
    c("0800000000000006", "0800000000000016", "0800000000000026")
  )
})

test_that("sort handles NA (pushed to end)", {
  cells <- a5_cell(c(NA, "0800000000000016", "0800000000000006"))
  sorted <- sort(cells, na.last = TRUE)
  expect_equal(format(sorted)[1:2], c("0800000000000006", "0800000000000016"))
  expect_true(is.na(sorted[3]))
})

test_that("order returns correct indices", {
  cells <- a5_cell(c(
    "0800000000000016",
    "0800000000000006",
    "0800000000000026"
  ))
  expect_equal(order(cells), c(2L, 1L, 3L))
})

test_that("comparison operators work", {
  a <- a5_cell("0800000000000006")
  b <- a5_cell("0800000000000016")
  expect_true(a < b)
  expect_false(a > b)
  expect_true(a == a)
  expect_true(a != b)
})

# -- memory -------------------------------------------------------------------

test_that("a5_cell memory is compact (8 raw vectors)", {
  cells <- a5_lonlat_to_cell(runif(1000, -180, 180), runif(1000, -80, 80), 10L)
  # Should be ~8 KB (8 x 1 byte x 1000), not ~56 KB+ from list-of-raw
  sz <- as.numeric(object.size(cells))
  expect_true(sz < 50000) # well under list-of-raw overhead
})

#' Get the resolution of A5 cell indices
#'
#' Extracts the resolution level (0--30) encoded in each cell index.
#'
#' @param cell An [a5_cell] vector.
#' @returns An integer vector of resolutions.
#'
#' @seealso [a5_cell_to_parent()], [a5_cell_to_children()]
#' @export
#' @examples
#' cell <- a5_lonlat_to_cell(-3.19, 55.95, resolution = 10)
#' a5_get_resolution(cell)
a5_get_resolution <- function(cell) {
  cell <- as_a5_cell(cell)
  a5_get_resolution_rs(cell_data(cell))
}

#' Navigate to parent cell(s)
#'
#' Returns the parent cell of each input cell. By default returns the
#' immediate parent (one resolution coarser). Optionally target a specific
#' coarser resolution.
#'
#' @param cell An [a5_cell] vector.
#' @param resolution Integer scalar target parent resolution, or `NULL` for
#'   the immediate parent.
#' @returns An [a5_cell] vector of parent cells.
#'
#' @seealso [a5_cell_to_children()], [a5_get_resolution()]
#' @export
#' @examples
#' cell <- a5_lonlat_to_cell(-3.19, 55.95, resolution = 10)
#' a5_cell_to_parent(cell)
#' a5_cell_to_parent(cell, resolution = 5)
a5_cell_to_parent <- function(cell, resolution = NULL) {
  cell <- as_a5_cell(cell)
  if (!is.null(resolution)) {
    resolution <- vctrs::vec_cast(resolution, integer())
    check_resolution(resolution)
    vctrs::vec_assert(resolution, size = 1L)
  }
  cells_from_rs(a5_cell_to_parent_rs(cell_data(cell), resolution))
}

#' Get child cells
#'
#' Returns the child cells of a single cell. By default returns the 4
#' immediate children (one resolution finer). Optionally target a specific
#' finer resolution.
#'
#' @param cell A single [a5_cell] value.
#' @param resolution Integer scalar target child resolution, or `NULL` for
#'   immediate children.
#' @returns An [a5_cell] vector of child cells.
#'
#' @seealso [a5_cell_to_parent()], [a5_get_resolution()]
#' @export
#' @examples
#' cell <- a5_lonlat_to_cell(-3.19, 55.95, resolution = 5)
#' a5_cell_to_children(cell)
a5_cell_to_children <- function(cell, resolution = NULL) {
  cell <- as_a5_cell(cell)
  vctrs::vec_assert(cell, size = 1L)
  if (!is.null(resolution)) {
    resolution <- vctrs::vec_cast(resolution, integer())
    check_resolution(resolution)
    vctrs::vec_assert(resolution, size = 1L)
  }
  cells_from_rs(a5_cell_to_children_rs(cell_data(cell), resolution))
}

test_that("get_resolution extracts resolution", {
  cell <- a5_lonlat_to_cell(0, 0, resolution = 10)
  expect_equal(a5_get_resolution(cell), 10L)
})

test_that("cell_to_parent returns coarser cell", {
  cell <- a5_lonlat_to_cell(0, 0, resolution = 10)
  parent <- a5_cell_to_parent(cell)
  expect_equal(a5_get_resolution(parent), 9L)
})

test_that("cell_to_parent with explicit resolution", {
  cell <- a5_lonlat_to_cell(0, 0, resolution = 10)
  parent <- a5_cell_to_parent(cell, resolution = 5)
  expect_equal(a5_get_resolution(parent), 5L)
})

test_that("cell_to_children returns 4 children", {
  cell <- a5_lonlat_to_cell(0, 0, resolution = 5)
  children <- a5_cell_to_children(cell)
  expect_length(children, 4L)
  expect_true(all(a5_get_resolution(children) == 6L))
})

test_that("get_resolution is vectorised", {
  cells <- a5_lonlat_to_cell(c(0, 10), c(0, 10), resolution = c(3L, 7L))
  expect_equal(a5_get_resolution(cells), c(3L, 7L))
})

test_that("cell_to_parent is vectorised", {
  cells <- a5_lonlat_to_cell(c(0, 10), c(0, 10), resolution = 10)
  parents <- a5_cell_to_parent(cells)
  expect_length(parents, 2L)
  expect_true(all(a5_get_resolution(parents) == 9L))
})

test_that("cell_to_children with explicit resolution", {
  cell <- a5_lonlat_to_cell(0, 0, resolution = 5)
  children <- a5_cell_to_children(cell, resolution = 7)
  expect_length(children, 16L)
  expect_true(all(a5_get_resolution(children) == 7L))
})

test_that("res0 cells", {
  cells <- a5_get_res0_cells()
  expect_length(cells, 12L)
  expect_true(all(a5_get_resolution(cells) == 0L))
})

test_that("get_num_cells returns positive double", {
  n <- a5_get_num_cells(0)
  expect_type(n, "double")
  expect_equal(n, 12)
})

test_that("get_num_cells increases with resolution", {
  n5 <- a5_get_num_cells(5)
  n10 <- a5_get_num_cells(10)
  expect_true(n10 > n5)
})

# -- NA propagation -----------------------------------------------------------

test_that("get_resolution propagates NA", {
  cells <- a5_cell(c("0800000000000006", NA))
  res <- a5_get_resolution(cells)
  expect_false(is.na(res[1]))
  expect_true(is.na(res[2]))
})

test_that("cell_to_parent propagates NA", {
  cells <- a5_cell(c("0800000000000006", NA))
  parents <- a5_cell_to_parent(cells)
  expect_false(is.na(parents[1]))
  expect_true(is.na(parents[2]))
})

#' wk methods for a5_cell
#'
#' Integration with the [wk][wk::wk-package] geometry framework. Allows
#' `a5_cell` vectors to be handled as geometry (via their boundary polygons)
#' and to report their CRS.
#'
#' @param handleable,x An [a5_cell] vector.
#' @param handler A [wk handler][wk::wk_handle].
#' @param ... Passed to underlying methods.
#' @returns
#' - `wk_handle()`: the result of the handler.
#' - `wk_crs()`: a [wk::wk_crs] object (WGS 84 lon/lat).
#' @name wk_methods
NULL

#' @exportS3Method wk::wk_handle
#' @rdname wk_methods
wk_handle.a5_cell <- function(handleable, handler, ...) {
  geom <- a5_cell_to_boundary(handleable, format = "wkb")
  wk::wk_handle(geom, handler, ...)
}

#' @exportS3Method wk::wk_crs
#' @rdname wk_methods
wk_crs.a5_cell <- function(x) {
  wk::wk_crs_longlat()
}

#' Distance between cell centroids
#'
#' Computes the distance between the centroids of pairs of A5 cells
#' using the specified method.
#'
#' @param from,to [a5_cell] vectors (recycled to common length).
#' @param units Character scalar specifying the distance unit (default
#'   `"m"`). Any unit convertible from metres via [units::set_units()] is
#'   accepted (e.g. `"km"`, `"mi"`). If NULL, the distance is returned as a
#'   numeric vector in metres.
#' @param method Distance calculation method. One of `"haversine"`
#'   (great-circle, default), `"geodesic"` (WGS84 ellipsoid via Karney
#'   2013), or `"rhumb"` (loxodrome / constant-bearing).
#' @returns A [units::units] vector of distances.
#'
#' @seealso [a5_cell_to_lonlat()] for cell centroids,
#'   [a5_cell_area()] for cell areas.
#' @export
#' @examples
#' a <- a5_lonlat_to_cell(-3.19, 55.95, resolution = 24)
#' b <- a5_lonlat_to_cell(-3.10, 55.90, resolution = 24)
#' a5_cell_distance(a, b)
#' a5_cell_distance(a, b, units = "km")
#' a5_cell_distance(a, b, method = "geodesic")
a5_cell_distance <- function(
  from,
  to,
  units = "m",
  method = c("haversine", "geodesic", "rhumb")
) {
  from <- as_a5_cell(from)
  to <- as_a5_cell(to)
  args <- vctrs::vec_recycle_common(from = from, to = to)
  method <- rlang::arg_match(method)
  if (!is.null(units) && !units::ud_are_convertible("m", units)) {
    cli::cli_abort(
      "{.arg units} must be a distance unit convertible from m (or NULL), not {.val {units}}."
    )
  }
  d <- a5_cell_distance_rs(cell_data(args$from), cell_data(args$to), method)

  if (is.null(units)) {
    return(d)
  }

  units::set_units(d, "m", mode = "standard") |>
    units::set_units(units, mode = "standard")
}

# Silence routine lifecycle deprecation warnings emitted by deprecated
# functions exercised by the test suite. testthat's `local_reproducible_output()`
# resets `lifecycle_verbosity` to "warning" at the start of every `test_that()`,
# so a global option set here has no effect inside tests. Use the helper below
# at the top of test bodies that call deprecated functions.
#
# Explicit deprecation tests use `lifecycle::expect_deprecated()`, which forces
# the warning regardless.
local_quiet_lifecycle <- function(.env = parent.frame()) {
  withr::local_options(lifecycle_verbosity = "quiet", .local_envir = .env)
}

---
title: "Traversal and distance"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Traversal and distance}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(a5R)
```

## Selecting cells around a point

Pick an origin cell and use `a5_spherical_cap()` to grab every cell whose
centre falls within a given great-circle radius. The result is compacted, so
pass it through `a5_uncompact()` for a uniform-resolution grid.

```{r}
origin <- a5_lonlat_to_cell(35, 55, resolution = 14)

disk <- a5_grid_disk(origin, k = 20, vertex = FALSE)
disk_vertex <- a5_grid_disk(origin, k = 20, vertex = TRUE)
cap_compact <- a5_spherical_cap(origin, radius = 5000)
cap <- a5_uncompact(cap_compact, resolution = 14)
```

`a5_grid_disk()` selects cells by hop count rather than metric distance.
With `vertex = FALSE` (default) only edge-sharing neighbours are traversed
(4-connected), producing a diamond shape. Setting `vertex = TRUE` includes
corner-sharing neighbours too (8-connected), giving a squarer footprint.

## Distance heatmap

Compute the distance from the origin to every cell in the cap (or disk) and 
visualise it as a colour gradient.


```{r heatmap, fig.width = 10, fig.height = 10}
dcap <- a5_cell_distance(origin, cap, units = "km")
dcap_compact <- a5_cell_distance(origin, cap_compact, units = "km")
ddisk <- a5_cell_distance(origin, disk, units = "km")
ddisk_vertex <- a5_cell_distance(origin, disk_vertex, units = "km")

pal <- hcl.colors(256, "Inferno")

oldpar <- par(mfrow = c(2, 2), mar = c(2, 2, 2, 1))
for (info in list(
  list(s = disk, d = ddisk, lab = "Grid disk (edges)"),
  list(s = disk_vertex, d = ddisk_vertex, lab = "Grid disk (vertices)"),
  list(s = cap_compact, d = dcap_compact, lab = "Spherical cap (compact)"),
  list(s = cap, d = dcap, lab = "Spherical cap")
)) {
  brk <- seq(0, max(as.numeric(info$d)), length.out = 257)
  cols <- pal[findInterval(as.numeric(info$d), brk, all.inside = TRUE)]

  plot(a5_cell_to_boundary(info$s), col = cols, border = NA, asp = 1,
       main = info$lab)
  plot(a5_cell_to_boundary(origin), border = "#ffffffff", add = TRUE)
}
```

```{r, include = FALSE}
par(oldpar)
```

Distance radiates smoothly from the centre. Cells near the origin are
dark, cells at the rim are bright. The origin cell is shown with a
white border for reference.

## Distance methods

a5R supports three distance methods: **haversine** (great-circle),
**geodesic** (WGS84 ellipsoid via Karney 2013), and **rhumb** (loxodrome /
constant-bearing). For short distances they are nearly identical; differences
grow with distance.


```{r methods, fig.width = 10, fig.height = 10}
wide_origin <- a5_lonlat_to_cell(35, 55, resolution = 8)
wide_cap <- a5_uncompact(
  a5_spherical_cap(wide_origin, radius = 500000),
  resolution = 8
)

hav <- a5_cell_distance(wide_origin, wide_cap, units = "m", method = "haversine")
geo <- a5_cell_distance(wide_origin, wide_cap, units = "m", method = "geodesic")

diff_m <- as.numeric(geo - hav)
brk <- seq(min(diff_m), max(diff_m), length.out = 257)
cols <- pal[findInterval(diff_m, brk, all.inside = TRUE)]

oldpar <- par(no.readonly = TRUE)
layout(matrix(c(1, 2), nrow = 1), widths = c(4, 1))
par(mar = c(2, 2, 2, 1))
plot(a5_cell_to_boundary(wide_cap), col = cols, border = NA, asp = 1,
     main = "Geodesic \u2212 Haversine (m)")

par(mar = c(9, 0, 9, 9))
image(1, seq(min(diff_m), max(diff_m), length.out = 256),
      t(seq_along(pal)), col = pal, axes = FALSE, xlab = "", ylab = "")
axis(4, las = 1)
```

```{r, include = FALSE}
par(oldpar)
```

The difference is near zero at the centre and grows to ~`r round(max(diff_m))` m
at the rim (~500 km away).

---
title: "Multi-threading"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Multi-threading}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(a5R)
```

a5R can parallelise vectorised operations using multiple threads via
[rayon](https://docs.rs/rayon). By default a5R uses a single thread, so
there is zero overhead. You opt in to parallelism when you need it.

## Setting the thread count

```{r}
# Check the current setting (default: 1)
a5_get_threads()

# Use 4 threads
a5_set_threads(4)
a5_get_threads()
```

```{r, include = FALSE}
a5_set_threads(1)
```

You can also set threads at package load time via an R option or environment
variable, useful for scripts and batch jobs:

```r
# In .Rprofile or at the top of a script
options(a5R.threads = 4)

# Or as an environment variable
# Sys.setenv(A5R_NUM_THREADS = 4) 
```

`a5_set_threads()` invisibly returns the previous value, making temporary
changes easy:

```{r}
old <- a5_set_threads(4)
# ... parallel work ...
a5_set_threads(old)
```

## What gets parallelised

Threading applies to **vectorised** functions that process each element
independently:

| Function | Per-element cost | Benefit |
|---|---|---|
| `a5_cell_to_boundary()` | Heavy (boundary + WKT/WKB) | High |
| `a5_polygon_to_cells()` | Heavy (per-part flood-fill, hole subtraction) | High |
| `a5_linestring_to_cells()` | Heavy (per-feature tracing) | High |
| `a5_lonlat_to_cell()` | Moderate (projection) | High |
| `a5_cell_distance()` | Moderate (2x projection + distance) | Medium |
| `a5_cell_to_lonlat()` | Moderate (reverse projection) | Medium |
| `a5_cell_to_parent()` | Light (bit ops + hex) | Low |
| `a5_get_resolution()` | Light (bit ops) | Low |
| `a5_is_valid()` | Light (hex parse) | Low |

Scalar and bulk operations (`a5_cell_to_children()`, `a5_compact()`,
`a5_cell_area()`, etc.) are unaffected: they are already fast or delegate
to algorithms that don't parallelise element-wise.

## When is it worthwhile?

Threading has a small fixed overhead (thread synchronisation, memory
allocation for intermediate results). For small vectors this can outweigh
the benefit. As a rule of thumb:

- **< 1,000 elements**: stick with 1 thread
- **1,000--10,000**: 2-4 threads helps for heavy ops (boundary, indexing)
- **> 10,000**: use as many threads as you have cores

Here's a quick comparison on 100k cells:

```{r, eval = FALSE}
cells <- a5_polygon_to_cells(wk::rct(-10, 50, 10, 60), resolution = 12)
length(cells)
#> [1] 704259

a5_set_threads(1)
system.time(a5_cell_to_boundary(cells, format = "wkt"))
#>   user  system elapsed
#>  3.124   0.000   3.122 

a5_set_threads(8)
system.time(a5_cell_to_boundary(cells, format = "wkt"))
#>   user  system elapsed
#> 6.195   1.289   1.667 
```

Note that `user` time increases (total CPU work across all threads) while
`elapsed` (wall-clock) time decreases: that is the parallelism at work.

## Thread safety

a5R uses a dedicated rayon thread pool, separate from R's own
parallelism. It is safe to use alongside `future`, `mirai`, etc. but think 
carefully about this nested parallelism as it can, if overloaded, degrade 
performance.

The thread pool is rebuilt each time you call `a5_set_threads()`, so
changing the count mid-session is fine (and cheap) but not free: ideally,
do it once at the start of your workflow rather than toggling per-call.

---
title: "How a5R stores cell IDs without strings"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{How a5R stores cell IDs without strings}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## The problem

An A5 cell ID is a 64-bit unsigned integer (`u64`).
R has no native `u64` type: its integers are 32-bit signed (`-2^31` to
`2^31 - 1`), and its doubles are 64-bit floating point. A `double`
can only represent integers exactly up to 2^53, while a `u64` can go up to
2^64 - 1.

The obvious workaround is to store cell IDs as hex strings
(`"0800000000000006"`). This works, but every trip across the R / Rust
boundary requires hex parsing and formatting: O(n) string allocation that
dominates the cost of lightweight operations like `a5_get_resolution()` or
`a5_cell_to_parent()`.

## The solution: eight raw-byte fields

A `u64` is exactly 8 bytes. We store each byte of the little-endian
representation as a separate `raw` vector field in a vctrs record type:

```
cell_id (u64):  0x0800000000000006

little-endian bytes:
  b1 = 0x06, b2 = 0x00, b3 = 0x00, b4 = 0x00,
  b5 = 0x00, b6 = 0x00, b7 = 0x00, b8 = 0x08
```

This is lossless: the eight bytes are the exact same bits as the original
`u64`, just stored across eight contiguous `raw` vectors. No precision loss,
no special-case handling. On the Rust side, reconstructing the `u64` from
the eight byte slices is a single `u64::from_le_bytes()` call. This also
avoids pointers, so there is no need to think about serialization when saving
an `a5_cell` object to disk.

## R-side: a vctrs record type

On the R side, `a5_cell` is a **vctrs record** (`vctrs::new_rcrd()`) with
eight fields (`b1` through `b8`):

```{r}
library(a5R)
cell <- a5_lonlat_to_cell(-3.19, 55.95, resolution = 10)
vctrs::field(cell, "b1")
vctrs::field(cell, "b8")
```

Each field is a plain `raw` vector: a contiguous block of memory with
no per-element overhead. Subsetting, combining, and NA propagation are all
handled automatically by vctrs.

Hex strings are only produced on demand:

```{r}
# Display calls format(), which converts to hex for readability
cell

# Explicit conversion
a5_u64_to_hex(cell)

# Round-trip from hex
a5_cell("0800000000000006")
```

## Why this matters

Compare memory for one million cells:

```{r}
set.seed(42)
cells <- a5_lonlat_to_cell(
  runif(1e6, -180, 180),
  runif(1e6, -80, 80),
  resolution = 10
)

# rcrd: eight contiguous raw vectors (8 × 1 byte × 1M ≈ 7.6 MB)
format(object.size(cells), units = "MB")

# equivalent hex strings would be ~81 MB
# (16 chars + 56-byte SEXP header per string)
hex <- a5_u64_to_hex(cells)
format(object.size(hex), units = "MB")
```

## NA handling

A5 cell IDs use 60 "quintants" (values 0–59) in their top 6 bits. Quintant
63 (binary `111111`) is invalid in the A5 system, so we use
`0xFC00000000000000` as a sentinel value for `NA`. In little-endian, the
last byte (`b8`) is `0xFC`, making NA detection a fast single-byte check.

On the Rust side, the sentinel is detected and mapped to `None`. Standard
R idioms work as expected:

```{r}
cells_with_na <- a5_cell(c("0800000000000006", NA))
is.na(cells_with_na)
```

## Summary

| Aspect | Hex strings | Raw bytes |
|--------|------------|-----------|
| R type | `character` vector | `vctrs_rcrd` (eight `raw` fields) |
| Memory (1M cells) | ~81 MB | ~7.6 MB |
| R-Rust crossing | O(n) hex parse/format | Zero-copy byte access |
| Human-readable | Always | On `format()` / `print()` |
| Lossless | Yes | Yes (exact byte representation) |

---
title: "Working with Arrow and Parquet"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Working with Arrow and Parquet}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## The uint64 problem

A5 cell IDs are 64-bit unsigned integers. R has no native `uint64` type,
and its `double` can only represent integers exactly up to 2^53. Nearly
half of all A5 cell IDs exceed this threshold, so converting them to
`double` silently corrupts the data.

This is a problem when reading Parquet files that store A5 cell IDs as
`uint64` columns: the standard format used by DuckDB, Python, and
[geoparquet.io](https://geoparquet.io/). By default, `arrow::read_parquet()`
converts `uint64` to R's `double`, losing precision:

```{r naive, eval = requireNamespace("arrow", quietly = TRUE), message = FALSE}
library(arrow)
library(tibble)
library(a5R)

# A real A5 cell: Edinburgh at resolution 20
cell <- a5_lonlat_to_cell(-3.19, 55.95, resolution = 20)
a5_u64_to_hex(cell)

# Write to Parquet as uint64 (the standard interchange format)
tf <- tempfile(fileext = ".parquet")
arrow::write_parquet(
  arrow::arrow_table(cell_id = a5_cell_to_arrow(cell)),
  tf
)

# Read it back naively: arrow silently converts uint64 to double
(naive <- tibble(arrow::read_parquet(tf)))

cell_as_dbl <- naive$cell_id

# The double can't distinguish this cell from nearby IDs
cell_as_dbl == cell_as_dbl + 1   # TRUE: silent corruption
cell_as_dbl == cell_as_dbl + 100 # still TRUE
```

## The solution: `a5_cell_from_arrow()` and `a5_cell_to_arrow()`

a5R provides two functions that bypass the lossy `double` conversion
entirely, using Arrow's zero-copy `View()` to reinterpret the raw bytes:

```{r bridge, eval = requireNamespace("arrow", quietly = TRUE)}
library(a5R)
library(tibble)

# Six cities across the globe; some will have bit 63 set (origin >= 6)
cities <- tibble(
  name = c("Edinburgh", "Tokyo", "São Paulo", "Nairobi", "Anchorage", "Sydney"),
  lon  = c(   -3.19,     139.69,     -46.63,     36.82,    -149.90,    151.21),
  lat  = c(   55.95,      35.69,     -23.55,     -1.29,      61.22,    -33.87)
)

cities$cell <- a5_lonlat_to_cell(cities$lon, cities$lat, resolution = 10)
cities
```

These cells work seamlessly in tibbles. Now let's enrich the data with
some A5 operations: cell resolution and distance from Edinburgh.

```{r enrich, eval = requireNamespace("arrow", quietly = TRUE)}
edinburgh <- cities$cell[1]

cities$resolution <- a5_get_resolution(cities$cell)
cities$dist_from_edinburgh_km <- as.numeric(
  a5_cell_distance(cities$cell, rep(edinburgh, nrow(cities)), units = "km")
)

cities
```

## Writing and reading Parquet

Convert to an Arrow table and write to Parquet. The cell column is stored
as native `uint64`, the same binary format used by DuckDB, Python, and
geoparquet.io:

```{r parquet_write, eval = requireNamespace("arrow", quietly = TRUE)}
tf <- tempfile(fileext = ".parquet")

arrow_tbl <- arrow::arrow_table(
  name = cities$name,
  cell_id = a5_cell_to_arrow(cities$cell),
  cell_res = cities$resolution,
  dist_from_edinburgh_km = cities$dist_from_edinburgh_km
)
arrow_tbl$schema
arrow::write_parquet(arrow_tbl, tf)
```

Read it back. `a5_cell_from_arrow()` recovers the exact cell IDs
without any precision loss:

```{r parquet_read, eval = requireNamespace("arrow", quietly = TRUE)}
pq <- arrow::read_parquet(tf, as_data_frame = FALSE)

# Recover cells from the uint64 column, bind with the rest of the data
recovered_cells <- a5_cell_from_arrow(pq$column(1))
result <- as.data.frame(pq)
result$cell <- recovered_cells
result <- tibble::as_tibble(result[c("name", "cell", "cell_res", "dist_from_edinburgh_km")])
result
```

Verify the round-trip is lossless:

```{r verify, eval = requireNamespace("arrow", quietly = TRUE)}
identical(format(cities$cell), format(result$cell))
```

## How it works under the hood

1. **`a5_cell_to_arrow()`**: packs the eight raw-byte fields into 8-byte
   little-endian blobs (one per cell), creates an Arrow
   `fixed_size_binary(8)` array, then uses `View(uint64)` to
   reinterpret the bytes as unsigned 64-bit integers. Zero-copy.

2. **`a5_cell_from_arrow()`**: does the reverse, applying
   `View(fixed_size_binary(8))` on the `uint64` array to get the raw
   bytes, then unpacks each 8-byte blob into the eight raw-byte fields
   used by `a5_cell`.

The raw bytes never pass through `double`, so there is no precision loss
at any step. See `vignette("internal-cell-representation")` for details
on the raw-byte representation.

---
title: "Getting started with a5R"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting started with a5R}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(a5R)
```

## Index a point

Map a longitude/latitude coordinate to a cell at a given resolution (0--30).
Higher resolutions produce smaller cells.

```{r}
cell <- a5_lonlat_to_cell(-3.19, 55.95, resolution = 10)
cell
```

Convert back to the cell centre point:

```{r}
a5_cell_to_lonlat(cell)
```

## Cell boundaries

Get the boundary polygon for one or more cells:

```{r boundary-plot, fig.width = 5, fig.height = 4}
boundary <- a5_cell_to_boundary(cell)
boundary

plot(boundary, col = "#206ead20", border = "#206ead", asp = 1)
```

Boundaries are returned as `wk_wkb` vectors by default (set `format = "wkt"` for
WKT). Both integrate directly with sf, terra, and other spatial tooling via the
wk package.

## Hierarchy and compaction

A5 is a hierarchical grid: every cell has a **parent** at a coarser
resolution and 4 **children** at the next finer resolution.

```{r}
parent <- a5_cell_to_parent(cell)
parent

children <- a5_cell_to_children(cell)
children
```

We can visualise the relationship: the parent (dark outline) contains
our cell (blue fill), which in turn contains its 4 children (orange):

```{r hierarchy-plot, fig.width = 8, fig.height = 8}
plot(NULL, xlim = c(-3.23, -3),  ylim = c(55.98, 55.99),               
       xlab = "", ylab = "", asp = 1
       )
plot(a5_cell_to_boundary(a5_cell_to_children(cell)),
     col = "#ad6e2020", border = "#ad6e20", add = TRUE)
plot(a5_cell_to_boundary(cell), col = "#206ead40", border = "#206ead",
     lwd = 2, add = TRUE)
plot(a5_cell_to_boundary(parent), border = "#333333", lwd = 2, add = TRUE)
```

Cell area decreases geometrically: each level is roughly 4x smaller.

```{r}
a5_cell_area(0:5)
```

### Compact and uncompact

When a complete set of siblings is present, `a5_compact()` merges them
back into their shared parent. This is the inverse of `a5_cell_to_children()`
and is useful for reducing the size of large cell sets without losing
coverage.

```{r}
children
a5_compact(children)

# round-trips back to the original
a5_uncompact(a5_compact(children), resolution = 11)
```

Many a5R functions return compacted output automatically. For example,
`a5_grid_disk()` and `a5_spherical_cap()` compact their results; use
`a5_uncompact()` when you need a uniform-resolution grid (see
[Traversal](#traversal) below).


## Traversal

Find neighbouring cells by hop count with `a5_grid_disk()`, or by
great-circle distance with `a5_spherical_cap()`:

```{r traversal, fig.width = 7}
disk <- a5_grid_disk(cell, k = 10)
cap <- a5_spherical_cap(cell, radius = 50000)

plot(a5_cell_to_boundary(cap), col = "#6ead2020", border = "#6ead20", asp = 1)
plot(a5_cell_to_boundary(disk), col = "#206ead20", border = "#206ead", asp = 1)
```

Both functions return a **compacted** cell vector: sibling groups are
merged into coarser parent cells to save space. To recover a uniform grid
at the original resolution, pass the result through `a5_uncompact()`:

```{r traversal-uncompact, fig.width = 7}
disk_grid <- a5_uncompact(disk, resolution = a5_get_resolution(cell))

plot(a5_cell_to_boundary(disk_grid), col = "#206ead20", border = "#206ead", asp = 1)
```


## Converting geometries to a5 cells

`a5_polygon_to_cells()` returns the cells whose centres lie inside a
polygon. It accepts any geometry that wk can handle (a `wk::rct()`
bounding box, a WKT or WKB polygon, an `sf` / `sfc` feature) and also
terra `SpatVector` polygons.

```{r grid-plot, fig.width = 7}
cells <- a5_polygon_to_cells(wk::rct(-3.3, 55.9, -3.1, 56.0), resolution = 12)
length(cells)

plot(a5_cell_to_boundary(cells), col = "#206ead20", border = "#206ead", asp = 1)
```

The returned vector is sorted and compacted: whenever four sibling
cells all sit inside the polygon, they are merged into their parent so
the result uses fewer slots without losing coverage. Call
`a5_uncompact()` to expand back to a uniform grid at the target
resolution:

```{r}
cells_uncom <- a5_uncompact(cells, 12)
length(cells_uncom)

plot(a5_cell_to_boundary(cells_uncom), col = "#206ead20", border = "#206ead", asp = 1)
```

Multi-part inputs are handled natively: a `MULTIPOLYGON` or an `sfc`
of multiple polygons returns the union of cells across all parts, and
a `POLYGON` with holes returns the outer ring's cells with the hole
cells properly subtracted.

```{r, eval = requireNamespace("sf", quietly = TRUE)}
library(sf)
demo(nc, ask = FALSE, echo = FALSE)
nca5 <- a5_polygon_to_cells(nc, resolution = 9) |>
  a5_uncompact(9)
plot(a5_cell_to_boundary(nca5), col = "#03030320", border = "#6d20adff", asp = 1)
```

## Tracing a line

`a5_linestring_to_cells()` returns the cells whose pentagons are
crossed by a great-circle polyline. Output is in discovery order along
the path. Consecutive waypoints are connected by great-circle arcs, so
antimeridian-crossing routes work transparently.

Multi-part inputs (`MULTILINESTRING` or an `sfc` of several linestrings)
are handled natively: per-feature outputs are concatenated in feature
order with first-seen deduplication.

To show what cell-by-cell tracing looks like, we can write a short
phrase as a `MULTILINESTRING` (one stroke per letter) and ask A5 to
fill in the cells:

```{r linestring-a5R, fig.width = 7, fig.height = 3}
a5R_strokes <- wk::wkt(
  "MULTILINESTRING (
    (1.8 52.5, 1.2 52.8, 0.6 52.5, 0.4 52, 0.6 51.5, 1.2 51.2, 1.8 51.5, 2 52, 1.8 52.5, 2 51.2),
    (5 53, 3.5 53, 3.5 52.2, 4 52.2, 4.7 52, 5 51.6, 4.7 51.2, 4 51, 3.5 51.1),
    (6.5 51, 6.5 53, 7.5 53, 8 52.8, 8.2 52.5, 8 52.2, 7.5 52, 6.5 52, 8.2 51)
  )"
)

cells <- a5_linestring_to_cells(a5R_strokes, resolution = 9)
length(cells)

plot(a5_cell_to_boundary(cells),
     col = "#206ead80", border = "#37af6d", asp = 1)
```

Each stroke is one continuous linestring; the function walks the
great-circle path, expanding cell neighbours along the way, and
returns the union across all strokes.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/regions.R
\name{a5_polygon_to_cells}
\alias{a5_polygon_to_cells}
\title{Cells whose centres lie inside a polygon}
\usage{
a5_polygon_to_cells(x, resolution)
}
\arguments{
\item{x}{A polygon-like geometry. One of:
\itemize{
\item Any geometry handleable by \code{\link[wk:wk_handle]{wk::wk_handle()}} (e.g. \code{\link[wk:wkt]{wk::wkt()}},
\code{\link[wk:wkb]{wk::wkb()}}, \code{\link[wk:rct]{wk::rct()}}, \code{sf}, \code{sfc}) containing one or more
\code{POLYGON} / \code{MULTIPOLYGON} features.
\item A \code{SpatVector} of polygons (requires the \code{terra} package).
\item A two-column numeric matrix (\code{cbind(lon, lat)}) of vertices,
interpreted as a single outer ring.
\item A \code{data.frame} with columns \code{lon} and \code{lat}, interpreted as a
single outer ring.
}}

\item{resolution}{Integer scalar target resolution (0-30).}
}
\value{
An \link{a5_cell} vector at or coarser than \code{resolution}.
}
\description{
Returns A5 cells at \code{resolution} whose centres fall inside the polygon.
Multi-feature inputs (a \code{MULTIPOLYGON}, an \code{sfc} of multiple polygons,
or a \code{POLYGON} with holes) are handled natively: per polygon part, the
outer-ring cells are computed and any hole-ring cells are subtracted,
then the results are unioned across parts. The final cell set is
compacted; use \code{\link[=a5_uncompact]{a5_uncompact()}} to expand to a uniform-resolution grid.
}
\details{
Membership is determined by \strong{centre-point containment}: a cell is
included iff its centroid lies inside the polygon, with hole rings
properly subtracted. This is distinct from \code{\link[=a5_grid]{a5_grid()}}'s
boundary-intersection semantics; for the same polygon the two
functions can return slightly different cell sets near the boundary.

Coordinates must be WGS 84 longitude/latitude in degrees. Rings are
closed automatically; a trailing duplicate vertex is dropped if
present.

Where no A5 cell centroids at the specified \code{resolution} fall within
the geometry, an empty \code{a5_cell} vector is returned.

Matrix and \code{data.frame} inputs are treated as a single ring; for
multi-feature data or polygons with holes, pass an \code{sf}, \code{sfc}, wk,
or \code{SpatVector} geometry instead.
}
\examples{
poly <- wk::wkt(
  "POLYGON ((-3.3 55.9, -3.1 55.9, -3.1 56, -3.3 56, -3.3 55.9))"
)
cells <- a5_polygon_to_cells(poly, resolution = 8)
length(cells)
}
\seealso{
\code{\link[=a5_linestring_to_cells]{a5_linestring_to_cells()}}, \code{\link[=a5_uncompact]{a5_uncompact()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hierarchy.R
\name{a5_cell_to_children}
\alias{a5_cell_to_children}
\title{Get child cells}
\usage{
a5_cell_to_children(cell, resolution = NULL)
}
\arguments{
\item{cell}{A single \link{a5_cell} value.}

\item{resolution}{Integer scalar target child resolution, or \code{NULL} for
immediate children.}
}
\value{
An \link{a5_cell} vector of child cells.
}
\description{
Returns the child cells of a single cell. By default returns the 4
immediate children (one resolution finer). Optionally target a specific
finer resolution.
}
\examples{
cell <- a5_lonlat_to_cell(-3.19, 55.95, resolution = 5)
a5_cell_to_children(cell)
}
\seealso{
\code{\link[=a5_cell_to_parent]{a5_cell_to_parent()}}, \code{\link[=a5_get_resolution]{a5_get_resolution()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compact.R
\name{a5_uncompact}
\alias{a5_uncompact}
\title{Uncompact a set of A5 cells to a target resolution}
\usage{
a5_uncompact(cells, resolution)
}
\arguments{
\item{cells}{An \link{a5_cell} vector.}

\item{resolution}{Integer scalar target resolution (0--30).}
}
\value{
An \link{a5_cell} vector of uncompacted cells.
}
\description{
Expands each cell to its descendants at the target resolution.
}
\examples{
cell <- a5_lonlat_to_cell(-3.19, 55.95, resolution = 5)
a5_uncompact(cell, resolution = 7)
}
\seealso{
\code{\link[=a5_compact]{a5_compact()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/arrow.R
\name{a5_cell_from_arrow}
\alias{a5_cell_from_arrow}
\alias{a5_cell_to_arrow}
\title{Convert between a5_cell and Arrow uint64 arrays}
\usage{
a5_cell_from_arrow(x)

a5_cell_to_arrow(x)
}
\arguments{
\item{x}{For \code{a5_cell_from_arrow()}, an Arrow \code{Array} or
\code{ChunkedArray} of type \code{uint64}. For \code{a5_cell_to_arrow()}, an
\link{a5_cell} vector.}
}
\value{
\code{a5_cell_from_arrow()} returns an \link{a5_cell} vector.
\code{a5_cell_to_arrow()} returns an Arrow \code{Array} of type \code{uint64}.
}
\description{
Losslessly convert between \link{a5_cell} vectors and Arrow \code{uint64}
arrays. This avoids the precision loss that occurs when Arrow
converts \code{uint64} to R's \code{double} (which can only represent
integers exactly up to 2^53, while A5 cell IDs span the full
0--2^64 range).
}
\details{
Internally these use Arrow's zero-copy \code{View()} to reinterpret
\code{uint64} bytes as \code{fixed_size_binary(8)}, then convert to/from the
raw-byte representation used by \link{a5_cell}. The resulting Arrow
arrays can be written directly to Parquet and read correctly by
DuckDB, Python, and other Arrow-compatible tools.
}
\examples{
\dontshow{if (requireNamespace("arrow", quietly = TRUE)) withAutoprint(\{ # examplesIf}
cell <- a5_lonlat_to_cell(135, 0, resolution = 10)
arr <- a5_cell_to_arrow(cell)
back <- a5_cell_from_arrow(arr)
identical(format(cell), format(back))
\dontshow{\}) # examplesIf}
\dontshow{if (requireNamespace("arrow", quietly = TRUE)) withAutoprint(\{ # examplesIf}
cells <- a5_lonlat_to_cell(c(-3.19, 135), c(55.95, 0), resolution = 10)
arr <- a5_cell_to_arrow(cells)
arr$type$ToString()
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=a5_u64_to_hex]{a5_u64_to_hex()}} for converting to hex strings instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compact.R
\name{a5_compact}
\alias{a5_compact}
\title{Compact a set of A5 cells}
\usage{
a5_compact(cells)
}
\arguments{
\item{cells}{An \link{a5_cell} vector.}
}
\value{
An \link{a5_cell} vector of compacted cells.
}
\description{
Merges complete sibling groups into their common parent, reducing the
number of cells while preserving coverage.
}
\examples{
cell <- a5_lonlat_to_cell(-3.19, 55.95, resolution = 5)
children <- a5_cell_to_children(cell)
a5_compact(children) # back to the parent
}
\seealso{
\code{\link[=a5_uncompact]{a5_uncompact()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/a5R-package.R
\docType{package}
\name{a5R-package}
\alias{a5R}
\alias{a5R-package}
\title{a5R: A5 Pentagonal Geospatial Index for R}
\description{
R bindings for the \href{https://a5geo.org/}{A5} pentagonal geospatial index /
discrete global grid system, powered by the a5 Rust crate via extendr.
}
\section{Cell type}{

\itemize{
\item \code{\link[=a5_cell]{a5_cell()}} --- construct cell ID vectors
\item \link[=a5_cell]{is_a5_cell()} / \link[=a5_cell]{a5_is_valid()} --- type test and
validation
\item \code{\link[=a5_u64_to_hex]{a5_u64_to_hex()}} / \link[=a5_u64_to_hex]{a5_hex_to_u64()} --- hex string
conversion
}
}

\section{Indexing}{

\itemize{
\item \code{\link[=a5_lonlat_to_cell]{a5_lonlat_to_cell()}} --- coordinates to cell IDs
\item \code{\link[=a5_cell_to_lonlat]{a5_cell_to_lonlat()}} --- cell IDs to centre coordinates
}
}

\section{Geometry}{

\itemize{
\item \code{\link[=a5_cell_to_boundary]{a5_cell_to_boundary()}} --- cell boundary polygons (WKB or WKT)
\item \code{\link[=a5_cell_area]{a5_cell_area()}} --- cell area at a given resolution
\item \code{\link[=a5_cell_distance]{a5_cell_distance()}} --- distance between cell centroids
\item \code{\link[=a5_get_num_cells]{a5_get_num_cells()}} --- total cell count at a resolution
\item \code{\link[=a5_get_num_children]{a5_get_num_children()}} --- child count between resolutions
}
}

\section{Hierarchy}{

\itemize{
\item \code{\link[=a5_get_resolution]{a5_get_resolution()}} --- extract resolution from cell IDs
\item \code{\link[=a5_cell_to_parent]{a5_cell_to_parent()}} --- navigate to coarser cells
\item \code{\link[=a5_cell_to_children]{a5_cell_to_children()}} --- navigate to finer cells
\item \code{\link[=a5_get_res0_cells]{a5_get_res0_cells()}} --- the 12 root cells
\item \code{\link[=a5_compact]{a5_compact()}} / \code{\link[=a5_uncompact]{a5_uncompact()}} --- compress and expand cell sets
}
}

\section{Traversal}{

\itemize{
\item \code{\link[=a5_grid_disk]{a5_grid_disk()}} --- neighbours by hop count
\item \code{\link[=a5_spherical_cap]{a5_spherical_cap()}} --- neighbours by great-circle distance
}
}

\section{Geometry indexing}{

\itemize{
\item \code{\link[=a5_polygon_to_cells]{a5_polygon_to_cells()}} --- cells whose centres lie inside a polygon
\item \code{\link[=a5_linestring_to_cells]{a5_linestring_to_cells()}} --- cells crossed by a great-circle polyline
}
}

\section{Arrow & Parquet}{

\itemize{
\item \code{\link[=a5_cell_from_arrow]{a5_cell_from_arrow()}} / \link[=a5_cell_from_arrow]{a5_cell_to_arrow()} ---
lossless conversion to/from Arrow \code{uint64}
}
}

\section{Configuration}{

\itemize{
\item \code{\link[=a5_set_threads]{a5_set_threads()}} / \code{\link[=a5_get_threads]{a5_get_threads()}} --- multi-threading control
}
}

\section{Vignettes}{

\itemize{
\item \code{vignette("a5R")} --- getting started
\item \code{vignette("multithreading")} --- parallel processing
\item \code{vignette("internal-cell-representation")} --- how cell IDs are stored
\item \code{vignette("arrow-parquet")} --- Arrow and Parquet interop
}
}

\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/belian-earth/a5R}
  \item \url{https://belian-earth.github.io/a5R/}
  \item Report bugs at \url{https://github.com/belian-earth/a5R/issues}
}

}
\author{
\strong{Maintainer}: Hugh Graham \email{hugh@belian.earth}

Authors:
\itemize{
  \item Hugh Graham \email{hugh@belian.earth}
}

Other contributors:
\itemize{
  \item belian.earth [copyright holder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/boundary.R
\name{a5_cell_to_boundary}
\alias{a5_cell_to_boundary}
\title{Get cell boundary polygons}
\usage{
a5_cell_to_boundary(
  cell,
  format = c("wkb", "wkt"),
  closed = TRUE,
  segments = NULL
)
}
\arguments{
\item{cell}{An \link{a5_cell} vector.}

\item{format}{Character scalar, either \code{"wkb"} (default) or \code{"wkt"}.}

\item{closed}{Logical scalar; if \code{TRUE} (default) the ring is closed
(first vertex repeated at end).}

\item{segments}{Integer scalar or \code{NULL}. Number of interpolation segments
per edge for geodesic accuracy. \code{NULL} uses the default (straight edges).}
}
\value{
A \code{wk_wkt} or \code{wk_wkb} vector of polygon geometries with
\code{wk::wk_crs_longlat()} CRS.
}
\description{
Returns the boundary of each cell as a \code{\link[wk:wkt]{wk::wkt()}} or \code{\link[wk:wkb]{wk::wkb()}}
polygon geometry. Boundaries are pentagonal polygons on the WGS 84
ellipsoid.
}
\examples{
cell <- a5_lonlat_to_cell(-3.19, 55.95, resolution = 5)
a5_cell_to_boundary(cell)
a5_cell_to_boundary(cell, format = "wkt")
}
\seealso{
\code{\link[=a5_cell_to_lonlat]{a5_cell_to_lonlat()}} for cell centroids.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/info.R
\name{a5_cell_area}
\alias{a5_cell_area}
\title{Cell area at a given resolution}
\usage{
a5_cell_area(resolution, units = "m^2")
}
\arguments{
\item{resolution}{Integer vector of resolutions (0--30).}

\item{units}{Character scalar specifying the output area unit (default
\code{"m^2"}). Any unit convertible from \code{m^2} via \code{\link[units:set_units]{units::set_units()}} is
accepted (e.g. \code{"km^2"}, \code{"ha"}, \code{"acre"}).  If NULL, the area is
returned as a numeric vector in m^2.}
}
\value{
A \link[units:units]{units::units} vector of areas.
}
\description{
Returns the area of a single cell in square metres at the given
resolution(s). Because A5 is an equal-area DGGS, all cells at the same
resolution have identical area.
}
\examples{
a5_cell_area(0:5)
a5_cell_area(5, units = "km^2")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/threads.R
\name{a5_set_threads}
\alias{a5_set_threads}
\alias{a5_get_threads}
\title{Set the number of threads used by a5R}
\usage{
a5_set_threads(n = 1L)

a5_get_threads()
}
\arguments{
\item{n}{Integer scalar. Number of threads. Must be >= 1.}
}
\value{
Invisibly returns the previous thread count.

Integer scalar.
}
\description{
Controls the number of threads used for parallel processing in vectorised
functions. Set to 1 (the default) for sequential processing with zero
overhead, or higher for parallel execution via rayon.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/info.R
\name{a5_get_num_children}
\alias{a5_get_num_children}
\title{Number of children between two resolutions}
\usage{
a5_get_num_children(parent_resolution, child_resolution)
}
\arguments{
\item{parent_resolution}{Integer scalar (0--30).}

\item{child_resolution}{Integer scalar (0--30), must be >=
\code{parent_resolution}.}
}
\value{
A numeric scalar. Returned as double because the count can
exceed R's integer range at large resolution deltas.
}
\description{
Returns the number of child cells each parent cell contains when
expanding from one resolution to another.
}
\examples{
a5_get_num_children(5, 8)   # 4^3 = 64
a5_get_num_children(0, 5)
}
\seealso{
\code{\link[=a5_get_num_cells]{a5_get_num_cells()}}, \code{\link[=a5_cell_to_children]{a5_cell_to_children()}},
\code{\link[=a5_uncompact]{a5_uncompact()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cell.R
\name{a5_cell}
\alias{a5_cell}
\alias{is_a5_cell}
\alias{as_a5_cell}
\alias{a5_is_valid}
\title{A5 Cell Index Vector}
\usage{
a5_cell(x = character())

is_a5_cell(x)

as_a5_cell(x)

a5_is_valid(x)
}
\arguments{
\item{x}{A character vector of hex-encoded A5 cell IDs, or an object
coercible to one.}
}
\value{
An \code{a5_cell} vector (\code{a5_cell}, \code{as_a5_cell}), a logical
scalar (\code{is_a5_cell}), or a logical vector (\code{a5_is_valid}).
}
\description{
Create, test, and coerce A5 cell index vectors. Cells are stored as
a record with eight raw-byte fields (\code{b1}--\code{b8}) representing the
little-endian bytes of the u64 cell ID. This avoids the precision
loss of floating-point storage and keeps memory compact.
}
\examples{
cells <- a5_cell(c("0800000000000006", "0800000000000016"))
cells
a5_is_valid(c("0800000000000006", "not_a_cell", NA))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distance.R
\name{a5_cell_distance}
\alias{a5_cell_distance}
\title{Distance between cell centroids}
\usage{
a5_cell_distance(
  from,
  to,
  units = "m",
  method = c("haversine", "geodesic", "rhumb")
)
}
\arguments{
\item{from, to}{\link{a5_cell} vectors (recycled to common length).}

\item{units}{Character scalar specifying the distance unit (default
\code{"m"}). Any unit convertible from metres via \code{\link[units:set_units]{units::set_units()}} is
accepted (e.g. \code{"km"}, \code{"mi"}). If NULL, the distance is returned as a
numeric vector in metres.}

\item{method}{Distance calculation method. One of \code{"haversine"}
(great-circle, default), \code{"geodesic"} (WGS84 ellipsoid via Karney
2013), or \code{"rhumb"} (loxodrome / constant-bearing).}
}
\value{
A \link[units:units]{units::units} vector of distances.
}
\description{
Computes the distance between the centroids of pairs of A5 cells
using the specified method.
}
\examples{
a <- a5_lonlat_to_cell(-3.19, 55.95, resolution = 24)
b <- a5_lonlat_to_cell(-3.10, 55.90, resolution = 24)
a5_cell_distance(a, b)
a5_cell_distance(a, b, units = "km")
a5_cell_distance(a, b, method = "geodesic")
}
\seealso{
\code{\link[=a5_cell_to_lonlat]{a5_cell_to_lonlat()}} for cell centroids,
\code{\link[=a5_cell_area]{a5_cell_area()}} for cell areas.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/traversal.R
\name{a5_grid_disk}
\alias{a5_grid_disk}
\title{Cells within k hops of a cell}
\usage{
a5_grid_disk(cell, k, vertex = FALSE)
}
\arguments{
\item{cell}{A single \link{a5_cell} value.}

\item{k}{Integer scalar, number of hops.}

\item{vertex}{Logical scalar. If \code{FALSE} (default), only edge-sharing
neighbours (4-connected) are traversed. If \code{TRUE}, vertex-sharing
neighbours are included (8-connected).}
}
\value{
A compacted \link{a5_cell} vector.
}
\description{
Returns all cells reachable within \code{k} edge hops of a centre cell,
including the centre cell itself.
}
\examples{
cell <- a5_lonlat_to_cell(-3.19, 55.95, resolution = 8)
a5_grid_disk(cell, k = 1)
}
\seealso{
\code{\link[=a5_spherical_cap]{a5_spherical_cap()}} for distance-based selection.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wk.R
\name{wk_methods}
\alias{wk_methods}
\alias{wk_handle.a5_cell}
\alias{wk_crs.a5_cell}
\title{wk methods for a5_cell}
\usage{
\method{wk_handle}{a5_cell}(handleable, handler, ...)

\method{wk_crs}{a5_cell}(x)
}
\arguments{
\item{handleable, x}{An \link{a5_cell} vector.}

\item{handler}{A \link[wk:wk_handle]{wk handler}.}

\item{...}{Passed to underlying methods.}
}
\value{
\itemize{
\item \code{wk_handle()}: the result of the handler.
\item \code{wk_crs()}: a \link[wk:wk_crs]{wk::wk_crs} object (WGS 84 lon/lat).
}
}
\description{
Integration with the \link[wk:wk-package]{wk} geometry framework. Allows
\code{a5_cell} vectors to be handled as geometry (via their boundary polygons)
and to report their CRS.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/traversal.R
\name{a5_spherical_cap}
\alias{a5_spherical_cap}
\title{Cells within a great-circle radius}
\usage{
a5_spherical_cap(cell, radius)
}
\arguments{
\item{cell}{A single \link{a5_cell} value.}

\item{radius}{Numeric scalar, great-circle radius in metres.}
}
\value{
A compacted \link{a5_cell} vector.
}
\description{
Returns all cells whose centres fall within a great-circle distance
of a given cell's centre.
}
\examples{
cell <- a5_lonlat_to_cell(-3.19, 55.95, resolution = 8)
a5_spherical_cap(cell, radius = 1000)
}
\seealso{
\code{\link[=a5_grid_disk]{a5_grid_disk()}} for hop-based selection.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/info.R
\name{a5_get_num_cells}
\alias{a5_get_num_cells}
\title{Total number of cells at a given resolution}
\usage{
a5_get_num_cells(resolution)
}
\arguments{
\item{resolution}{Integer scalar resolution (0--30).}
}
\value{
A numeric scalar (double) giving the total count. Returned as
double because the count can exceed R's integer range.
}
\description{
Total number of cells at a given resolution
}
\examples{
a5_get_num_cells(0)
a5_get_num_cells(10)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/indexing.R
\name{a5_lonlat_to_cell}
\alias{a5_lonlat_to_cell}
\title{Convert coordinates to A5 cell indices}
\usage{
a5_lonlat_to_cell(lon, lat, resolution)
}
\arguments{
\item{lon}{Numeric vector of longitudes in degrees.}

\item{lat}{Numeric vector of latitudes in degrees.}

\item{resolution}{Integer scalar or vector of resolutions (0--30).}
}
\value{
An \link{a5_cell} vector of cell indices.
}
\description{
Maps longitude/latitude coordinates to A5 cell indices at the specified
resolution.
}
\examples{
a5_lonlat_to_cell(-3.19, 55.95, resolution = 5)
}
\seealso{
\code{\link[=a5_cell_to_lonlat]{a5_cell_to_lonlat()}} for the inverse operation.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/info.R
\name{a5_get_res0_cells}
\alias{a5_get_res0_cells}
\title{Get all resolution-0 root cells}
\usage{
a5_get_res0_cells()
}
\value{
An \link{a5_cell} vector of length 12.
}
\description{
Returns the 12 root cells corresponding to the 12 faces of the
dodecahedron.
}
\examples{
a5_get_res0_cells()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hierarchy.R
\name{a5_cell_to_parent}
\alias{a5_cell_to_parent}
\title{Navigate to parent cell(s)}
\usage{
a5_cell_to_parent(cell, resolution = NULL)
}
\arguments{
\item{cell}{An \link{a5_cell} vector.}

\item{resolution}{Integer scalar target parent resolution, or \code{NULL} for
the immediate parent.}
}
\value{
An \link{a5_cell} vector of parent cells.
}
\description{
Returns the parent cell of each input cell. By default returns the
immediate parent (one resolution coarser). Optionally target a specific
coarser resolution.
}
\examples{
cell <- a5_lonlat_to_cell(-3.19, 55.95, resolution = 10)
a5_cell_to_parent(cell)
a5_cell_to_parent(cell, resolution = 5)
}
\seealso{
\code{\link[=a5_cell_to_children]{a5_cell_to_children()}}, \code{\link[=a5_get_resolution]{a5_get_resolution()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/indexing.R
\name{a5_cell_to_lonlat}
\alias{a5_cell_to_lonlat}
\title{Convert A5 cell indices to coordinates}
\usage{
a5_cell_to_lonlat(cell, as_dataframe = FALSE)
}
\arguments{
\item{cell}{An \link{a5_cell} vector (or character coercible to one).}

\item{as_dataframe}{Logical scalar controlling the return container.
When \code{FALSE} (default), centroids are returned as a \code{\link[wk:xy]{wk::xy()}} vector
with WGS 84 CRS, the geographic-typed form that plugs into wk/sf
pipelines. When \code{TRUE}, centroids are returned as a base \code{data.frame}
with columns \code{lon} and \code{lat}.}
}
\value{
A \code{\link[wk:xy]{wk::xy()}} vector (if \code{as_dataframe = FALSE}) or a
\code{data.frame} with columns \code{lon} and \code{lat} (if \code{as_dataframe = TRUE}).
}
\description{
Returns the centre-point longitude and latitude of each cell.
}
\examples{
cell <- a5_lonlat_to_cell(-3.19, 55.95, resolution = 5)
a5_cell_to_lonlat(cell)

# Data frame output
cell2 <- a5_lonlat_to_cell(114.8, 4.1, resolution = 5)
a5_cell_to_lonlat(cell2, as_dataframe = TRUE)
}
\seealso{
\code{\link[=a5_lonlat_to_cell]{a5_lonlat_to_cell()}} for the inverse operation,
\code{\link[=a5_cell_to_boundary]{a5_cell_to_boundary()}} for full cell polygons.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hierarchy.R
\name{a5_get_resolution}
\alias{a5_get_resolution}
\title{Get the resolution of A5 cell indices}
\usage{
a5_get_resolution(cell)
}
\arguments{
\item{cell}{An \link{a5_cell} vector.}
}
\value{
An integer vector of resolutions.
}
\description{
Extracts the resolution level (0--30) encoded in each cell index.
}
\examples{
cell <- a5_lonlat_to_cell(-3.19, 55.95, resolution = 10)
a5_get_resolution(cell)
}
\seealso{
\code{\link[=a5_cell_to_parent]{a5_cell_to_parent()}}, \code{\link[=a5_cell_to_children]{a5_cell_to_children()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/regions.R
\name{a5_linestring_to_cells}
\alias{a5_linestring_to_cells}
\title{Cells traced by a great-circle linestring}
\usage{
a5_linestring_to_cells(x, resolution)
}
\arguments{
\item{x}{A linestring-like geometry. One of:
\itemize{
\item Any geometry handleable by \code{\link[wk:wk_handle]{wk::wk_handle()}} containing one or
more \code{LINESTRING} / \code{MULTILINESTRING} features.
\item A \code{SpatVector} of linestrings (requires the \code{terra} package).
\item A two-column numeric matrix (\code{cbind(lon, lat)}) of waypoints.
\item A \code{data.frame} with columns \code{lon} and \code{lat}.
}}

\item{resolution}{Integer scalar target resolution (0--30).}
}
\value{
An \link{a5_cell} vector at \code{resolution}.
}
\description{
Returns the A5 cells at \code{resolution} whose pentagons are intersected by
the great-circle polyline connecting the supplied waypoints. Output is
uncompacted, in discovery order along the path, with duplicates removed
first-seen. Multi-feature inputs (a \code{MULTILINESTRING} or an \code{sfc} of
multiple linestrings) are handled natively: per-feature cell sequences
are concatenated in feature order with first-seen deduplication.
}
\details{
Consecutive waypoints are connected by great-circle arcs (not rhumb
lines or planar segments), so antimeridian-crossing paths work
transparently when written in unwrapped lon/lat.

Matrix and \code{data.frame} inputs are treated as a single linestring;
for multi-feature data, pass an \code{sf}, \code{sfc}, wk, or \code{SpatVector}
geometry instead.
}
\examples{
line <- wk::wkt("LINESTRING (2.35 48.86, -0.13 51.51)")
cells <- a5_linestring_to_cells(line, resolution = 6)
length(cells)
}
\seealso{
\code{\link[=a5_polygon_to_cells]{a5_polygon_to_cells()}}, \code{\link[=a5_grid_disk]{a5_grid_disk()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cell.R
\name{a5_u64_to_hex}
\alias{a5_u64_to_hex}
\alias{a5_hex_to_u64}
\title{Coerce between hex strings and A5 cell vectors}
\usage{
a5_u64_to_hex(x)

a5_hex_to_u64(x)
}
\arguments{
\item{x}{For \code{a5_u64_to_hex()}, an \link{a5_cell} vector (or object
coercible to one). For \code{a5_hex_to_u64()}, a character vector of
hex-encoded cell IDs.}
}
\value{
\code{a5_u64_to_hex()} returns a character vector. \code{a5_hex_to_u64()}
returns an \link{a5_cell} vector.
}
\description{
\code{a5_u64_to_hex()} converts an \link{a5_cell} vector to 16-character
zero-padded hex strings. \code{a5_hex_to_u64()} converts hex strings to
an \link{a5_cell} vector.
}
\details{
These are named to match \code{u64_to_hex} / \code{hex_to_u64} in the upstream
Python, JavaScript, and DuckDB A5 bindings. In those languages the
functions convert between a native 64-bit unsigned integer and its hex
representation. Because R has no native \code{uint64} type, \code{a5_u64_to_hex()}
accepts an \link{a5_cell} (which stores the \code{u64} internally as eight raw
bytes) instead of a bare integer.
}
\examples{
cell <- a5_lonlat_to_cell(-3.19, 55.95, resolution = 5)
hex <- a5_u64_to_hex(cell)
hex

a5_hex_to_u64(hex)
}
\seealso{
\code{\link[=a5_cell_from_arrow]{a5_cell_from_arrow()}} and \code{\link[=a5_cell_to_arrow]{a5_cell_to_arrow()}} for lossless
conversion between \link{a5_cell} and Arrow \code{uint64} arrays.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid.R
\name{a5_grid}
\alias{a5_grid}
\title{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}}
\usage{
a5_grid(x, resolution)
}
\arguments{
\item{x}{An area specification. One of:
\itemize{
\item A numeric vector of length 4 (\code{c(xmin, ymin, xmax, ymax)}) interpreted
as a WGS 84 bounding box.
\item Any geometry handleable by \code{\link[wk:wk_handle]{wk::wk_handle()}} (e.g. \code{\link[wk:wkt]{wk::wkt()}},
\code{\link[wk:wkb]{wk::wkb()}}, \code{sfc}, \code{sf}, \link{a5_cell}).
}}

\item{resolution}{Integer scalar target resolution (0--30).}
}
\value{
An \link{a5_cell} vector of cells at \code{resolution} that intersect \code{x}.
}
\description{
Generate a grid of A5 cells covering an area
}
\details{
Returns all cells at the target resolution that intersect the given
geometry. Uses hierarchical flood-fill: starting from the 12 resolution-0
root cells, the algorithm repeatedly expands and prunes by spatial
intersection until the target resolution is reached.

Grid generation runs entirely in Rust via hierarchical flood-fill with
bounding-box pruning. For non-bbox geometry inputs, an exact intersection
filter removes cells that fall outside the target shape. No cell count
limit is imposed — high resolutions over large areas can consume
significant memory.

Input geometries must use WGS 84 coordinates; projected geometries are
not reprojected and will produce incorrect results. Multiple geometries
are collected into a GEOMETRYCOLLECTION automatically.
Antimeridian-crossing bounding boxes are supported: when \code{xmin > xmax}
(e.g. \code{c(170, -50, -170, -30)}), the bbox is split at the antimeridian.

\strong{Limitation:} spatial filtering uses planar geometry on lon/lat
coordinates, which can produce incomplete results very close to the poles
(above ~88° latitude) or the antimeridian. Use a larger target geometry
to ensure complete coverage in these areas.
}
\examples{
# Grid from a bounding box
cells <- a5_grid(c(-3.3, 55.9, -3.1, 56.0), resolution = 5)
cells

# Grid from a WKT polygon
poly <- wk::wkt("POLYGON ((-3.3 55.9, -3.1 55.9, -3.1 56, -3.3 56, -3.3 55.9))")
cells <- a5_grid(poly, resolution = 5)
}
\seealso{
\code{\link[=a5_cell_to_boundary]{a5_cell_to_boundary()}} to convert result cells to geometries.
}
